@@ -25,11 +25,34 @@
 #include <string.h>
 #include <unistd.h>
 #include <fcntl.h>
+
+/* samhainctl */
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <sys/wait.h>
+#include <signal.h>
+#include <errno.h>
+
+
+#if TIME_WITH_SYS_TIME
+#include <sys/time.h>
 #include <time.h>
+#else
+#if HAVE_SYS_TIME_H
+#include <sys/time.h>
+#else
+#include <time.h>
+#endif
+#endif
+
 #ifdef HAVE_MEMORY_H
 #include <memory.h>
 #endif
 
+#ifdef HAVE_SETPRIORITY
+#include <sys/resource.h>
+#endif
+
 
 #include "samhain.h"
 #include "sh_files.h"
@@ -53,6 +76,7 @@
 #endif
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
 #include "sh_modules.h"
+#include "sh_ignore.h"
 #endif
 
 #undef  FIL__
@@ -75,11 +99,14 @@ extern UINT32  ErrFlag[2];
  **************************************************/
 
 volatile  int      sig_raised;
-volatile  int      sig_debug_on;           /* SIGUSR1 */
-volatile  int      sig_debug_off;          /* SIGUSR2 */
+volatile  int      sig_urgent;
+volatile  int      sig_debug_switch;       /* SIGUSR1 */
+volatile  int      sig_suspend_switch;     /* SIGUSR2 */
+volatile  int      sh_global_suspend_flag;
 volatile  int      sig_fresh_trail;        /* SIGIOT  */
 volatile  int      sig_config_read_again;  /* SIGHUP  */
 volatile  int      sig_terminate;          /* SIGQUIT */
+volatile  int      sig_termfast;           /* SIGTERM */
 long int           eintr__result;
 char               sh_sig_msg[SH_MINIBUF];
 
@@ -97,16 +124,16 @@ char               sh_sig_msg[SH_MINIBUF];
 #endif
 
 #ifndef GLOB_LEN
-#define GLOB_LEN 255
+#define GLOB_LEN 511
 #endif
 
 char * globber(char * str)
 {
-  int i;
-  int j;
+  size_t i;
+  size_t j;
 
-  static   int  items = 0;
-  static   int  count = 0;
+  static   size_t  items = 0;
+  static   size_t  count = 0;
   static   char glob[SH_MAX_GLOBS * (GLOB_LEN+1)];
 
   if (str == NULL)
@@ -150,7 +177,9 @@ void sh_do_encode (char * str, int len)
   /* this is a symmetric operation
    */
   for (i = 0; i < len; ++i)
-    str[i] = str[i] ^ XOR_CODE;
+    {
+      str[i] = str[i] ^ XOR_CODE;
+    }
   return;
 }
 
@@ -163,7 +192,7 @@ void sh_do_encode (char * str, int len)
  **************************************************/
 
 sh_struct   sh;
-sh_key_t  * skey = NULL;
+/*@null@*/ sh_key_t  * skey = NULL;
 
 extern unsigned char TcpFlag[8][PW_LEN+1];
 
@@ -173,6 +202,8 @@ extern unsigned char TcpFlag[8][PW_LEN+1];
  *
  **************************************************/
 
+static int is_samhainctl_init = S_FALSE;
+
 static
 void sh_init (void)
 {
@@ -183,6 +214,8 @@ void sh_init (void)
   char            q[SH_PATHBUF];
 #endif
 
+  SL_ENTER(_("sh_init"));
+
 #ifdef MKA_09
   ErrFlag[0] |= (1 << 8);
 #endif
@@ -210,12 +243,14 @@ void sh_init (void)
 
   /* Signal handling.
    */
-  sig_raised            = 0;
-  sig_config_read_again = 0;           /* SIGHUP  */
-  sig_debug_on          = 0;           /* SIGUSR1 */
-  sig_debug_off         = 0;           /* SIGUSR2 */
-  sig_fresh_trail       = 0;           /* SIGIOT  */
-  sig_terminate         = 0;           /* SIGQUIT */
+  sig_raised             = 0;
+  sig_config_read_again  = 0;           /* SIGHUP  */
+  sig_debug_switch       = 0;           /* SIGUSR1 */
+  sig_suspend_switch     = 0;           /* SIGUSR2 */
+  sh_global_suspend_flag = 0;           /* SIGUSR2 */
+  sig_fresh_trail        = 0;           /* SIGIOT  */
+  sig_terminate          = 0;           /* SIGQUIT */
+  sig_termfast           = 0;           /* SIGTERM */
   strcpy ( sh_sig_msg, _("None"));
 
 #ifdef MKB_01
@@ -253,10 +288,12 @@ void sh_init (void)
 
   /* The flags.
    */
-  sh.flag.checkSum        = SH_CHECK_NONE;
+  if (is_samhainctl_init == S_FALSE)
+    sh.flag.checkSum        = SH_CHECK_NONE;
   sh.flag.update          = S_FALSE;
   sh.flag.opts            = S_FALSE;
-  sh.flag.isdaemon        = S_FALSE;
+  if (is_samhainctl_init == S_FALSE)
+    sh.flag.isdaemon        = S_FALSE;
   sh.flag.isserver        = S_FALSE;
   sh.flag.islocked        = S_FALSE;
   sh.flag.smsg            = S_FALSE;
@@ -264,10 +301,12 @@ void sh_init (void)
   sh.flag.reportonce      = S_TRUE;
   sh.flag.fulldetail      = S_FALSE;
   sh.flag.audit           = S_FALSE;
+  sh.flag.nice            = 0;
   sh.flag.aud_mask        = 0xFFFFFFFFUL;
   sh.flag.client_severity = S_FALSE;
   sh.flag.client_class    = S_FALSE;
   sh.flag.hidefile        = S_FALSE;
+  sh.flag.loop            = S_FALSE;
 
 #ifdef MKB_09
   ErrFlag[1] |= (1 << 8);
@@ -296,14 +335,14 @@ void sh_init (void)
 
   /* The stats.
    */
-  sh.stat.openfiles    = 0;
-  sh.stat.bytes_hashed = 0;
-  sh.stat.mail_success = 0;
-  sh.stat.mail_failed  = 0;
-  sh.stat.log_success  = 0;
-  sh.stat.log_failed   = 0;
-  sh.stat.time_start   = time(NULL);
-  sh.stat.time_check   = (time_t) 0;
+  sh.statistics.openfiles    = 0;
+  sh.statistics.bytes_hashed = 0;
+  sh.statistics.mail_success = 0;
+  sh.statistics.mail_failed  = 0;
+  sh.statistics.log_success  = 0;
+  sh.statistics.log_failed   = 0;
+  sh.statistics.time_start   = time(NULL);
+  sh.statistics.time_check   = (time_t) 0;
 
 #ifdef MKC_01
   ErrFlag[0] |= (1 << 16);
@@ -333,7 +372,7 @@ void sh_init (void)
 
   /* The local host.
    */
-  sl_strlcpy (sh.host.name,  _("localhost"),  SH_MINIBUF);
+  (void) sl_strlcpy (sh.host.name,  _("localhost"),  SH_MINIBUF);
   sh.host.system[0]     = '\0';
   sh.host.release[0]    = '\0';
   sh.host.machine[0]    = '\0';
@@ -365,9 +404,9 @@ void sh_init (void)
 
   /* The paths.
    */
-  sl_strlcpy (sh.conf.path,  DEFAULT_CONFIGFILE,    SH_PATHBUF);
+  (void) sl_strlcpy (sh.conf.path,  DEFAULT_CONFIGFILE,    SH_PATHBUF);
   sh.conf.hash[0] = '\0';
-  sl_strlcpy (sh.data.path,  DEFAULT_DATA_FILE,     SH_PATHBUF);
+  (void) sl_strlcpy (sh.data.path,  DEFAULT_DATA_FILE,     SH_PATHBUF);
   sh.data.hash[0] = '\0';
   sh.exec.path[0] = '\0';
   sh.exec.hash[0] = '\0';
@@ -402,13 +441,13 @@ void sh_init (void)
 #if defined(SH_WITH_MAIL)
   if (0 == strcmp (DEFAULT_MAILADDRESS, _("NULL")))
     {
-      sl_strncpy(q, DEFAULT_MAILADDRESS, SH_PATHBUF);
+      (void) sl_strncpy(q, DEFAULT_MAILADDRESS, SH_PATHBUF);
       p = strtok (q, ", \t");
       if (p)
 	{
-	  sh_mail_setaddress_int (p);
+	  (void) sh_mail_setaddress_int (p);
 	  while (NULL != (p = strtok (NULL, ", \t")))
-	    sh_mail_setaddress_int (p);
+	    (void) sh_mail_setaddress_int (p);
 	}
     }
 #endif
@@ -442,7 +481,6 @@ void sh_init (void)
   else
     sl_strlcpy (sh.srvlog.name,  DEFAULT_ERRFILE,      SH_PATHBUF);
 
-
   if (0 == strcmp (ALT_CONSOLE, _("NULL")))
     sh.srvcons.alt[0] = '\0';
   else
@@ -495,19 +533,21 @@ void sh_init (void)
 
   sh.looptime     = 60;
 
-
   /* The struct to hold privileged information.
    */
-  skey = (sh_key_t *) my_malloc (sizeof(sh_key_t));
+  skey = (sh_key_t *) malloc (sizeof(sh_key_t));
   if (skey == NULL) 
     {
       perror(_("sh_init"));
       _exit (EXIT_FAILURE);
     }
 
-  skey->mlock_failed = GOOD;
+  skey->mlock_failed = SL_FALSE;
   skey->rngI         = BAD;
 
+  for (i = 0; i < KEY_BYT; ++i)
+    skey->poolv[i] = '\0';
+
   skey->poolc        = 0;
 
   skey->ErrFlag[0]   = ErrFlag[0];
@@ -518,13 +558,13 @@ void sh_init (void)
   dez = &(TcpFlag[POS_TF-1][0]);
   for (i = 0; i < PW_LEN; ++i)
     { 
-      skey->pw[i] = (*dez); 
+      /*@i@*/ skey->pw[i] = (*dez); 
       (*dez)      = '\0';
       ++dez; 
     }
 
-  sh_unix_mlock();
-  return;
+  sh_unix_memlock();
+  SL_RET0(_("sh_init"));
 }
 
 
@@ -532,6 +572,10 @@ void sh_init (void)
 #include <sys/mman.h>
 #endif
 
+#if defined(SH_USE_XML)
+extern int    sh_log_file    (char * message, char * inet_peer);
+#endif
+
 /*******************************************************
  * 
  * Exit Handler
@@ -543,7 +587,18 @@ void exit_handler(void)
    */
 #if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
   int modnum;
+#endif
+#if defined(SH_WITH_SERVER)
+  extern int sh_socket_remove ();
+#endif
 
+  SL_ENTER(_("exit_handler"));
+
+#if defined(SH_WITH_SERVER)
+  sh_socket_remove ();
+#endif
+
+#if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
   for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
     {
       if (modList[modnum].initval == GOOD)
@@ -581,6 +636,10 @@ void exit_handler(void)
   sh_mem_stat();
 #endif
 
+#ifdef MEM_DEBUG
+  sh_unix_count_mlock();
+#endif
+
   /* --- Checksum of executable. ---
    */
   sh_unix_self_check();
@@ -590,6 +649,9 @@ void exit_handler(void)
    */
   sh_error_handle ((-1), FIL__, __LINE__, sh.flag.exit, MSG_EXIT_NORMAL, 
 		   sh.prg_name, sh_sig_msg);
+#ifdef SH_USE_XML
+  sh_log_file (NULL, NULL);
+#endif
 
 
   /* --- Restrict error logging to stderr. ---
@@ -602,15 +664,388 @@ void exit_handler(void)
 
   /* --- Remove lock, delete critical information. ---
    */
-  sh_unix_unlock ();
+  sh_unix_rm_lock_file (sh.srvlog.name);
+  sh_unix_rm_pid_file ();
   if (skey != NULL)
     memset (skey, '\0', sizeof(sh_key_t));
   
   /* --- Exit. ---
    */
-  return;
+  SL_RET0(_("exit_handler"));
 }
 
+/***********************************************************
+ *
+ */
+#ifndef SIGHUP
+#define SIGHUP   1
+#endif
+#ifndef SIGTERM
+#define SIGTERM 15
+#endif
+#ifndef SIGKILL
+#define SIGKILL  9
+#endif
+
+#if defined(__linux__) || defined(sun) || defined(__sun) || defined(__sun__)
+#include <dirent.h>
+static pid_t * procdirSamhain ()
+{
+  pid_t        * pidlist = malloc(sizeof(pid_t) * 65535);
+  struct dirent * d;
+  DIR *        dp;
+  long         ino;
+  struct stat  buf;
+  int          i = 0;
+  pid_t        pid, mypid = getpid();
+  char       * tail;
+  char         exef[128];
+
+  for (i = 0; i < 65535; ++i) pidlist[i] = 0;
+  i = 0;
+
+  if (0 != stat(SH_INSTALL_PATH, &buf))
+    return NULL;
+
+  ino = (long) buf.st_ino;
+    
+  if (NULL == (dp = opendir("/proc")))
+    return NULL;
+  while (NULL != (d = readdir(dp)) && i < 65535)
+    {
+      if (0 != strcmp(d->d_name, ".") && 0 != strcmp(d->d_name, ".."))
+	{
+	  errno = 0;
+	  pid = (pid_t) strtol (d->d_name, &tail, 0);
+	  if (*tail != '\0' || errno != 0)
+	    continue;
+	  if (pid == mypid)
+	    continue;
+#if defined(__linux__) 
+          sprintf(exef, _("/proc/%d/exe"), (int) pid); /* known to fit  */
+#else
+          sprintf(exef, _("/proc/%d/object/a.out"),    /* known to fit  */
+		  (int) pid);
+#endif
+	  if (0 == stat(exef, &buf) && ino == (long) buf.st_ino)
+	    { pidlist[i] = (pid_t) pid; ++i; }
+	}
+    }
+  closedir(dp);
+  return pidlist;
+}
+#else
+static pid_t * procdirSamhain ()
+{
+  return NULL;
+}
+#endif
+
+static int killprocSamhain (pid_t pid)
+{
+  /* fprintf(stderr, "Killing %d\n", pid); */
+  if (pid > 0 && 0 == kill (pid, SIGTERM))
+    {
+      sleep(1);
+      if (0 != kill (pid, 0) && errno == ESRCH)
+	return (0);
+      sleep(2);
+      if (0 != kill (pid, 0) && errno == ESRCH)
+	return (0);
+      sleep(2);
+      if (0 != kill (pid, 0) && errno == ESRCH)
+	return (0);
+      
+      kill (pid, SIGKILL);
+      return (0);
+    }
+  if (pid > 0)
+    {
+      if (errno == ESRCH)
+	return 7;
+      if (errno == EPERM)
+	return 4;
+      return 1;
+    }
+  else
+    return (7);
+}
+
+static pid_t pidofSamhain (int flag)
+{
+  FILE      * fp;
+  char        line[256];
+  char      * tail;
+  char      * p;
+  pid_t       pid;
+  long        inpid;
+  struct stat buf;
+ 
+  fp = fopen (DEFAULT_ERRLOCK, "r");
+
+  if (!fp)
+    { if (errno != ENOENT) perror(_("fopen")); return 0; }
+  if (NULL == fgets(line, 255, fp))
+    { perror(_("fgets")); fclose(fp); return 0; }
+  fclose(fp); 
+  p = line; 
+  while (*p == ' '  || *p == '\f' || *p == '\n' || 
+	 *p == '\r' || *p == '\t' || *p == '\v')
+    ++p;
+  errno = 0;
+  inpid = strtol (p, &tail, 0);
+  if (p == tail || errno != 0)
+    { perror(_("strtol")); return 0; }
+
+  pid = (pid_t) inpid;
+  if (inpid != (long) pid)
+    { perror(_("strtol")); return 0; }
+
+  /* remove stale pid file
+   */
+  if (flag == 1 && pid > 0 && 0 != kill(pid, 0) && errno == ESRCH)
+    {
+      /*@i@*/ lstat (DEFAULT_ERRLOCK, &buf);
+      if (S_ISREG(buf.st_mode)) unlink(DEFAULT_ERRLOCK);
+
+      pid = 0;
+    }
+  return pid;
+}
+
+/* 1: start 2:stop 3:reload 4:status
+ */
+static int samhainctl(int ctl, int * argc, char * argv[])
+{
+  char * fullpath;
+  pid_t  pid;
+  int    status;
+  int    res;
+  pid_t  respid;
+  int    times;
+  char * argp[32];
+  char * envp[2];
+  pid_t       * pidlist;
+  int         i;
+
+
+  fullpath = malloc(strlen(SH_INSTALL_PATH)+1);
+  if (fullpath == NULL)
+    { perror(_("malloc")); exit (1); }
+  else
+    strcpy(fullpath, SH_INSTALL_PATH);                 /* known to fit  */
+
+  argp[0] = malloc(strlen(SH_INSTALL_PATH)+1);
+  if (argp[0] == NULL)
+    { perror(_("malloc")); exit (1); }
+  else
+    strcpy(argp[0], SH_INSTALL_PATH);                  /* known to fit  */
+
+
+  for (times = 1; times < 32; ++times)  argp[times] = NULL;
+
+  res = (*argc > 32) ? 32 : *argc;
+
+  for (times = 2; times < res; ++times)  
+    {
+      argp[times-1] = malloc(strlen(argv[times])+1);
+      if (argp[times-1] == NULL)
+	{ perror(_("malloc")); exit (1); }
+      else
+	strcpy(argp[times-1], argv[times]);              /* known to fit  */
+    }
+
+  if (getenv("TZ") != NULL)
+    {
+      envp[0] = 
+	malloc(strlen(getenv("TZ"))+1);
+      if (envp[0] == NULL)
+	{ perror(_("malloc")); exit (1); }
+      else
+	strcpy(envp[0], getenv("TZ"));                   /* known to fit  */
+    }
+  else
+    {
+      envp[0] = NULL;
+    }
+  envp[1] = NULL;
+
+  if (ctl == 1)
+    {
+      pid = pidofSamhain(1);
+
+      if (pid != 0 && 0 == kill (pid, 0)) /* already started */
+	exit (0);
+
+      pid = fork();
+      switch (pid) {
+      case ((pid_t) -1):
+	perror(_("fork"));
+	exit (1);
+      case  0:
+	close (0);
+	execve(fullpath, argp, envp);
+	if (errno == EPERM)
+	  _exit(4);
+	else if (errno == ENOENT)
+	  _exit(5);
+	_exit (1);
+      default:
+	times = 0;
+	while (times < 10) {
+	  respid = waitpid(pid, &status, WNOHANG|WUNTRACED);
+	  if ((pid_t)-1 == respid)
+	    {
+	      perror(_("waitpid"));
+	      exit (1);
+	    }
+	  else if (pid == respid)
+	    {
+	      if (0 != WIFEXITED(status))
+		{
+		  res = WEXITSTATUS(status);
+		  exit (res == 0 ? 0 : res );
+		}
+	      else
+		exit (1);
+	    }
+	  ++times;
+	  sleep(1);
+	}
+	exit (1);
+      }
+    }
+
+  pid = pidofSamhain(0);
+
+  if (ctl == 2)  /* stop */
+    {
+      pidlist = procdirSamhain ();
+      if (pid == 0 && NULL == pidlist) /* pid file not found */ 
+	return (0);
+	  
+      status = 0;
+      if (pid != 0)
+	 status = killprocSamhain(pid);
+      if (pidlist != NULL)
+	{
+	  i = 0; 
+	  while (i < 65535 && pidlist[i] != 0)
+	    { 
+	      if (pidlist[i] != pid) 
+		status = killprocSamhain(pidlist[i]);
+	      ++i;
+	    }
+	}
+      if (status == 7)
+	return 0;
+      else
+	return status;
+    }
+	
+  if (ctl == 3)  /* reload */
+    {
+      if (pid == 0)
+        exit (7);
+      if (0 == kill (pid, SIGHUP))
+        exit (0);
+      else
+	{
+	  if (errno == EPERM)
+	    exit (4);
+	  if (errno == ESRCH)
+	    exit (7);
+	  exit (1);
+	}
+    }
+
+  if (ctl == 4)  /* status */
+    {
+      if (pid == 0)
+	exit (3);
+      if (0 == kill (pid, 0))
+	exit (0);
+      else
+	{
+	  if (errno == EPERM)
+	    exit (4);
+	  if (errno == ESRCH)
+	    exit (1);
+	}
+    }
+
+  exit (1); /* no exit handler installed yet */
+  return (0); /* notreached */
+}
+
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
+#include "sh_schedule.h"
+static sh_schedule_t * FileSchedOne = NULL;
+static sh_schedule_t * FileSchedTwo = NULL;
+
+/* free a linked list of schedules
+ */
+sh_schedule_t *  free_sched (sh_schedule_t * isched)
+{
+  sh_schedule_t * current = isched;
+  sh_schedule_t * next    = NULL;
+
+  while (current != NULL)
+    {
+      next = current->next;
+      SH_FREE(current);
+      current = next;
+    }
+  return NULL;
+}
+
+/* Add a new schedule to the linked list of schedules
+ */
+sh_schedule_t * sh_set_schedule_int (char * str, sh_schedule_t * FileSchedIn, 
+				     int * status)
+{
+  sh_schedule_t * FileSched;
+
+  SL_ENTER(_("sh_set_schedule_int"));
+
+  if (0 == sl_strncmp(str, _("NULL"), 4))
+    {
+      free_sched(FileSchedIn);
+      FileSchedIn = NULL;
+      return 0;
+    }
+
+  FileSched = SH_ALLOC(sizeof(sh_schedule_t));
+  *status = create_sched(str, FileSched);
+  if (*status != 0)
+    {
+      SH_FREE(FileSched);
+      FileSched = NULL;
+      SL_RETURN(FileSchedIn , _("sh_set_schedule_int"));
+    }
+  FileSched->next = FileSchedIn;
+  SL_RETURN(FileSched , _("sh_set_schedule_int"));
+}
+
+/* Add a new schedule to the linked list FileSchedOne
+ */
+int sh_set_schedule_one (char * str)
+{
+  int status;
+  FileSchedOne = sh_set_schedule_int (str, FileSchedOne, &status);
+  return status;
+}
+
+/* Add a new schedule to the linked list FileSchedTwo
+ */
+int sh_set_schedule_two (char * str)
+{
+  int status;
+  FileSchedTwo = sh_set_schedule_int (str, FileSchedTwo, &status);
+  return status;
+}
+
+#endif
 
 /*******************************************************
  * 
@@ -619,6 +1054,12 @@ void exit_handler(void)
  *******************************************************/
 int main(int argc, char * argv[])
 {
+#if defined(INET_SYSLOG)
+  extern int    create_syslog_socket (int flag);
+#endif
+#if defined(SH_WITH_SERVER)
+  extern int    sh_create_tcp_socket();
+#endif
 
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
   int           modnum;
@@ -626,11 +1067,15 @@ int main(int argc, char * argv[])
   float         st_1, st_2;
   int           status;
   long          cct = 0; /* main loop iterations */
+
+  int           flag_check_1 = 0;
+  int           flag_check_2 = 0;
 #endif
 
   unsigned long told;
   unsigned long tcurrent;
   size_t        tzlen;
+  int           res;
 
 #if defined (SH_STEALTH_NOCL)
   char  * command_line;
@@ -638,6 +1083,54 @@ int main(int argc, char * argv[])
   char  * my_argv[32];
 #endif
 
+  SL_ENTER(_("main"));
+
+  if (argc >= 2 && 0 == getuid())
+    {
+      /* return codes:
+       * 0    Success
+       * 1    Can not send signal / start program
+       * 2    Pid file does not exist
+       */
+      if      (0 == strcmp(argv[1], _("start")))
+	{
+	  samhainctl (1, &argc, argv);
+	}
+      else if (0 == strcmp(argv[1], _("stop")))
+        return (samhainctl (2, &argc, argv));
+      else if (0 == strcmp(argv[1], _("reload")))
+	samhainctl (3, &argc, argv);
+      else if (0 == strcmp(argv[1], _("force-reload")))
+	samhainctl (3, &argc, argv);
+      else if (0 == strcmp(argv[1], _("status")))
+	samhainctl (4, &argc, argv);
+      else if (0 == strcmp(argv[1], _("restart")))
+	{
+	  res = samhainctl (2, &argc, argv);
+	  if (res == 0 || res == 7)
+	    {
+	      samhainctl (1, &argc, argv);
+	    }
+	  else
+	    return (res);
+	}
+    }
+  
+  /* if fd 0 is closed, presume that we want to be daemon and
+   * run in check mode
+   */
+  if ((-1) == retry_fcntl(FIL__, __LINE__, 0, F_GETFL, 0) && 
+	   errno == EBADF)
+    {
+      sh.flag.opts = S_TRUE;
+      sh_unix_setdeamon(NULL);
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
+      sh_util_setchecksum(_("check"));
+#endif
+      is_samhainctl_init = S_TRUE;
+      sh.flag.opts = S_FALSE;
+    }
+
 
   /* --- Install the exit handler. ---
    */
@@ -672,55 +1165,54 @@ int main(int argc, char * argv[])
   ErrFlag[0] |= (1 << 7);
 #endif
 
-  (void) sh_derr();
+#if defined(SCREW_IT_UP)
+  BREAKEXIT(sh_sigtrap_prepare);
+  (void) sh_sigtrap_prepare();
+#endif
 
   /* Save the timezone.
    */
   if (getenv("TZ") != NULL)
     {
       tzlen       = strlen(getenv("TZ"));
-      sh.timezone = my_malloc (tzlen + 1);
+      sh.timezone = malloc (tzlen + 1);
       if (sh.timezone != NULL)
 	sl_strlcpy (sh.timezone, getenv("TZ"), tzlen + 1);
     }
   else
     sh.timezone = NULL;
-  
 
 
   /* --------  INIT  --------    
    */
-  SL_REQUIRE(sl_policy_get_user(DEFAULT_IDENT) == SL_ENONE, 
-	     _("sl_policy_get_user(DEFAULT_IDENT) == SL_ENONE")); 
 
   /* Restrict error logging to stderr.
    */
   sh_error_only_stderr (S_TRUE);
 
+  BREAKEXIT(sh_derr);
+  (void) sh_derr();
+
   /* Check that first three descriptors are open.
    */
-  if ( retry_fcntl(0, F_GETFL, 0) == (-1) || 
-       retry_fcntl(1, F_GETFL, 0) == (-1) || 
-       retry_fcntl(2, F_GETFL, 0) == (-1))
-    {
-      sh_unix_closeall (0);
-      aud_open(FIL__, __LINE__, SL_YESPRIV, _("/dev/null"), O_RDWR, 0);
-      aud_dup (FIL__, __LINE__, 0); 
-      aud_dup (FIL__, __LINE__, 0);
-    }
-
+  if ( retry_fcntl(FIL__, __LINE__, 0, F_GETFL, 0) == (-1))
+    (void) aud_open(FIL__, __LINE__, SL_NOPRIV, _("/dev/null"), O_RDWR, 0);
+  if ( retry_fcntl(FIL__, __LINE__, 1, F_GETFL, 0) == (-1))
+    (void) aud_open(FIL__, __LINE__, SL_NOPRIV, _("/dev/null"), O_RDWR, 1);
+  if ( retry_fcntl(FIL__, __LINE__, 2, F_GETFL, 0) == (-1))
+    (void) aud_open(FIL__, __LINE__, SL_NOPRIV, _("/dev/null"), O_RDWR, 2);
 
   /* --- Set default values. ---
    */
-  sh_init ();
-
-
+  BREAKEXIT(sh_init);
+  sh_init ();    /* we are still privileged here, so we can mlock skey */
 #if (defined (SH_WITH_SERVER) && !defined (SH_WITH_CLIENT))
   sh.flag.isserver = S_TRUE;
 #endif
 
   /* --- Get local hostname. ---
    */
+  BREAKEXIT(sh_unix_localhost);
   sh_unix_localhost();
 
   /* --- Read the command line. ---
@@ -738,7 +1230,8 @@ int main(int argc, char * argv[])
 	{
 	  my_argv[0] = argv[0]; ++my_argc;  
 	  command_line[0] = '\0';
-	  fgets (command_line, 255, stdin);
+	  (void*) fgets (command_line, 255, stdin);
+	  command_line[255] = '\0';
 	  do {
 	    my_argv[my_argc] = 
 	      strtok( (my_argc == 1) ? command_line : NULL, " \n"); 
@@ -761,9 +1254,14 @@ int main(int argc, char * argv[])
   sh.flag.opts = S_FALSE;
   
 
+  /* close all other files
+   */
+  sh_unix_closeall(3);
+
+
   /* --- Get user info. ---
    */
-  TPT((0, FIL__, __LINE__, _("msg=<Get user hostname.>\n")))
+  TPT((0, FIL__, __LINE__, _("msg=<Get user name.>\n")))
   if (0 != sh_unix_getUser ())
     {
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
@@ -779,35 +1277,45 @@ int main(int argc, char * argv[])
    * *****************************/
 
   TPT((0, FIL__, __LINE__, _("msg=<Read the configuration file.>\n")))
+  BREAKEXIT(sh_readconf_read);
   (void) sh_readconf_read ();
 
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
+  if (sh.flag.checkSum == SH_CHECK_NONE)
+    {
+      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN,
+		       _("No action specified"), _("main"));
+      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
+		       sh.prg_name);
+      aud_exit (FIL__, __LINE__, EXIT_FAILURE);
+    }
+#endif
+
   /* do not append to database if run SUID
    */
   if (sh.flag.checkSum == SH_CHECK_INIT && sl_is_suid()) 
     {
+      dlog(1, FIL__, __LINE__, 
+	   _("Cannot initialize database when running with SUID credentials.\nYou need to run this with the user ID %d.\nYour current user ID is %d."), 
+	   (int) geteuid(), (int) sh.real.uid);
       sh_error_handle ((-1), FIL__, __LINE__, EACCES, MSG_ACCESS,
 		       (long) sh.real.uid, sh.data.path);
-      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
-		       sh.prg_name);
       aud_exit(FIL__, __LINE__, EXIT_FAILURE);
     }
 
   /* avoid daemon mode for initialization 
    */
   if (sh.flag.checkSum == SH_CHECK_INIT)
-    sh.flag.isdaemon = S_FALSE;
+    {
+      sh.flag.isdaemon = S_FALSE;
+      sh.flag.loop     = S_FALSE;
+    }
 
   /* initialize signal handling etc
    */
   if (sh.flag.isdaemon == S_TRUE)
-    { 
-      sh_error_only_stderr (BAD);
-#if defined(WITH_TRACE) || defined(WITH_TPT) 
-      dbg_use_console();
-#endif
-    }
+    sh_error_only_stderr (BAD);
 
-  TPT((0, FIL__, __LINE__, _("msg=<Initialize signal handling etc.>\n")))
   if (sh_unix_init(sh.flag.isdaemon) == -1) 
     {
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
@@ -815,6 +1323,36 @@ int main(int argc, char * argv[])
       aud_exit(FIL__, __LINE__, EXIT_FAILURE);
     }
 
+  /* --- drop privileges eventually ---
+   */
+#if defined(SH_WITH_SERVER)
+  sh_create_tcp_socket ();
+#if defined(INET_SYSLOG)
+  create_syslog_socket (S_TRUE);
+#endif
+  SL_REQUIRE(sl_policy_get_real(DEFAULT_IDENT) == SL_ENONE, 
+	     _("sl_policy_get_real(DEFAULT_IDENT) == SL_ENONE"));
+#else
+  SL_REQUIRE(sl_policy_get_user(DEFAULT_IDENT) == SL_ENONE, 
+	     _("sl_policy_get_user(DEFAULT_IDENT) == SL_ENONE"));
+#endif
+
+  /* --- Get user info (again). ---
+   */
+  TPT((0, FIL__, __LINE__, _("msg=<Get user name.>\n")))
+  if (0 != sh_unix_getUser ())
+    {
+      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
+		       sh.prg_name);
+      aud_exit(FIL__, __LINE__, EXIT_FAILURE);
+    }
+
+  /* --- now check whether we really wanted it; if not, close ---
+   */
+#if defined(INET_SYSLOG) && defined(SH_WITH_SERVER)
+  create_syslog_socket (S_FALSE);
+#endif
+
 
   /* checksum of database
    */
@@ -822,7 +1360,7 @@ int main(int argc, char * argv[])
   TPT((0, FIL__, __LINE__, _("msg=<Get checksum of the database.>\n")))
   if (sh.flag.checkSum == SH_CHECK_CHECK) 
     {
-      if (0 == strcmp(file_path('D', 'R'), _("REQ_FROM_SERVER")))
+      if (0 == sl_strcmp(file_path('D', 'R'), _("REQ_FROM_SERVER")))
 	{
 	  /* fetch the file from server to get checksum
 	   */
@@ -860,7 +1398,7 @@ int main(int argc, char * argv[])
   sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_1H,
 		   sh.prg_name, (long) sh.real.uid, 
 		   (sh.flag.hidefile == S_TRUE) ? 
-		   _("(hidden)") : file_path('C', 'R'), 
+		   _("(hidden)") : file_path('C','R'), 
 		   sh.conf.hash);
 #endif
 
@@ -872,6 +1410,8 @@ int main(int argc, char * argv[])
    *
    ****************************************************/
 
+  BREAKEXIT(sh_error_handle);
+
   if (sh.flag.checkSum == SH_CHECK_CHECK) 
     {
 #if (defined(WITH_GPG) || defined(WITH_PGP))
@@ -904,7 +1444,7 @@ int main(int argc, char * argv[])
 #endif
 
  
-  if (skey->mlock_failed == BAD)
+  if (skey->mlock_failed == SL_TRUE)
     sh_error_handle ((-1), FIL__, __LINE__, EPERM, MSG_MLOCK);
 
   /* timer
@@ -943,7 +1483,8 @@ int main(int argc, char * argv[])
    *   CLIENT/STANDALONE
    *
    ****************************************************/
-#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
+
 
   /* --- Initialize modules. ---
    */
@@ -952,9 +1493,16 @@ int main(int argc, char * argv[])
     {
       if (0 != (status = modList[modnum].mod_init()))
 	{
-	  sh_error_handle ((-1), FIL__, __LINE__, status, MSG_MOD_FAIL,
-			   _(modList[modnum].name),
-			   status);
+	  if (status == (-1)) {
+	    sh_error_handle (SH_ERR_NOTICE, FIL__, __LINE__, status, 
+			     MSG_MOD_FAIL,
+			     _(modList[modnum].name),
+			     status);
+	  } else {
+	    sh_error_handle ((-1), FIL__, __LINE__, status, MSG_MOD_FAIL,
+			     _(modList[modnum].name),
+			     status);
+	  }
 	  modList[modnum].initval = S_FALSE;
 	}
       else
@@ -970,12 +1518,28 @@ int main(int argc, char * argv[])
   sh_files_setrec();
   sh_files_test_setup();
 
+
+  /* --------  NICE LEVEL   ---------
+   */
+  if (0 != sh.flag.nice)
+    {
+      sl_set_suid();
+#ifdef HAVE_SETPRIORITY
+      /*@i@*/ setpriority(PRIO_PROCESS, 0, sh.flag.nice);
+#else
+      nice(sh.flag.nice);
+#endif
+      sl_unset_suid();
+    }
+
   /*  --------  MAIN LOOP  ---------
    */
   while (1) 
     {
       ++cct;
 
+      BREAKEXIT(sh_error_handle);
+
       TPT((0, FIL__, __LINE__, _("msg=<Start main loop.>, iter=<%ld>\n"), cct))
 
       tcurrent                   = (unsigned long) time (NULL);
@@ -985,6 +1549,13 @@ int main(int argc, char * argv[])
 
 	  TPT((0, FIL__, __LINE__, _("msg=<Process a signal.>\n")))
 
+	  if (sig_termfast == 1)  /* SIGTERM */
+	    {
+	      TPT((0, FIL__, __LINE__, _("msg=<Terminate.>\n")));
+	      strncpy (sh_sig_msg, _("SIGTERM"), 20);
+	      aud_exit (FIL__, __LINE__, EXIT_SUCCESS);
+	    }
+	  
 	  if (sig_config_read_again == 1) /* SIGHUP */
 	    {
 	      TPT((0, FIL__, __LINE__, _("msg=<Re-read configuration.>\n")))
@@ -1000,37 +1571,102 @@ int main(int argc, char * argv[])
 	       */
 	      sh_files_deldirstack ();
 	      sh_files_delfilestack ();
+	      sh_ignore_clean ();
 	      hash_full_tree ();
 
 #if defined(SH_WITH_CLIENT)
 	      reset_count_dev_server();
 #endif
+#if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
+
+
+	      FileSchedOne = free_sched(FileSchedOne);
+	      FileSchedTwo = free_sched(FileSchedTwo);
+
+	      for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
+		{
+		  if (modList[modnum].initval == GOOD)
+		    modList[modnum].mod_reconf();
+		}
+#endif
+
 #if defined(SH_WITH_MAIL)
 	      reset_count_dev_mail();
 #endif
 	      reset_count_dev_console();
 	      reset_count_dev_time();
 
+	      sh_unix_maskreset();
+ 
 	      /* Should this be included ??? 
-		 (i.e. should we reload the database ?)
-	      */
+	       * (i.e. should we reload the database ?)
+	       */
 #ifdef RELOAD_DATABASE
 	      sh_hash_hashdelete();
 #endif
+	      sl_trust_purge_user();
 
 	      (void) sh_readconf_read ();
 	      sig_config_read_again = 0;
 	      sh_files_setrec();
 	      sh_files_test_setup();
+	      if (0 != sh.flag.nice)
+		{
+		  sl_set_suid();
+#ifdef HAVE_SETPRIORITY
+		  setpriority(PRIO_PROCESS, 0, sh.flag.nice);
+#else
+		  nice(sh.flag.nice);
+#endif
+		  sl_unset_suid();
+		}
+
+	      if (sh.flag.checkSum == SH_CHECK_INIT)
+		{
+		  sh.flag.isdaemon = S_FALSE;
+		  sh.flag.loop     = S_FALSE;
+		}
+
+	      /* --- Initialize modules. ---
+	       */
+	      TPT((0, FIL__, __LINE__, _("msg=<Initialize modules.>\n")));
+	      for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
+		{
+		  if (0 != (status = modList[modnum].mod_init()))
+		    {
+		      if (status == (-1)) {
+			sh_error_handle (SH_ERR_NOTICE, FIL__, __LINE__, 
+					 status, MSG_MOD_FAIL,
+					 _(modList[modnum].name),
+					 status);
+		      } else {
+			sh_error_handle ((-1), FIL__, __LINE__, 
+					 status, MSG_MOD_FAIL,
+					 _(modList[modnum].name),
+					 status);
+		      }
+		      modList[modnum].initval = S_FALSE;
+		    }
+		  else
+		    {
+		      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_MOD_OK,
+				       _(modList[modnum].name));
+		      modList[modnum].initval = S_TRUE;
+		    }
+		}
+	      
 	    }
 	  
 	  if (sig_fresh_trail == 1) /* SIGIOT */
 	    {
 	      /* Logfile access 
 	       */
-	      TPT((0, FIL__, __LINE__, _("msg=<Logfile stop/restart.>\n")))
+#ifdef SH_USE_XML
+	      sh_log_file (NULL, NULL);
+#endif
+	      TPT((0, FIL__, __LINE__, _("msg=<Logfile stop/restart.>\n")));
 	      sh_error_only_stderr (S_TRUE);
-	      sh_unix_unlock();
+	      sh_unix_rm_lock_file(sh.srvlog.name);
 	      sleep(3);
 	      sh.flag.log_start = S_TRUE;
 	      sh_error_only_stderr (S_FALSE);
@@ -1039,78 +1675,131 @@ int main(int argc, char * argv[])
 	  
 	  if (sig_terminate == 1)  /* SIGQUIT */
 	    {
-	      TPT((0, FIL__, __LINE__, _("msg=<Terminate.>\n")))
+	      TPT((0, FIL__, __LINE__, _("msg=<Terminate.>\n")));
 	      strncpy (sh_sig_msg, _("SIGQUIT"), 20);
 	      aud_exit (FIL__, __LINE__, EXIT_SUCCESS);
 	    }
 	  
-	  if (sig_debug_on == 1)  /* SIGUSR1 */
+	  if (sig_debug_switch == 1)  /* SIGUSR1 */
 	    {
-	      TPT((0, FIL__, __LINE__, _("msg=<Debug on.>\n")))
-	      sh_error_dbg_on();
-	      sig_debug_on = 0;
+	      TPT((0, FIL__, __LINE__, _("msg=<Debug switch.>\n")));
+	      sh_error_dbg_switch();
+	      sig_debug_switch = 0;
 	    }
 	  
-	  if (sig_debug_off == 1)  /* SIGUSR2 */
+	  if (sig_suspend_switch == 1)  /* SIGUSR2 */
 	    {
-	      TPT((0, FIL__, __LINE__, _("msg=<Debug off.>\n")))
-	      sh_error_dbg_off();
-	      sig_debug_off = 0;
+	      TPT((0, FIL__, __LINE__, _("msg=<Suspend switch.>\n")));
+	      if (sh_global_suspend_flag == 1) {
+		sh_global_suspend_flag = 0;
+	      } else {
+		sh_error_handle((-1), FIL__, __LINE__, 0, MSG_SUSPEND, 
+				sh.prg_name);
+		sh_global_suspend_flag = 1;
+	      }
+	      sig_suspend_switch = 0;
 	    }
 	  sig_raised = 0;
-	  TPT((0, FIL__, __LINE__, _("msg=<End signal processing.>\n")))
+	  TPT((0, FIL__, __LINE__, _("msg=<End signal processing.>\n")));
 	}
       
+      if (sh_global_suspend_flag == 1)
+	{
+	  (void) sleep (1);
+	  continue;
+	}
       
       /* see whether its time to check files
        */
+      if      (sh.flag.checkSum == SH_CHECK_INIT)
+	{
+	  flag_check_1 = 1;
+	  if (FileSchedTwo != NULL) 
+	    flag_check_2 = 1;
+	}
+      else if (sh.flag.checkSum == SH_CHECK_CHECK || 
+	       (sh.flag.update == S_TRUE && 
+		(sh.flag.isdaemon == S_TRUE || sh.flag.loop == S_TRUE)))
+	{
+	  if (FileSchedOne == NULL)
+	    {
+	      /* use interval if we have no schedule
+	       */
+	      if (tcurrent - sh.fileCheck.alarm_last >= 
+		  sh.fileCheck.alarm_interval)
+		flag_check_1 = 1;
+	    }
+	  else
+	    {
+	      flag_check_1 = test_sched(FileSchedOne);
+	      if (FileSchedTwo != NULL) 
+		flag_check_2 = test_sched(FileSchedTwo);
+	      if (flag_check_2 == 1) 
+		flag_check_1 = 1;
+	    }
+	}
+
       if (sh.flag.checkSum != SH_CHECK_NONE &&
-	  tcurrent - sh.fileCheck.alarm_last >= sh.fileCheck.alarm_interval) 
+	  (flag_check_1 == 1 || flag_check_2 == 1))
 	{
 	  /* 
 	   * check directories and files
 	   * ORDER IS IMPORTANT -- DIRZ FIRST
 	   */
-	  sh.stat.bytes_hashed  = 0;
-	  sh.stat.time_start    = time (NULL);
-	  TPT((0, FIL__, __LINE__, _("msg=<Check directories.>\n")))
-	  sh.stat.dirs_checked  = sh_dirs_chk  (); 
-	  TPT((0, FIL__, __LINE__, _("msg=<Check files.>\n")))
-	  sh.stat.files_checked = sh_files_chk ();
+	  sh.statistics.bytes_hashed  = 0;
+	  sh.statistics.time_start    = time (NULL);
 
-	  if (sig_terminate == 1)
+	  TPT((0, FIL__, __LINE__, _("msg=<Check directories.>\n")))
+	  BREAKEXIT(sh_dirs_chk);
+	  if (flag_check_1 == 1)
+	    sh.statistics.dirs_checked  = sh_dirs_chk  (1); 
+	  if (flag_check_2 == 1)
+	    sh.statistics.dirs_checked  = sh_dirs_chk  (2); 
+	  TPT((0, FIL__, __LINE__, _("msg=<Check files.>\n")))
+	  BREAKEXIT(sh_files_chk);
+	  if (flag_check_1 == 1)
+	    sh.statistics.files_checked = sh_files_chk ();
+
+	  if (sig_urgent == 1)
 	    continue;
 
 	  /*
 	   * check for files not visited
 	   */
-	  TPT((0, FIL__, __LINE__, _("msg=<Check for missing files.>\n")))
-	  sh_hash_unvisited (ShDFLevel[SH_ERR_T_FILE]);
+	  if (flag_check_2 == 1 || FileSchedTwo == NULL)
+	    {
+	      TPT((0, FIL__, __LINE__, _("msg=<Check for missing files.>\n")))
+	      sh_hash_unvisited (ShDFLevel[SH_ERR_T_FILE]);
+	    }
 
-	  if (sig_terminate == 1)
+	  if (sig_urgent == 1)
 	    continue;
 
 	  /* reset
 	   */
 	  TPT((0, FIL__, __LINE__, _("msg=<Reset status.>\n")))
 	  sh_dirs_reset  ();
-	  if (sig_terminate == 1)
+	  if (sig_urgent == 1)
 	    continue;
 
 	  sh_files_reset ();
-	  if (sig_terminate == 1)
+	  flag_check_1 = 0;
+	  flag_check_2 = 0;
+
+	  if (sig_urgent == 1)
 	    continue;
 
-	  runtim = time(NULL) - sh.stat.time_start;
+	  runtim = time(NULL) - sh.statistics.time_start;
 
 	
-	  if (sh.stat.dirs_checked == 0 && sh.stat.files_checked == 0)
+	  if ((sh.statistics.dirs_checked == 0) && 
+	      (sh.statistics.files_checked == 0))
 	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_CHECK_0);
 
 	  else
 	    {
-	      st_1 = sh.stat.bytes_hashed;
-	      st_2 = runtim;
+	      st_1 = (float) sh.statistics.bytes_hashed;
+	      st_2 = (float) runtim;
 
 
 	      if (st_1 > 0.0 && st_2 > 0.0) 
@@ -1127,7 +1816,7 @@ int main(int argc, char * argv[])
 	    }
 	  sh.fileCheck.alarm_last = (unsigned long) time (NULL);
 
-	  if (sig_terminate == 1)
+	  if (sig_urgent == 1)
 	    continue;
 
 	  /*
@@ -1139,7 +1828,7 @@ int main(int argc, char * argv[])
 #endif
 	}
       
-      if (sig_terminate == 1)
+      if (sig_urgent == 1)
 	continue;
       
       /* execute modules
@@ -1154,11 +1843,17 @@ int main(int argc, char * argv[])
 			       _(modList[modnum].name), (long) status);
 	}
       
+      /* 27.05.2002 avoid empty database
+       * 22.10.2002 moved here b/o suid check initialization
+       */ 
+      if      (sh.flag.checkSum == SH_CHECK_INIT)
+	sh_hash_pushdata (NULL, NULL);
+
       /* no loop if not daemon
        */
-      if (sh.flag.isdaemon != S_TRUE)
+      if (sh.flag.isdaemon != S_TRUE && sh.flag.loop == S_FALSE)
 	break; 
-      if (sig_terminate == 1)
+      if (sig_urgent == 1)
 	continue;
       
       /* see whether its time to send mail
@@ -1171,17 +1866,18 @@ int main(int argc, char * argv[])
 	  sh.mailTime.alarm_last = (unsigned long) time (NULL);
 	}
 #endif
-      if (sig_terminate == 1)
+      if (sig_urgent == 1)
 	continue;
             
       /* log the timestamp
        */
-      if ((tcurrent - told) > (unsigned int) sh.looptime )
+      if ((tcurrent - told) >= (unsigned int) sh.looptime )
 	{
 	  TPT((0, FIL__, __LINE__, _("msg=<Log the timestamp.>\n")))
 	  told = tcurrent;
 #ifdef MEM_DEBUG
 	  sh_mem_check();
+	  sh_unix_count_mlock();
 #else
 	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_STAMP);
 #endif
@@ -1191,19 +1887,22 @@ int main(int argc, char * argv[])
        */
       (void) taus_seed();
       
-      if (sig_terminate == 1)
+      if (sig_urgent == 1)
 	continue;
       
       /* go to sleep
        */
       (void) sleep (1);
 
+      BREAKEXIT(sh_derr);
       (void) sh_derr();
     }
   
   /*   ------  END  -----------
    */
 
+
+
   /*
    * cleanup
    */
@@ -1219,5 +1918,5 @@ int main(int argc, char * argv[])
 #endif
 
   aud_exit (FIL__, __LINE__, EXIT_SUCCESS);
-  return (0);
+  SL_RETURN(0, _("main"));
 }
