@@ -67,6 +67,212 @@
 #undef  FIL__
 #define FIL__  _("sh_entropy.c")
 
+#if defined (HAVE_EGD_RANDOM)
+/* rndegd.c  -  interface to the EGD
+ *      Copyright (C) 1999, 2000, 2001 Free Software Foundation, Inc.
+ */
+#include <stddef.h>
+#include <sys/socket.h>
+#include <sys/un.h>
+
+static int
+do_write( int fd, void *buf, size_t nbytes )
+{
+    size_t nleft = nbytes;
+    int nwritten;
+
+    while( nleft > 0 ) {
+        nwritten = write( fd, buf, nleft);
+        if( nwritten < 0 ) {
+            if( errno == EINTR )
+                continue;
+            return -1;
+        }
+        nleft -= nwritten;
+        buf = (char*)buf + nwritten;
+    }
+    return 0;
+}
+
+static int
+do_read( int fd, void *buf, int nbytes )
+{
+    int n, nread = 0;
+
+    if (nbytes < 0)
+      return 0;
+
+    do {
+        do {
+            n = read(fd, (char*)buf + nread, nbytes );
+        } while( n == -1 && errno == EINTR );
+        if( n == -1 )
+            return -1;
+        nread += n;
+    } while( nread < nbytes );
+    return nbytes;
+}
+
+
+int sh_entropy(int getbytes, char * nbuf)
+{
+    static int fd = -1;
+    int n;
+    byte buffer[256+2];
+    int nbytes;
+    int do_restart = 0;
+    int myerror = 0;
+    int length;
+    char * p = nbuf;
+    int i;
+
+    SL_ENTER(_("sh_entropy"));
+
+    if( getbytes <= 0)
+        SL_RETURN( -1, _("sh_entropy"));
+    if (getbytes > KEY_BYT)
+      getbytes = KEY_BYT;
+    length = getbytes;
+
+  restart:
+    if( do_restart ) {
+        if( fd != -1 ) {
+            close( fd );
+            fd = -1;
+        }
+    }
+    if( fd == -1 ) {
+        const char *bname = NULL;
+        char *name;
+        struct sockaddr_un addr;
+        int addr_len;
+
+      #ifdef EGD_SOCKET_NAME
+        bname = EGD_SOCKET_NAME;
+      #endif
+        if ( !bname || !*bname )
+            bname = _("=entropy");
+
+        if ( *bname == '=' && bname[1] )
+            name = sh_util_strconcat ( DEFAULT_DATAROOT, "/", bname+1 , NULL );
+        else
+            name = sh_util_strconcat ( bname , NULL );
+
+        if ( strlen(name)+1 >= sizeof(addr.sun_path) )
+	  {
+	    sh_error_handle ((-1), FIL__, __LINE__, ENAMETOOLONG, MSG_E_SUBGEN,
+			     _("EGD socketname is too long"),
+			     _("sh_entropy") ); 
+	    SH_FREE(name);
+	    SL_RETURN( -1, _("sh_entropy") );
+	  }
+
+        memset( &addr, 0, sizeof(addr) );
+        addr.sun_family = AF_UNIX;
+        strcpy( addr.sun_path, name );              /* known to fit  */
+        addr_len = offsetof( struct sockaddr_un, sun_path )
+                   + strlen( addr.sun_path );
+
+        fd = socket(AF_UNIX, SOCK_STREAM, 0);
+        if( fd == -1 )
+	  {
+	    myerror = errno;
+	    sh_error_handle ((-1), FIL__, __LINE__, myerror, MSG_E_SUBGEN,
+			     _("cannot create unix domain socket"),
+			     _("sh_entropy") ); 
+	    SH_FREE(name);
+	    SL_RETURN( -1, _("sh_entropy") );
+	  }
+        if( connect( fd, (struct sockaddr*)&addr, addr_len) == -1 )
+	  {
+	    myerror = errno;
+	    sh_error_handle ((-1), FIL__, __LINE__, myerror, MSG_E_SUBGEN,
+			     _("cannot connect to unix domain socket"),
+			     _("sh_entropy") ); 
+	    SH_FREE(name);
+	    SL_RETURN( -1, _("sh_entropy") );
+	  }
+        SH_FREE(name);
+    }
+    do_restart = 0;
+
+    nbytes = length < 255? length : 255;
+    /* first time we do it with a non blocking request */
+    buffer[0] = 1; /* non blocking */
+    buffer[1] = nbytes;
+    if( do_write( fd, buffer, 2 ) == -1 )
+	  {
+	    myerror = errno;
+	    sh_error_handle ((-1), FIL__, __LINE__, myerror, MSG_E_SUBGEN,
+			     _("cannot write to EGD"),
+			     _("sh_entropy") ); 
+	    SL_RETURN( -1, _("sh_entropy") );
+	  }
+    n = do_read( fd, buffer, 1 );
+    if( n == -1 ) {
+        myerror = errno;
+        sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, myerror, MSG_E_SUBGEN,
+			 _("read error on EGD"),
+			 _("sh_entropy") ); 
+        do_restart = 1;
+        goto restart;
+    }
+    n = buffer[0];
+    if( n ) {
+        n = do_read( fd, buffer, n );
+        if( n == -1 ) {
+            myerror = errno;
+            sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, myerror,MSG_E_SUBGEN,
+			     _("read error on EGD"),
+			     _("sh_entropy") ); 
+            do_restart = 1;
+            goto restart;
+        }
+	for (i = 0; i < n; ++i)
+	  {
+	    if (getbytes >= 0)
+	      { *p = buffer[i]; ++p; --getbytes; }
+	  }
+        length -= n;
+    }
+
+    while( length ) {
+        nbytes = length < 255? length : 255;
+
+        buffer[0] = 2; /* blocking */
+        buffer[1] = nbytes;
+        if( do_write( fd, buffer, 2 ) == -1 )
+	  {
+	    myerror = errno;
+	    sh_error_handle ((-1), FIL__, __LINE__, myerror, MSG_E_SUBGEN,
+			     _("cannot write to EGD"),
+			     _("sh_entropy") ); 
+	    SL_RETURN( -1, _("sh_entropy") );
+	  }
+        n = do_read( fd, buffer, nbytes );
+        if( n == -1 ) {
+            myerror = errno;
+            sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, myerror,MSG_E_SUBGEN,
+			     _("read error on EGD"),
+			     _("sh_entropy") ); 
+            do_restart = 1;
+            goto restart;
+        }
+	for (i = 0; i < n; ++i)
+	  {
+	    if (getbytes >= 0)
+	      { *p = buffer[i]; ++p; --getbytes; }
+	  }
+        length -= n;
+    }
+    memset(buffer, 0, sizeof(buffer) );
+
+    SL_RETURN( 0, _("sh_entropy") ); /* success */
+}
+
+/* HAVE_EGD_RANDOM */
+#endif
+
 #if defined (HAVE_URANDOM)
 
 #include <setjmp.h>
@@ -93,9 +299,18 @@ int read_mbytes(int timeout_val, char * path, char * nbuf, int nbytes)
   int fd2;
   struct  sigaction  def_act;
   struct  sigaction  new_act;
-  def_act.sa_handler = SIG_IGN;
-  new_act.sa_handler = sh_entropy_alarmhandle;
+  sigset_t           unblock;
 
+  def_act.sa_handler = SIG_IGN;
+
+  new_act.sa_handler = sh_entropy_alarmhandle;
+  sigemptyset( &new_act.sa_mask );       /* set an empty mask       */
+  new_act.sa_flags = 0;                  /* init sa_flags           */
+
+  sigemptyset(&unblock);
+  sigaddset  (&unblock, SIGALRM);
+
+  SL_ENTER(_("read_mbytes"));
 
   if ((fd2 = aud_open (FIL__, __LINE__, SL_NOPRIV, path, O_RDONLY, 0)) >= 0) 
     {
@@ -109,15 +324,18 @@ int read_mbytes(int timeout_val, char * path, char * nbuf, int nbytes)
 	   */
 	  if (setjmp(entropy_timeout) != 0)
 	    {
+	      alarm(0);
+	      sigaction (SIGALRM, &def_act, NULL);
+	      sigprocmask(SIG_UNBLOCK, &unblock, NULL);
 	      TPT((0,FIL__,__LINE__, _("msg=<read_mbytes: timeout>\n"))); 
 	      close (fd2);
-	      return 0;
+	      SL_RETURN(0, _("read_mbytes"));
 	    }
 
 	  /* timeout after 30 seconds
 	   */
 	  alarm(0);
-	  sigaction (SIGALRM, &new_act, NULL);
+	  sigaction (SIGALRM, &new_act, &def_act);
 	  alarm(timeout_val);
 
 	  m_count = 0;
@@ -158,7 +376,7 @@ int read_mbytes(int timeout_val, char * path, char * nbuf, int nbytes)
   TPT((0, FIL__, __LINE__, _("msg=<read_mbytes: OK>\n"))); 
   sigaction (SIGALRM, &def_act, NULL); 
   
-  return m_count;
+  SL_RETURN(m_count, _("read_mbytes"));
 }
 
 /* Read nbytes bytes from /dev/random, mix them with 
@@ -171,6 +389,7 @@ int sh_entropy(int nbytes, char * nbuf)
   char * keybuf;
   char   addbuf[2 * KEY_BYT];
 
+  SL_ENTER(_("sh_entropy"));
 
   ASSERT((nbytes <= KEY_BYT), _("nbytes <= KEY_BYT"))
 
@@ -186,9 +405,15 @@ int sh_entropy(int nbytes, char * nbuf)
 #endif
 
   if (m_count == 0)
-    sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_NODEV, 
-		     (long) sh.real.uid, NAME_OF_DEV_RANDOM);
-
+    {
+#ifdef NAME_OF_DEV_URANDOM
+      sh_error_handle (SH_ERR_NOTICE, FIL__, __LINE__, EIO, MSG_NODEV, 
+		       (long) sh.real.uid, NAME_OF_DEV_RANDOM);
+#else
+      sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_NODEV, 
+		       (long) sh.real.uid, NAME_OF_DEV_RANDOM);
+#endif
+    }
 
 #ifdef NAME_OF_DEV_URANDOM
   if (m_count < nbytes)
@@ -213,7 +438,7 @@ int sh_entropy(int nbytes, char * nbuf)
       for (i = 0; i < KEY_BYT; ++i)
 	addbuf[i+KEY_BYT] = skey->poolv[i];
       keybuf = (char *) sh_tiger_hash_uint32 (addbuf, 
-					      TIGER_DATA, sizeof(addbuf));
+					      TIGER_DATA, 2 * KEY_BYT);
       memset(addbuf, '\0', sizeof(addbuf));
       
       /* -- Give out nbytes bytes from the new pool. --
@@ -226,14 +451,18 @@ int sh_entropy(int nbytes, char * nbuf)
 	}
       memset (keybuf, '\0', KEY_BYT);
       
-      return 0;
+      SL_RETURN(0, _("sh_entropy"));
     }
   else
-    return (-1);
+    {
+      SL_RETURN((-1), _("sh_entropy"));
+    }
 }
 
-  
-#else
+/* HAVE_URANDOM */
+#endif
+
+#ifdef HAVE_UNIX_RANDOM
 
 #ifndef FD_SET
 #define NFDBITS         32
@@ -340,6 +569,7 @@ sourcetable_t source[] = {
 };
 
 
+extern struct passwd * sh_getpwnam(const char *name);
 
 static FILE * sh_popen (sourcetable_t  *source, char * command)
 {
@@ -350,6 +580,8 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
   char * arg[4];
   char * envp[2];
 
+  SL_ENTER(_("sh_popen"));
+
   arg[0] = _("/bin/sh");
   arg[1] = _("-c");
   arg[2] = command;
@@ -357,8 +589,11 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
 
   if (sh.timezone != NULL)
     {
-      envp[0] = my_malloc (sl_strlen(sh.timezone) + 4);
-      sprintf (envp[0], "TZ=%s", sh.timezone);      /* known to fit  */
+      envp[0] = malloc (sl_strlen(sh.timezone) + 4);
+      if (envp[0] != NULL)
+	sprintf (envp[0], "TZ=%s", sh.timezone);         /* known to fit  */
+      else
+	envp[0] = NULL;
       envp[1] = NULL;
     }
   else
@@ -370,8 +605,8 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
   /* Create the pipe 
    */
   if (aud_pipe(FIL__, __LINE__, pipedes) < 0) {
-    if (envp[0] != NULL) my_free(envp[0]);
-    return (NULL);
+    if (envp[0] != NULL) free(envp[0]);
+    SL_RETURN(NULL, _("sh_popen"));
   }
   
   source->pid = aud_fork(FIL__, __LINE__);
@@ -381,8 +616,8 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
   if (source->pid == (pid_t) - 1) {
     close(pipedes[0]);
     close(pipedes[1]);
-    if (envp[0] != NULL) my_free(envp[0]);
-    return (NULL);
+    if (envp[0] != NULL) free(envp[0]);
+    SL_RETURN(NULL, _("sh_popen"));
   }
 
   if (source->pid == (pid_t) 0) 
@@ -412,8 +647,12 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
       i = 0; 
       if (0 == geteuid()) {  
 	tempres = getpwnam(DEFAULT_IDENT);
+	if (NULL == tempres) 
+	  tempres = sh_getpwnam(DEFAULT_IDENT);
 	if (NULL != tempres) {
-	  i = aud_setgid(FIL__, __LINE__, tempres->pw_gid);  
+	  i = aud_setgid(FIL__, __LINE__, tempres->pw_gid); 
+	  if (i == 0)
+	    i = sh_unix_initgroups(DEFAULT_IDENT ,tempres->pw_gid);
 	  if (i == 0) 
 	    i = aud_setuid(FIL__, __LINE__, tempres->pw_uid);
 	  /* make sure we cannot get root again
@@ -444,10 +683,10 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
     /* parent
      */
     if (envp[0] != NULL) 
-      my_free(envp[0]);
+      free(envp[0]);
 
     close (pipedes[STDOUT_FILENO]);
-    retry_fcntl (pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
+    retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
 
     MBLK( outf = fdopen (pipedes[STDIN_FILENO], "r"); )
 
@@ -457,10 +696,10 @@ static FILE * sh_popen (sourcetable_t  *source, char * command)
 	close (pipedes[STDOUT_FILENO]);
         waitpid (source->pid, NULL, 0);
         source->pid = 0;
-        return (NULL);
+	SL_RETURN(NULL, _("sh_popen"));
       }
 
-    return (outf);
+    SL_RETURN(outf, _("sh_popen"));
 }
 
 
@@ -468,19 +707,23 @@ static int sh_pclose (sourcetable_t *source)
 {
     int status = 0;
 
-    MBLK( status = fclose(source->pipe) )
+    SL_ENTER(_("sh_pclose"));
+
+    status = fclose(source->pipe);
     if (status)
-      return (-1);
+      {
+	SL_RETURN((-1), _("sh_pclose"));
+      }
 
     if (waitpid(source->pid, NULL, 0) != source->pid)
       status = -1;
 
     source->pipe = NULL;
     source->pid = 0;
-    return (status);
+    SL_RETURN(status, _("sh_pclose"));
 }
 
-#define BUF_ENT 8190
+#define BUF_ENT 32766
 
 /* Poll the system for randomness, mix results with 
  * previous reads using a hash function, and give out
@@ -503,6 +746,8 @@ int sh_entropy(int nbytes, char * nbuf)
   int    maxFD = 0;
   int    imax, selcount;
 
+  SL_ENTER(_("sh_entropy"));
+
   ASSERT((nbytes <= KEY_BYT), _("nbytes <= KEY_BYT"))
 
   if (nbytes > KEY_BYT)
@@ -519,7 +764,7 @@ int sh_entropy(int nbytes, char * nbuf)
 	  nbuf[i] = skey->poolv[i+j];
 	  --skey->poolc;
 	}
-      return (0);
+      SL_RETURN(0, _("sh_entropy"));
     }
 
 
@@ -565,7 +810,7 @@ int sh_entropy(int nbytes, char * nbuf)
 			 combuf, (long) source[i].pipeFD);
 
 	maxFD = (source[i].pipeFD > maxFD) ? source[i].pipeFD : maxFD;
-	retry_fcntl( source[i].pipeFD, F_SETFL, O_NONBLOCK);
+	retry_fcntl( FIL__, __LINE__, source[i].pipeFD, F_SETFL, O_NONBLOCK);
 	FD_SET( source[i].pipeFD, &fds );
 	source[i].isset = 1;
 	++icount;
@@ -614,10 +859,14 @@ int sh_entropy(int nbytes, char * nbuf)
 		       source[i].pipe );
 	if (count == 0) 
 	  {
+	    if (0 != feof(source[i].pipe))
+	      sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_ENCLOS,
+			       (long) source[i].pipeFD);
+	    else
+	      sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_ENCLOS1,
+			       (long) source[i].pipeFD);
 	    source[i].isset = 0;
 	    sh_pclose ( &source[i] );
-	    sh_error_handle ((-1), FIL__, __LINE__, EIO, MSG_ENCLOS,
-			     (long) source[i].pipeFD);
 	    --icount;
 	  }
 	else
@@ -683,7 +932,7 @@ int sh_entropy(int nbytes, char * nbuf)
   else 
     {
       SH_FREE(buffer);
-      return (-1);
+      SL_RETURN((-1), _("sh_entropy"));
     }
 
   /* give out nbytes Bytes from the entropy pool
@@ -694,10 +943,10 @@ int sh_entropy(int nbytes, char * nbuf)
       --skey->poolc;
     }
 
-  return (0);
+  SL_RETURN(0, _("sh_entropy"));
 }
 
-/* HAVE_URANDOM */
+/* HAVE_UNIX_RANDOM */
 #endif
 
 
