@@ -44,12 +44,18 @@
 #include "samhain.h"
 
 #include "sh_cat.h"
+#include "sh_database.h"
 #include "sh_error.h"
 #include "sh_utils.h"
 #include "sh_unix.h"
 #include "sh_tiger.h"
 #include "sh_mail.h"
 #include "sh_forward.h"
+#include "sh_prelude.h"
+
+#if defined(WITH_DATABASE)
+#include "sh_tools.h"
+#endif
 
 #if defined(WITH_EXTERNAL)
 #include "sh_extern.h"
@@ -60,13 +66,15 @@
 
 extern int clt_class;
 
+int flag_err_debug = SL_FALSE;
 
 int  ShDFLevel[SH_ERR_T_END];
 
 typedef struct _log_t {
   char file[SH_PATHBUF];
   char format[SH_PATHBUF];
-  char msg[SH_BUFSIZE];
+  char * msg;
+  int  msg_len;
   int  severity;
   int  class;
   int  pid;
@@ -79,14 +87,15 @@ typedef struct _log_t {
 struct  _errFlags  errFlags;
 
 static int  sh_error_init (void);
+
+inline
 static char * get_format(unsigned long msg_id, int * priority, int * class);
 
 static int sh_error_string (struct _log_t * lmsg, va_list vl);
 
 extern int  sh_log_console (char *message);
 extern int  sh_log_syslog  (int  severity, char *message);
-extern int  sh_log_file    (char *message);
-
+extern int  sh_log_file    (char *message, char * inet_peer);
 
 static int  IsInitialized = BAD;
 
@@ -103,7 +112,7 @@ int sh_error_verify (char * s)
 {
   char * foo;
 
-  if (s[0] == '/')  /* FIXME  a filename */
+  if (s[0] == '/')
     {
       foo = sh_tiger_hash_gpg (s, TIGER_FILE, 0);
       fprintf (stdout, _("%s\n"),  foo);
@@ -134,39 +143,59 @@ void sh_error_only_stderr (int flag)
 static int dbg_store = 0;
 static int dbg_flag  = 0;
 
-void sh_error_dbg_on()
+void compute_flag_err_debug()
 {
-  if (dbg_flag == 1)
-    return;
-
-  dbg_store           = errFlags.printlevel;
-  errFlags.printlevel = (SH_ERR_ALL    | SH_ERR_INFO  | SH_ERR_NOTICE | 
-			 SH_ERR_WARN   | SH_ERR_STAMP | SH_ERR_ERR    | 
-			 SH_ERR_SEVERE | SH_ERR_FATAL);
-  dbg_flag  = 1;
+  if ((errFlags.loglevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.printlevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.maillevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.exportlevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.sysloglevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.externallevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else if ((errFlags.databaselevel & SH_ERR_ALL) != 0)
+    flag_err_debug = SL_TRUE;
+  else
+    flag_err_debug = SL_FALSE;
   return;
 }
 
-void sh_error_dbg_off()
+void sh_error_dbg_switch()
 {
   if (dbg_flag == 0)
-    return;
-
-  errFlags.printlevel = dbg_store;
-  dbg_store = 0;
-  dbg_flag  = 0;
+    {
+      dbg_store           = errFlags.printlevel;
+      errFlags.printlevel = (SH_ERR_ALL    | SH_ERR_INFO  | SH_ERR_NOTICE | 
+			     SH_ERR_WARN   | SH_ERR_STAMP | SH_ERR_ERR    | 
+			     SH_ERR_SEVERE | SH_ERR_FATAL);
+      dbg_flag  = 1;
+      flag_err_debug = SL_TRUE;
+    }
+  else {
+    errFlags.printlevel = dbg_store;
+    dbg_store = 0;
+    dbg_flag  = 0;
+    compute_flag_err_debug();
+  }
   return;
 }
 
+
 static int sh_error_set_classmask (char * c, int * facility_mask)
 {
   char * p;
   int    num = 0;
   int    i;
   int    len;
+
+  SL_ENTER(_("sh_error_set_classmask"));
   
   if (c == NULL)
-    return -1;
+    SL_RETURN( -1, _("sh_error_set_classmask"));
 
   if (IsInitialized == BAD) 
     (void) sh_error_init();
@@ -193,13 +222,24 @@ static int sh_error_set_classmask (char * c, int * facility_mask)
 
     for (i = 0; i < SH_CLA_MAX; ++i)
       {
-	if (0 == strcmp(p, _(class_cat[i])))
-	  *facility_mask |= (1 << i);
+	if (i < SH_CLA_RAW_MAX) {
+	  if (0 == strcmp(p, _(class_cat[i])))
+	    *facility_mask |= (1 << i);
+	} else {
+	  if (0 == strcmp(p, _(class_cat[SH_CLA_RAW_MAX + 0])))
+	    *facility_mask |= OTHER_CLA;
+	  if (0 == strcmp(p, _(class_cat[SH_CLA_RAW_MAX + 1])))
+	    *facility_mask |= RUN_NEW;
+	  if (0 == strcmp(p, _(class_cat[SH_CLA_RAW_MAX + 2])))
+	    *facility_mask |= FIL_NEW;
+	  if (0 == strcmp(p, _(class_cat[SH_CLA_RAW_MAX + 3])))
+	    *facility_mask |= ERROR_CLA;
+	}	  
       }
 
   } while (p);
 
-  return 0;
+  SL_RETURN( 0, _("sh_error_set_classmask"));
 }
 
 int sh_error_log_mask (char * c)
@@ -226,6 +266,14 @@ int sh_error_external_mask (char * c)
 {
   return (sh_error_set_classmask(c, &(errFlags.external_class)));
 }
+int sh_error_database_mask (char * c)
+{
+  return (sh_error_set_classmask(c, &(errFlags.database_class)));
+}
+int sh_error_prelude_mask (char * c)
+{
+  return (sh_error_set_classmask(c, &(errFlags.prelude_class)));
+}
   
 
 
@@ -370,9 +418,11 @@ int sh_error_convert_level (char * str_s)
 {
   int i;
   int level = (-1);
+
+  SL_ENTER(_("sh_error_convert_level"));
   
   if (str_s == NULL)
-    return -1;
+    SL_RETURN( -1, _("sh_error_convert_level"));
 
   for (i = 0; i < SH_EEF_MAX; ++i)
     {
@@ -384,7 +434,7 @@ int sh_error_convert_level (char * str_s)
 	}
     }
 
-  return level;
+  SL_RETURN( level, _("sh_error_convert_level"));
 }
 
 
@@ -393,6 +443,8 @@ int sh_error_convert_level (char * str_s)
 int sh_error_set_iv (int iv, char *  str_s)
 {
   int level = (-1);
+
+  SL_ENTER(_("sh_error_set_iv"));
   
   if (IsInitialized == BAD) 
     (void) sh_error_init();
@@ -404,7 +456,7 @@ int sh_error_set_iv (int iv, char *  str_s)
       sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS,
 		       _("severity"), 
 		       str_s != NULL ? str_s : _("(NULL)"));
-      return (-1);
+      SL_RETURN (-1, _("sh_error_set_iv"));
     }
 
   if (iv > SH_ERR_T_START && iv < SH_ERR_T_END) 
@@ -415,9 +467,9 @@ int sh_error_set_iv (int iv, char *  str_s)
     {
       sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALL, 
 		       _("severity"), (long) iv);
-      return (-1);
+      SL_RETURN (-1, _("sh_error_set_iv"));
     }
-  return (0);
+  SL_RETURN (0, _("sh_error_set_iv"));
 }
 
 int sh_error_set_level(char * str_s, int * facility)
@@ -426,6 +478,8 @@ int sh_error_set_level(char * str_s, int * facility)
 
   int  old_facility;
 
+  SL_ENTER(_("sh_error_set_level"));
+
   if (IsInitialized == BAD) 
     (void) sh_error_init();
 
@@ -436,7 +490,7 @@ int sh_error_set_level(char * str_s, int * facility)
     {
       if (0 == sl_strncmp(str_s, _(eef_tab[0].str), sl_strlen(eef_tab[0].str)))
 	{
-	  *facility |= eef_tab[0].val;
+	  *facility |= eef_tab[0].val;  /* This is 'none' */
 	  f = GOOD;
 	}
       else if (str_s[0] == '*')
@@ -482,9 +536,10 @@ int sh_error_set_level(char * str_s, int * facility)
       *facility = old_facility; 
       sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS, 
 		       _("priority"), str_s);
-      return (-1);
+      SL_RETURN (-1, _("sh_error_set_level"));
     }
-  return (0);
+  compute_flag_err_debug();
+  SL_RETURN (0, _("sh_error_set_level"));
 }
 
 #ifdef SH_WITH_CLIENT
@@ -505,16 +560,26 @@ int sh_error_setexport(char *  str_s)
 
 /* set severity for printing
  */
+extern void dlog_set_active(int flag);
+
 int sh_error_setprint(char *  str_s)
 {
   static int reject = 0;
+  int        retval;
+
   if (reject == 1)
     return (0);
 
   if (sh.flag.opts == S_TRUE)   
     reject = 1;
 
-  return (sh_error_set_level(str_s, &errFlags.printlevel));
+  retval = sh_error_set_level(str_s, &errFlags.printlevel);
+
+  if (0 != (errFlags.printlevel & SH_ERR_INFO))
+    dlog_set_active(1);
+  if (0 != (errFlags.printlevel & SH_ERR_ALL))
+    dlog_set_active(2);
+  return retval;
 }
 
 
@@ -563,6 +628,38 @@ int sh_error_set_external (char * str_s)
 }
 #endif
 
+#if defined(WITH_DATABASE)
+/* set severity for database
+ */
+int sh_error_set_database (char * str_s)
+{
+  static int reject = 0;
+  if (reject == 1)
+    return (0);
+
+  if (sh.flag.opts == S_TRUE)  
+    reject = 1;
+
+  return (sh_error_set_level(str_s, &errFlags.databaselevel));
+}
+#endif
+
+#if defined(HAVE_LIBPRELUDE)
+/* set severity for prelude
+ */
+int sh_error_set_prelude (char * str_s)
+{
+  static int reject = 0;
+  if (reject == 1)
+    return (0);
+
+  if (sh.flag.opts == S_TRUE)  
+    reject = 1;
+
+  return (sh_error_set_level(str_s, &errFlags.preludelevel));
+}
+#endif
+
 /* set severity for mailing
  */
 int sh_error_setseverity (char * str_s)
@@ -577,6 +674,17 @@ int sh_error_setseverity (char * str_s)
   return (sh_error_set_level(str_s, &errFlags.maillevel));
 }
 
+#ifdef SH_WITH_SERVER
+static char inet_peer[SH_MINIBUF] = { '\0' };
+
+void sh_error_set_peer(const char * str)
+{
+  if (str == NULL)
+    memset(inet_peer, '\0', SH_MINIBUF);
+  else
+    sl_strlcpy(inet_peer, str, SH_MINIBUF);
+}
+#endif
   
 
 /**********************************************************
@@ -602,9 +710,18 @@ void sh_error_handle (int sev, char * file, long line,
   int    flag_inet  = S_FALSE;
   int    class_inet = clt_class;      /* initialize from global */
 
+#ifdef SH_WITH_SERVER
+  char   local_inet_peer[SH_MINIBUF];
+#endif
+
 #ifdef SH_WITH_CLIENT
   char   * ex_msg;
 #endif
+#if defined(WITH_DATABASE)
+  char   * escape_msg;
+#endif
+
+  static int    own_block = 0;
 
   /* 
    * Block a facility for errors generated
@@ -622,17 +739,46 @@ void sh_error_handle (int sev, char * file, long line,
 #if defined(WITH_EXTERNAL)
   static int external_block = 0;
 #endif
+#if defined(WITH_DATABASE)
+  static int database_block = 0;
+#endif
+#ifdef HAVE_LIBPRELUDE
+  static int prelude_block = 0;
+#endif
+
+  SL_ENTER(_("sh_error_handle"));
+
+#ifdef SH_WITH_SERVER
+  /* copy the global string into a local array
+   */
+  if ((msg_id == MSG_TCP_MSG) && (inet_peer[0] != '\0'))
+    {
+      sl_strlcpy(local_inet_peer, inet_peer, SH_MINIBUF);
+      sh_error_set_peer(NULL);
+    }
+  else
+    local_inet_peer[0] = '\0';
+#endif
 
   clt_class = (-1);      /* reset global */
 
+  if (own_block == 1)
+    {
+      SL_RET0(_("sh_error_handle"));
+    }
+
   /* --- Initialize to default values. ---
    */
+  own_block = 1;
   if (IsInitialized == BAD) 
     sh_error_init();
+  own_block = 0;
 
   /* --- Consistency checks. ---
    */
+  own_block = 1;
   fmt = get_format (msg_id, &severity, &class);
+  own_block = 0;
 
   if (class_inet != (-1))
     class = class_inet;
@@ -643,7 +789,7 @@ void sh_error_handle (int sev, char * file, long line,
       fprintf(stderr, 
 	      _("ERROR: msg=<NULL format>, file=<%s>, line=<%ld>\n"), 
 	      file, line);
-      return;
+      SL_RET0(_("sh_error_handle"));
     }
 
   /* --- Override the catalogue severity. ---
@@ -671,23 +817,37 @@ void sh_error_handle (int sev, char * file, long line,
 #ifdef WITH_EXTERNAL
        ( (errFlags.externallevel  & severity    ) == 0 ||
 	 (errFlags.external_class & (1 << class)) == 0 )     &&
+#endif
+#ifdef HAVE_LIBPRELUDE
+       ( (errFlags.preludelevel   & severity    ) == 0 ||
+	 (errFlags.prelude_class  & (1 << class)) == 0 )     &&
+#endif
+#ifdef WITH_DATABASE
+       ( (errFlags.databaselevel  & severity    ) == 0 ||
+	 (errFlags.database_class & (1 << class)) == 0 )     &&
 #endif
        ( (errFlags.maillevel     & severity    ) == 0 ||
 	 (errFlags.mail_class    & (1 << class)) == 0 )
 #ifdef SH_WITH_SERVER
       && (flag_inet == S_FALSE)
 #endif
-      )
-    return;
+       )
+    {
+      SL_RET0(_("sh_error_handle"));
+    }
 
   if ((severity & SH_ERR_NOT) != 0)
-    return;
+    {
+      SL_RET0(_("sh_error_handle"));
+    }
+
 
   /* Allocate space for the message.
    */
+  own_block = 1;
   lmsg = (struct _log_t *) SH_ALLOC(sizeof(struct _log_t));
   MLOCK( (char *) lmsg, sizeof(struct _log_t));
-  lmsg->msg[0] = '\0';
+  lmsg->msg = NULL;
 
   sl_strlcpy(lmsg->format, fmt, SH_PATHBUF);
   sl_strlcpy(lmsg->file, file, SH_PATHBUF);
@@ -695,13 +855,17 @@ void sh_error_handle (int sev, char * file, long line,
   lmsg->class    = class;
   lmsg->line     = line;
   lmsg->status   = status;
+  own_block = 0;
 
 
   /* Format the log message with timestamp etc.
+   * Allocate lmsg->msg
    */
+  own_block = 1;
   va_start (vl, msg_id);
   sh_error_string (lmsg, vl);
   va_end (vl);
+  own_block = 0;
 
 
   /* Log to stderr.
@@ -716,7 +880,10 @@ void sh_error_handle (int sev, char * file, long line,
     { 
       if (print_block == 0 && (errFlags.printlevel & SH_ERR_NOT) == 0) 
 	{
+	  /* no truncation
+	   */
 	  print_block = 1;
+	  TPT(( 0, FIL__, __LINE__, lmsg->msg)); 
 	  sh_log_console (lmsg->msg);
 	  print_block = 0;
 	}
@@ -732,9 +899,13 @@ void sh_error_handle (int sev, char * file, long line,
        */
       if ( (errFlags.sysloglevel  & severity)      != 0 &&
 	   (errFlags.syslog_class & (1 << class))  != 0 &&
+#ifndef INET_SYSLOG
 	   (flag_inet != S_TRUE)                        && /* !inet->syslog */
+#endif
 	   (errFlags.sysloglevel  & SH_ERR_NOT)    == 0 ) 
 	{
+	  /* will truncate to 1023 bytes 
+	   */
 	  if (syslog_block == 0)
 	    {
 	      syslog_block = 1;
@@ -754,6 +925,8 @@ void sh_error_handle (int sev, char * file, long line,
 	{
 	  if (external_block == 0)
 	    {
+	      /* no truncation
+	       */
 	      external_block = 1;
 	      sh_ext_execute ('l', 'o', 'g', lmsg->msg, 0);
 	      external_block = 0;
@@ -761,6 +934,46 @@ void sh_error_handle (int sev, char * file, long line,
 	}
 #endif
 
+#if defined(WITH_DATABASE)
+      /* 
+       * -- database facility 
+       */
+      if ((errFlags.databaselevel  & severity)     != 0 && 
+	  (errFlags.database_class & (1 << class)) != 0 &&
+	  (errFlags.databaselevel  & SH_ERR_NOT)   == 0 &&
+	  class != AUD)
+	{
+	  if (database_block == 0)
+	    {
+	      /* truncates; query_max is 16k
+	       */
+	      database_block = 1;
+#ifndef SH_STANDALONE
+	      if (msg_id == MSG_TCP_MSG 
+#ifdef INET_SYSLOG
+		  || msg_id == MSG_INET_SYSLOG
+#endif
+		  )
+		{
+		  /* do not escape twice
+		   */
+		  /* fprintf(stderr, "FIXME a\n"); */
+		  sh_database_insert (lmsg->msg);
+		}
+	      else
+#endif
+		{
+		  /* fprintf(stderr, "FIXME b\n"); */
+
+		  escape_msg = sh_tools_safe_name(lmsg->msg, 0);
+		  sh_database_insert (escape_msg);
+		  SH_FREE(escape_msg);
+		}
+	      database_block = 0;
+	    }
+	}
+#endif
+
       /****************************************************
        * Optionally include client code for TCP forwarding
        * to log server
@@ -777,6 +990,8 @@ void sh_error_handle (int sev, char * file, long line,
         {
           if (export_block == 0)
             {
+	      /* will truncate to 65280 bytes 
+	       */
               export_block = 1;
 	      ex_msg = SH_ALLOC (64 + sl_strlen(lmsg->msg) + 1);
 	      sprintf(ex_msg, _("%d?%d?%s"),             /* known to fit  */
@@ -800,8 +1015,11 @@ void sh_error_handle (int sev, char * file, long line,
 	{
 	  if (mail_block == 0)
 	    {
+	      /* will truncate to 998 bytes 
+	       */
 	      mail_block = 1;
 
+	      BREAKEXIT(sh_mail_msg);
 	      if ( (severity & SH_ERR_FATAL) == 0) 
 		sh_mail_pushstack (lmsg->msg);
 	      else 
@@ -812,6 +1030,30 @@ void sh_error_handle (int sev, char * file, long line,
 	}
 #endif
 
+#ifdef HAVE_LIBPRELUDE
+      if (((errFlags.preludelevel  & severity  )   != 0  &&
+	   (errFlags.prelude_class & (1 << class)) != 0  &&
+	   (errFlags.preludelevel  & SH_ERR_NOT)   == 0  &&
+	   (class != AUD)) 
+#ifdef SH_WITH_SERVER
+	     || (flag_inet == S_TRUE)
+#endif
+	  )
+	{
+	  if (prelude_block == 0)
+	    {
+	      /* will truncate to 998 bytes 
+	       */
+	      prelude_block = 1;
+
+	      BREAKEXIT(sh_prelude_alert);
+	      sh_prelude_alert (severity, class, lmsg->msg);
+
+	      prelude_block = 0;
+	    }
+	}
+#endif
+
       /* Log to logfile
        */
 
@@ -829,8 +1071,22 @@ void sh_error_handle (int sev, char * file, long line,
 	{
 	  if (log_block == 0)
 	    {
+	      /* no truncation
+	       */
 	      log_block = 1;
-	      sh_log_file (lmsg->msg);
+	      BREAKEXIT(sh_log_file);
+#ifdef SH_WITH_SERVER
+	      if (0 != sl_ret_euid())
+		{
+		  if (local_inet_peer[0] == '\0')
+		    sh_log_file (lmsg->msg, NULL);
+		  else
+		    sh_log_file (lmsg->msg, local_inet_peer);
+		}
+#else
+              sh_log_file (lmsg->msg, NULL);
+#endif
+	      /* sh_log_file (lmsg->msg); */
 	      log_block = 0;
 	    }
 	}
@@ -839,11 +1095,17 @@ void sh_error_handle (int sev, char * file, long line,
 
   /* Cleanup.
    */
+  own_block = 1;
+
+  if (lmsg->msg)
+    SH_FREE( lmsg->msg );
+
   memset ( lmsg, '\0', sizeof(struct _log_t) );
   MUNLOCK( (char *) lmsg,       sizeof(struct _log_t) );
   SH_FREE( lmsg );
+  own_block = 0;
 
-  return;
+  SL_RET0(_("sh_error_handle"));
 }
 
 
@@ -857,10 +1119,12 @@ void sh_error_handle (int sev, char * file, long line,
 
 /* --- Get the format from the message catalog. ---
  */
+inline
 static char * get_format(unsigned long msg_id, int * priority, int * class)
 {
   int i = 0;
 
+  SL_ENTER(_("get_format"));
   while (1)
     {
       if ( msg_cat[i].format == NULL )
@@ -870,25 +1134,31 @@ static char * get_format(unsigned long msg_id, int * priority, int * class)
 	{
 	  *priority = (int) msg_cat[i].priority;
 	  *class    = (int) msg_cat[i].class;
-	  return _(msg_cat[i].format);
+	  SL_RETURN (_(msg_cat[i].format), _("get_format"));
 	}
       ++i;
     }
-  return NULL;
+  SL_RETURN (NULL, _("get_format"));
 }
 
 static char * ehead_format = NULL;
 
-/* define message header
+/* allocate space for user-defined message header
  */
 int sh_error_ehead (char * str_s)
 {
   int size;
   char * s;
 
-  if (str_s == NULL)
-    return (-1);
+  SL_ENTER(_("sh_error_ehead"));
 
+  if (str_s == NULL)
+    {
+      SL_RETURN (-1, _("sh_error_ehead"));
+    }
+
+  /* ascii 34 ist t\"ttelchen
+   */
   if (str_s[0] == 34) s = &str_s[1];
   else s = str_s;
   
@@ -901,20 +1171,32 @@ int sh_error_ehead (char * str_s)
   ehead_format = SH_ALLOC(size+1);
   sl_strlcpy(ehead_format, s, size+1);
 
-  return 0;
+  SL_RETURN( 0, _("sh_error_ehead"));
 }
 
+#if !defined (VA_COPY)
+#if defined (__GNUC__) && defined (__PPC__) && (defined (_CALL_SYSV) || defined (_WIN32))
+#define VA_COPY(ap1, ap2)     (*(ap1) = *(ap2))
+#elif defined (VA_COPY_AS_ARRAY)
+#define VA_COPY(ap1, ap2)     memmove ((ap1), (ap2), sizeof (va_list))
+#else /* va_list is a pointer */
+#define VA_COPY(ap1, ap2)     ((ap1) = (ap2))
+#endif
+#endif 
+
 
 /* print an error  into string
  */
 static int sh_error_string (struct _log_t * lmsg, va_list vl)
 {
   int len;
+  int required;
   unsigned long line;
   char sev[16];
   char cla[16];
   char tst[64];
   char *p;
+  va_list       vl2;
 
   st_format rep_ehead_tab[] = {
     { 'S', S_FMT_STRING,  0, 0, NULL},  /* severity  */
@@ -926,19 +1208,55 @@ static int sh_error_string (struct _log_t * lmsg, va_list vl)
     {'\0', S_FMT_ULONG,   0, 0, NULL},
   };
 
+  SL_ENTER(_("sh_error_string"));
+
   if (ehead_format == NULL)
     {
       ehead_format = SH_ALLOC(64);
+#ifdef SH_USE_XML
+      if ((errFlags.printlevel & SH_ERR_ALL) == 0) 
+	sl_strlcpy(ehead_format, _("<log sev=\"%S\" tstamp=\"%T\" "), 64);
+      else
+	sl_strlcpy(ehead_format, _("<log sev=\"%S\" tstamp=\"%T\" p.f=\"%F\" p.l=\"%L\" p.s=\"%E\" "), 64);
+#else
       if ((errFlags.printlevel & SH_ERR_ALL) == 0) 
 	sl_strlcpy(ehead_format, _("%S %T "), 64);
       else
-	sl_strlcpy(ehead_format, _("%S %T (%F, %L) "), 64);
+	sl_strlcpy(ehead_format, _("%S %T (%F, %L, %E) "), 64);
+#endif
     }
 
   /* header of error message
    */
+#ifdef SH_USE_XML
+  if      ( (lmsg->severity & SH_ERR_INET) != 0)
+    sl_strlcpy (sev, _("RCVT"), 11);
+  else if ( (lmsg->severity & SH_ERR_ALL) != 0)
+    sl_strlcpy (sev, _("DEBG"), 11);
+  else if ( (lmsg->severity & SH_ERR_INFO) != 0)
+    sl_strlcpy (sev, _("INFO"), 11);
+  else if ( (lmsg->severity & SH_ERR_NOTICE) != 0)
+    sl_strlcpy (sev, _("NOTE"), 11);
+  else if ( (lmsg->severity & SH_ERR_WARN) != 0)
+    sl_strlcpy (sev, _("WARN"), 11);
+  else if ( (lmsg->severity & SH_ERR_STAMP) != 0)
+    sl_strlcpy (sev, _("MARK"), 11);
+  else if ( (lmsg->severity & SH_ERR_ERR) != 0)
+    sl_strlcpy (sev, _("ERRO"), 11);
+  else if ( (lmsg->severity & SH_ERR_SEVERE) != 0)
+    sl_strlcpy (sev, _("CRIT"), 11);
+  else if ( (lmsg->severity & SH_ERR_FATAL) != 0)
+    sl_strlcpy (sev, _("ALRT"), 11);
+  else {
+    sl_strlcpy (sev, _("????"), 11);
+#else
+#if defined(INET_SYSLOG)
+  if      ( (lmsg->severity & SH_ERR_INET) != 0)
+    sl_strlcpy (sev, _("<NET>  : "), 11);
+#else
   if      ( (lmsg->severity & SH_ERR_INET) != 0)
     sl_strlcpy (sev, _("<TCP>  : "), 11);
+#endif
   else if ( (lmsg->severity & SH_ERR_ALL) != 0)
     sl_strlcpy (sev, _("DEBUG  : "), 11);
   else if ( (lmsg->severity & SH_ERR_INFO) != 0)
@@ -957,6 +1275,7 @@ static int sh_error_string (struct _log_t * lmsg, va_list vl)
     sl_strlcpy (sev, _("ALERT  : "), 11);
   else {
     sl_strlcpy (sev, _("???    : "), 11);
+#endif
   }
 
   sl_strlcpy (tst, sh_unix_time (0), 64);
@@ -968,10 +1287,15 @@ static int sh_error_string (struct _log_t * lmsg, va_list vl)
   rep_ehead_tab[2].data_str   = lmsg->file;
   rep_ehead_tab[3].data_ulong = line;
   rep_ehead_tab[4].data_str   = cla;
-  rep_ehead_tab[5].data_ulong = lmsg->status;
+  rep_ehead_tab[5].data_long  = lmsg->status;
   
   p = sh_util_formatted(ehead_format, rep_ehead_tab);
 
+  /* ---  copy the header to lmsg->msg  ---
+   */
+  lmsg->msg     = SH_ALLOC(SH_BUFSIZE);
+  lmsg->msg_len = SH_BUFSIZE;
+
   if (p)
     {
       sl_strlcpy (lmsg->msg, p, SH_BUFSIZE);
@@ -982,20 +1306,40 @@ static int sh_error_string (struct _log_t * lmsg, va_list vl)
       lmsg->msg[0] = '\0';
     }
 
+  /* --- copy message to lmsg->msg ---
+   */
   if ( NULL == strchr(lmsg->format, '%') ) 
     {
-      sl_strlcat (lmsg->msg, lmsg->format, SH_BUFSIZE);
+      sl_strlcat (lmsg->msg, lmsg->format, lmsg->msg_len);
     }
-
-  /* use VA_COPY 
-   */
   else 
     {
-      len = sl_strlen(lmsg->msg);
-      sl_vsnprintf(&(lmsg->msg[len]), (SH_BUFSIZE-len), lmsg->format, vl);
+      /* use VA_COPY */
+      VA_COPY(vl2, vl);
+      len      = sl_strlen(lmsg->msg);
+      required = sl_vsnprintf(&(lmsg->msg[len]), 
+			      (lmsg->msg_len - len), lmsg->format, vl);
+      if ( (required + len) > (lmsg->msg_len - 4) )
+	{
+	  p = SH_ALLOC(required + len + 4);
+	  sl_strlcat(p, lmsg->msg, required + len + 1);
+	  SH_FREE(lmsg->msg);
+	  lmsg->msg = p;
+	  lmsg->msg_len = required + len + 4;
+	  sl_vsnprintf(&(lmsg->msg[len]), 
+			      (required + len + 1), lmsg->format, vl2);
+	}
+      va_end(vl2);
     }
 
-  return(0);
+#ifdef SH_USE_XML
+  /* closing tag
+   */
+  if (lmsg->msg[sl_strlen(lmsg->msg)-1] != '>')
+    sl_strlcat (lmsg->msg, _(" />"), lmsg->msg_len);
+#endif
+
+  SL_RETURN(0, _("sh_error_string"));
 }
 
      
@@ -1007,11 +1351,20 @@ static int  sh_error_init ()
 {
   register int j;
 
+  SL_ENTER(_("sh_error_init"));
+
   errFlags.debug          = 0;
   errFlags.HaveLog        = GOOD;
   errFlags.sysloglevel    = SH_ERR_NOT;
+#if defined(SH_STEALTH)
   errFlags.loglevel       = SH_ERR_NOT;
+#else
+  errFlags.loglevel       = (SH_ERR_STAMP | SH_ERR_ERR    | SH_ERR_SEVERE |
+			     SH_ERR_FATAL);
+#endif
   errFlags.externallevel  = SH_ERR_NOT;
+  errFlags.databaselevel  = SH_ERR_NOT;
+  errFlags.preludelevel   = SH_ERR_NOT;
   errFlags.maillevel      = SH_ERR_FATAL;
 #if defined(SH_STEALTH)
   errFlags.printlevel     = SH_ERR_NOT;
@@ -1029,12 +1382,13 @@ static int  sh_error_init ()
   errFlags.export_class   = 0xFFFF;
   errFlags.syslog_class   = 0xFFFF;
   errFlags.external_class = 0xFFFF;
+  errFlags.database_class = 0xFFFF;
+  errFlags.prelude_class  = 0xFFFF;
 
 
   for (j = 0; j < SH_ERR_T_END; ++j) 
-    ShDFLevel[j] = SH_ERR_NOT;
+    ShDFLevel[j] = SH_ERR_SEVERE;
 
   IsInitialized = GOOD;
-  return (0);
+  SL_RETURN (0, _("sh_error_init"));
 }
-
