@@ -113,7 +113,10 @@ FILE * pdbgc;
 
 
 #ifndef SH_UID
-#define SH_UID long
+#define SH_UID uid_t
+#endif
+#ifndef SH_GID
+#define SH_GID gid_t
 #endif
 
 #undef  FIL__
@@ -143,7 +146,7 @@ typedef struct _sh_com_t
 
   SH_UID   trusted_users[32];
   SH_UID   run_user_uid;
-  SH_UID   run_user_gid;
+  SH_GID   run_user_gid;
   int      privileged;
 
   int      pipeFD;
@@ -177,10 +180,14 @@ sh_com_t * command_init(void)
   uid_t       ff_euid;
   sh_com_t  * ext_com = NULL;
 
+  SL_ENTER(_("command_init"));
+
   ext_com = (sh_com_t *) SH_ALLOC(sizeof(sh_com_t));
 
   if (!ext_com)
-    return NULL;
+    {
+      SL_RETURN( NULL, ("command_init"));
+    }
 
   set3(ext_com->type, 'l', 'o', 'g');
   ext_com->command      = NULL;
@@ -206,20 +213,20 @@ sh_com_t * command_init(void)
       ext_com->fnot_v[i]        = NULL;
       ext_com->argv[i]          = NULL;
       ext_com->envv[i]          = NULL;
-      ext_com->trusted_users[i] = -1;
+      ext_com->trusted_users[i] = (uid_t) -1;
     }
 
-  ext_com->run_user_uid     = (SH_UID) getuid();
-  ext_com->run_user_gid     = (SH_UID) getgid();
+  ext_com->run_user_uid     = (uid_t) getuid();
+  ext_com->run_user_gid     = (gid_t) getgid();
 
   sl_get_euid(&ff_euid);
-  ext_com->trusted_users[0] = 0;
-  ext_com->trusted_users[1] = (SH_UID) (ff_euid);
+  ext_com->trusted_users[0] = (uid_t) 0;
+  ext_com->trusted_users[1] = (uid_t) (ff_euid);
 
 
   ext_com->next             = NULL;
 
-  return ext_com;
+  SL_RETURN( ext_com, ("command_init"));
 }
 
 /* the list of external commands
@@ -236,10 +243,14 @@ int sh_ext_add_envv(char * key, char * val)
 {
   int sk = 0, sv = 0, si;
 
+  SL_ENTER(_("sh_ext_add_envv"));
+
   if (ext_coms == NULL || ext_failed == (-1) || 
       (key == NULL      && val == NULL)      || 
       ext_coms->envc >= 30)
-    return (-1);
+    {
+      SL_RETURN (-1, _("sh_ext_add_envv"));
+    }
 
   if (key != NULL)
     sk = strlen(key) + 1;
@@ -251,7 +262,9 @@ int sh_ext_add_envv(char * key, char * val)
   ext_coms->envv[si] = SH_ALLOC(sk + sv);
 
   if (ext_coms->envv[si] == NULL)
-    return (-1);
+    {
+      SL_RETURN (-1, _("sh_ext_add_envv"));
+    }
 
   if (key != NULL)
     {
@@ -265,7 +278,7 @@ int sh_ext_add_envv(char * key, char * val)
 
   ++ext_coms->envc;
 
-  return (0);
+  SL_RETURN (0, _("sh_ext_add_envv"));
 }
 
 
@@ -276,17 +289,27 @@ int sh_ext_init(char * command)
   sh_com_t * retval;
   int        size;
 
+  SL_ENTER(_("sh_ext_init"));
+
   if (command == NULL)
-    return (-1);
+    {
+      SL_RETURN (-1, _("sh_ext_init"));
+    }
   size = strlen(command);
   if (command[0] != '/' || size < 2)
-    return (-1);
+    {
+      SL_RETURN (-1, _("sh_ext_init"));
+    }
 
   if (NULL == (retval = command_init()))
-    return (-1);
+    {
+      SL_RETURN (-1, _("sh_ext_init"));
+    }
   retval->command = SH_ALLOC(size + 1);
   if (retval->command == NULL)
-    return (-1);
+    {
+      SL_RETURN (-1, _("sh_ext_init"));
+    }
   sl_strlcpy(retval->command, command, size+1);
 
   if (sh.timezone != NULL)
@@ -296,27 +319,33 @@ int sh_ext_init(char * command)
 
   retval->next = ext_coms;
   ext_coms     = retval;
-  return 0;
+  SL_RETURN (0, _("sh_ext_init"));
 }
 
+extern struct passwd * sh_getpwnam(const char *name);
 
 static
 int sh_ext_uid (char * user, uid_t * uid, gid_t * gid)
 {
   struct passwd * tempres;
 
-  if (user == NULL)
-    return (-1);
+  SL_ENTER(_("sh_ext_uid"));
 
+  if (user == NULL)
+    {
+      SL_RETURN (-1, _("sh_ext_uid"));
+    }
   tempres = getpwnam(user);
+  if (NULL == tempres)
+    tempres = sh_getpwnam(user);
   if (NULL != tempres) 
     {
       *uid = tempres->pw_uid;  
       *gid = tempres->pw_gid;
-      return 0;
+      SL_RETURN (0, _("sh_ext_uid"));
     } 
-  else 
-    return (-1);
+
+  SL_RETURN (-1, _("sh_ext_uid"));
 }
 
 
@@ -327,8 +356,12 @@ int sh_ext_add (char * argstring, int * ntok, char * stok[])
   int    s;
   char * p;
 
+  SL_ENTER(_("sh_ext_add"));
+
   if (NULL == argstring)
-    return (-1);
+    {
+      SL_RETURN((-1), _("sh_ext_add")); 
+    }
 
   do
     {
@@ -353,7 +386,7 @@ int sh_ext_add (char * argstring, int * ntok, char * stok[])
 
   *ntok = i;
 
-  return (0);
+  SL_RETURN (0, _("sh_ext_add"));
 }
 
 /*********************************************************
@@ -369,11 +402,13 @@ int sh_ext_add (char * argstring, int * ntok, char * stok[])
 int sh_ext_setcommand(char * cmd)
 {
   int i;
+
+  SL_ENTER(_("sh_ext_setcommand"));
   if ( (i = sh_ext_init(cmd)) < 0)
     ext_failed = -1;
   else
     ext_failed = 0;
-  return i;
+  SL_RETURN( i, _("sh_ext_setcommand"));
 }
 
 
@@ -385,6 +420,8 @@ int sh_ext_cleanup(void)
   int i;
   sh_com_t * retval;
 
+  SL_ENTER(_("sh_ext_cleanup"));
+
   while (ext_coms != NULL)
     {
       retval   = ext_coms;
@@ -405,7 +442,7 @@ int sh_ext_cleanup(void)
 
     }
 
-  return (0);
+  SL_RETURN (0, _("sh_ext_cleanup"));
 }
 
 /*
@@ -457,13 +494,21 @@ int sh_ext_add_default (char * dummy)
    * an 'unused parameter' warning
    */
   char * p = (dummy == NULL ? dummy : NULL);
+  int    i;
 
+  SL_ENTER(_("sh_ext_add_default"));
+  if (dummy[0] == 'n' ||  dummy[0] == 'N' ||
+      dummy[0] == 'f' ||  dummy[0] == 'F' || dummy[0] == '0')
+    {
+      SL_RETURN(0, _("sh_ext_add_default"));
+    }
   p = sh_unix_getUIDdir (SH_ERR_ERR, (uid_t) ext_coms->run_user_uid);
   if (p)
     sh_ext_add_envv (_("HOME"), p);
-   sh_ext_add_envv (_("SHELL"), _("/bin/sh")); 
-   sh_ext_add_envv (_("PATH"),  _("/sbin:/usr/sbin:/bin:/usr/bin")); 
-   return (p == NULL ? (-1) :  0);
+  sh_ext_add_envv (_("SHELL"), _("/bin/sh")); 
+  sh_ext_add_envv (_("PATH"),  _("/sbin:/usr/sbin:/bin:/usr/bin")); 
+  i = (p == NULL ? (-1) :  0);
+  SL_RETURN(i, _("sh_ext_add_default"));
 }
 
 /*
@@ -471,7 +516,10 @@ int sh_ext_add_default (char * dummy)
  */
 int sh_ext_add_environ (char * str)
 {
-  return (sh_ext_add_envv (NULL, str));
+  int i;
+  SL_ENTER(_("sh_ext_add_environ"));
+  i = sh_ext_add_envv (NULL, str);
+  SL_RETURN(i, _("sh_ext_add_environ"));
 }
 
 /*
@@ -482,15 +530,20 @@ int sh_ext_deadtime (char * str)
   long    deadtime = 0;
   char  * tail     = NULL;
 
-  if (ext_coms == NULL || ext_failed == (-1) || str == NULL)
-    return (-1);
+  SL_ENTER(_("sh_ext_deadtime"));
 
+  if (ext_coms == NULL || ext_failed == (-1) || str == NULL)
+    {
+      SL_RETURN (-1, ("sh_ext_deadtime"));
+    }
   deadtime = strtol(str, &tail, 10);
   if (tail == str || deadtime < 0 || deadtime == LONG_MAX)
-    return (-1);
+    {
+      SL_RETURN (-1, ("sh_ext_deadtime"));
+    }
   
   ext_coms->deadtime = (time_t) deadtime;  
-  return (0);
+  SL_RETURN (0, ("sh_ext_deadtime"));  
 }
 
 /*
@@ -498,11 +551,17 @@ int sh_ext_deadtime (char * str)
  */
 int sh_ext_type (char * str)
 {
+  SL_ENTER(_("sh_ext_type"));
+
   if (ext_coms == NULL || ext_failed == (-1) || str == NULL)
-    return (-1);
+    {
+      SL_RETURN((-1), _("sh_ext_type"));
+    }
 
   if (strlen(str) != 3)
-    return (-1);
+    {
+      SL_RETURN((-1), _("sh_ext_type"));
+    }
 
   set3(ext_coms->type, str[0], str[1], str[2]);
 
@@ -513,9 +572,11 @@ int sh_ext_type (char * str)
   else if (str[0] == 'm' && str[1] == 'o' && str[2] == 'n')
     ext_coms->rw = 'r';
   else
-    return (-1);
+    {
+      SL_RETURN((-1), _("sh_ext_type"));
+    }
 
-  return (0);
+  SL_RETURN(0, _("sh_ext_type"));
 } 
   
 
@@ -525,15 +586,20 @@ int sh_ext_type (char * str)
  */
 int sh_ext_checksum (char * str)
 {
+  SL_ENTER(_("sh_ext_checksum"));
   if (ext_coms == NULL || ext_failed == (-1) || str == NULL)
-    return (-1);
+    {
+      SL_RETURN((-1), _("sh_ext_checksum"));
+    }
 
   if (sl_strlen(str) != KEY_LEN)
-    return (-1);
+    {
+      SL_RETURN((-1), _("sh_ext_checksum"));
+    }
 
   sl_strlcpy (ext_coms->checksum, str, KEY_LEN+1);
 
-  return (0);
+  SL_RETURN((0), _("sh_ext_checksum"));
 }
 
 /*
@@ -545,16 +611,17 @@ int sh_ext_priv (char * c)
   uid_t m_uid;
   gid_t m_gid;
 
+  SL_ENTER(_("sh_ext_priv"));
   if (0 == sh_ext_uid (c, &m_uid, &m_gid))
     {
-      ext_coms->run_user_uid = (SH_UID) m_uid;
-      ext_coms->run_user_gid = (SH_UID) m_gid;
+      ext_coms->run_user_uid = (uid_t) m_uid;
+      ext_coms->run_user_gid = (gid_t) m_gid;
       if (m_uid != (uid_t) 0)
 	ext_coms->privileged   = 0;
-      return (0);
+      SL_RETURN((0), _("sh_ext_priv"));
     }
 
-  return (-1);
+  SL_RETURN (-1, _("sh_ext_priv"));
 }
 
 
@@ -581,6 +648,8 @@ int sh_ext_popen (sh_com_t * task)
   int         pfd;
 #endif
   
+  SL_ENTER(_("sh_ext_popen"));
+
   /* Linux, HP-UX and FreeBSD will happily accept envp = argp = NULL
    * Solaris (and probably some other Unices) 
    *         needs a valid *envp[] with envp[0] = NULL;
@@ -613,7 +682,7 @@ int sh_ext_popen (sh_com_t * task)
 			  sl_error_string(status), tmp);
 	}
       SH_FREE(tmp);
-      return (-1);
+      SL_RETURN ((-1), _("sh_ext_popen"));
     }
 
   PDBG(1);
@@ -632,7 +701,7 @@ int sh_ext_popen (sh_com_t * task)
 	  tmp  = sh_util_safe_name (task->command);
 	  sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_HASH, tmp);
 	  SH_FREE(tmp);
-	  return (-1);
+	  SL_RETURN ((-1), _("sh_ext_popen"));
  	}
     }
 
@@ -647,7 +716,7 @@ int sh_ext_popen (sh_com_t * task)
       errnum = errno;
       sh_error_handle((-1), FIL__, __LINE__, errnum, MSG_E_SUBGEN, 
 		      sh_error_message(errnum), _("pipe"));
-      return (-1);
+      SL_RETURN ((-1), _("sh_ext_popen"));
     }
 
   PDBG(3);
@@ -664,7 +733,7 @@ int sh_ext_popen (sh_com_t * task)
       errnum = errno;
       sh_error_handle((-1), FIL__, __LINE__, errnum, MSG_E_SUBGEN, 
 		      sh_error_message(errnum), _("fork"));
-      return (-1);
+      SL_RETURN ((-1), _("sh_ext_popen"));
     }
   
   PDBG(4);
@@ -774,7 +843,7 @@ int sh_ext_popen (sh_com_t * task)
 
 	      PDBGC(5);
 	      pfd = get_the_fd(fd);
-	      sprintf(pname, _("/proc/self/fd/%d"),      /*  known to fit */
+	      sprintf(pname, _("/proc/self/fd/%d"),      /* known to fit  */
 			   pfd);
 	      if (0 == access(pname, R_OK|X_OK))
 		{
@@ -840,29 +909,30 @@ int sh_ext_popen (sh_com_t * task)
     {
       PDBG_S("is w");
       close (pipedes[STDIN_FILENO]);
-      retry_fcntl (pipedes[STDOUT_FILENO], F_SETFD, FD_CLOEXEC);
+      retry_fcntl (FIL__, __LINE__, pipedes[STDOUT_FILENO], F_SETFD, FD_CLOEXEC);
       MBLK( outf = fdopen (pipedes[STDOUT_FILENO], "w"); )
     }
   else
     {
       PDBG_S("is r");
       close (pipedes[STDOUT_FILENO]);
-      retry_fcntl (pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
+      retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
       MBLK( outf = fdopen (pipedes[STDIN_FILENO], "r"); )
     }
 
   
   if (outf == NULL) 
     {
+      errnum = errno;
       PDBG_S("outf == NULL");
       tmp  = sh_util_safe_name (task->command);
       
       if (task->privileged == 0 && 0 == getuid())
-	sh_error_handle((-1), FIL__, __LINE__, MSG_NOEXEC,
-			(long) task->run_user_uid, tmp);
+	sh_error_handle((-1), FIL__, __LINE__, errnum, MSG_NOEXEC,
+			(UID_CAST) task->run_user_uid, tmp);
       else
-	sh_error_handle((-1), FIL__, __LINE__, MSG_NOEXEC,
-			(long) getuid(), tmp);
+	sh_error_handle((-1), FIL__, __LINE__, errnum, MSG_NOEXEC,
+			(UID_CAST) getuid(), tmp);
 
       SH_FREE(tmp);
 
@@ -871,7 +941,7 @@ int sh_ext_popen (sh_com_t * task)
       waitpid (task->pid, NULL, 0);
       task->pid = 0;
 
-      return (-1);
+      SL_RETURN ((-1), _("sh_ext_popen"));
     }
   
   if (task->rw == 'w')
@@ -881,14 +951,14 @@ int sh_ext_popen (sh_com_t * task)
 
   PDBG_D(task->pipeFD);
 
-  flags = retry_fcntl (task->pipeFD, F_GETFL, 0);
+  flags = retry_fcntl (FIL__, __LINE__, task->pipeFD, F_GETFL, 0);
   if (flags != (-1))
-    retry_fcntl (task->pipeFD, F_SETFL, flags|O_NONBLOCK);
+    retry_fcntl (FIL__, __LINE__, task->pipeFD, F_SETFL, flags|O_NONBLOCK);
   task->pipe     = outf;
 
   PDBG_S("return from popen");
   
-  return (0);
+  SL_RETURN (0, _("sh_ext_popen"));
 }
 
 
@@ -899,6 +969,8 @@ int sh_ext_pclose (sh_com_t * task)
 {
   int status = 0;
 
+  SL_ENTER(_("sh_ext_pclose"));
+
   PDBG_S(" -> pclose");
   fflush(task->pipe);
   MBLK( fclose(task->pipe); )
@@ -908,7 +980,7 @@ int sh_ext_pclose (sh_com_t * task)
   task->pipeFD   = (-1);
 
   PDBG_S(" <--");
-  return (status);
+  SL_RETURN (status, _("sh_ext_pclose"));
 }
 
 /*
@@ -920,13 +992,19 @@ int sh_ext_filter (char * message, sh_com_t * task)
   int j = 0;
   time_t now_time;
 
+  SL_ENTER(_("sh_ext_filter"));
+
   /* Presence of any of these keywords prevents execution.
    */
   if (task->fnot_c > 0)
     {
       for (i = 0; i < task->fnot_c; ++i)
-	if (NULL != sl_strstr(message, task->fnot_v[i]))
-	  return (-1);
+	{
+	  if (NULL != sl_strstr(message, task->fnot_v[i]))
+	    {
+	      SL_RETURN ((-1), _("sh_ext_filter"));
+	    }
+	}
     }
 
   /* Presence of all of these keywords is required for execution.
@@ -940,7 +1018,10 @@ int sh_ext_filter (char * message, sh_com_t * task)
 	  ++j;
 
       if (j != task->fand_c)
-	return (-1);
+	{
+	  SL_RETURN ((-1), _("sh_ext_filter"));
+	}
+
     }
 
   /* Presence of at least one of these keywords is required for execution.
@@ -961,20 +1042,20 @@ int sh_ext_filter (char * message, sh_com_t * task)
 		    }
 		  else if ((time_t)(now_time-task->last_run) < task->deadtime)
 		    {
-		      return (-1);
+		      SL_RETURN ((-1), _("sh_ext_filter"));
 		    }
 		  else
 		    {
 		      task->last_run = now_time;
 		    }
 		}
-	      return (0);
+	      SL_RETURN ((0), _("sh_ext_filter"));
 	    }
 	}
-      return (-1);
+      SL_RETURN ((-1), _("sh_ext_filter"));
     }
 
-  return (0);
+  SL_RETURN ((0), _("sh_ext_filter"));
 }
 
 
@@ -993,10 +1074,14 @@ int sh_ext_execute (char t1, char t2, char t3, char * message, int msg_siz)
   struct  sigaction  new_act;
   struct  sigaction  old_act;
 
+  SL_ENTER(_("sh_ext_execute"));
+
   PDBG_OPEN;
 
   if (listval == NULL)
-    return (-1);
+    {
+      SL_RETURN ((-1), _("sh_ext_execute"));
+    }
 
   PDBG(-1);
 
@@ -1007,7 +1092,7 @@ int sh_ext_execute (char t1, char t2, char t3, char * message, int msg_siz)
   /* ignore SIGPIPE (instead get EPIPE if connection is closed)
    */
   new_act.sa_handler = SIG_IGN;
-  retry_sigaction (SIGPIPE, &new_act, &old_act);
+  retry_sigaction (FIL__, __LINE__, SIGPIPE, &new_act, &old_act);
 
   while (listval != NULL)
     {
@@ -1062,11 +1147,11 @@ int sh_ext_execute (char t1, char t2, char t3, char * message, int msg_siz)
 			{
 			  if (listval->privileged == 0 && 
 			      (0 == getuid() || sl_is_suid()) )
-			    sh_error_handle((-1), FIL__, __LINE__, MSG_NOEXEC,
-					    (long) listval->run_user_uid, tmp);
+			    sh_error_handle((-1), FIL__, __LINE__, 0, MSG_NOEXEC,
+					    (UID_CAST) listval->run_user_uid, tmp);
 			  else
-			    sh_error_handle((-1), FIL__, __LINE__, MSG_NOEXEC,
-					    (long) getuid(), tmp);
+			    sh_error_handle((-1), FIL__, __LINE__, 0, MSG_NOEXEC,
+					    (UID_CAST) getuid(), tmp);
 			  
 			  SH_FREE(tmp);
 			}
@@ -1089,10 +1174,11 @@ int sh_ext_execute (char t1, char t2, char t3, char * message, int msg_siz)
 
   /* restore old signal handler
    */
-  retry_sigaction (SIGPIPE, &old_act, NULL);
+  retry_sigaction (FIL__, __LINE__, SIGPIPE, &old_act, NULL);
   PDBG_S("return");
   PDBG_CLOSE;
-  return 0;
+
+  SL_RETURN ((0), _("sh_ext_execute"));
 }
   
   
