@@ -65,6 +65,7 @@
 #include "sh_files.h"
 #include "sh_tiger.h"
 #include "sh_hash.h"
+#include "sh_ignore.h"
 
 #undef  FIL__
 #define FIL__  _("sh_files.c")
@@ -99,7 +100,7 @@ typedef struct dir_struct {
   long    NumBDev;
   long    NumAll;
   long    TotalBytes;
-  char    DirPath[PATH_MAX+2];
+  char    DirPath[PATH_MAX];
 } dir_type;
 
 typedef struct dirstack_entry {
@@ -113,7 +114,12 @@ typedef struct dirstack_entry {
   struct dirstack_entry * next;
 } dirstack_t;
 
-static dirstack_t * dirList      = NULL;
+/* new 1.4.8
+ */
+static dirstack_t * dirListOne   = NULL;
+static dirstack_t * dirListTwo   = NULL;
+
+/* static dirstack_t * dirList      = NULL; */
 static dirstack_t * fileList     = NULL;
 
 static int        sh_files_fullpath  (char * testdir, char * d_name, 
@@ -170,8 +176,10 @@ unsigned long sh_files_chk ()
 
   while (ptr) 
     {
-      if (sig_terminate == 1)
+
+      if (sig_urgent == 1) {
 	SL_RETURN(fcount, _("sh_files_chk"));
+      }
 
       if (ptr->checked == S_FALSE)
 	{
@@ -181,13 +189,14 @@ unsigned long sh_files_chk ()
 	  
 	  sh_error_handle ((-1),  FIL__, __LINE__, 0, MSG_FI_CHK,
 			   tmp);
-	  
+
+	  BREAKEXIT(sh_files_filecheck);
 	  status = sh_files_filecheck (ptr->class, base, file, 
 				       (int *) &(ptr->reported));
 	  
 	  TPT(( 0, FIL__, __LINE__, 
-		_("msg=<filecheck complete: %s> status=<%d>\n"), 
-		tmp, status));
+		_("msg=<filecheck complete: %s> status=<%d> reported=<%d>\n"), 
+		tmp, status, ptr->reported));
 
 	  if (status == SH_FILE_UNKNOWN && ptr->reported == S_FALSE) 
 	    {
@@ -196,49 +205,81 @@ unsigned long sh_files_chk ()
 
 	      if ( sh.flag.checkSum == SH_CHECK_INIT  || 
 		  sh_hash_have_it (ptr->name) >= 0)
-		  sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
-				   ShDFLevel[ptr->class] : 
-				   ShDFLevel[SH_ERR_T_FILE],
-				   FIL__, __LINE__, 0, MSG_FI_MISS,
-				   tmp);
-
+		{
+		  if (S_FALSE == sh_ignore_chk_del(ptr->name))
+		    {
+		      sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
+				       ShDFLevel[ptr->class] : 
+				       ShDFLevel[SH_ERR_T_FILE],
+				       FIL__, __LINE__, 0, MSG_FI_MISS,
+				       tmp);
+		    }
+		}
 	      else /* not there at init, and still missing */
+		{
 		  sh_error_handle (SH_ERR_NOTICE,
 				   FIL__, __LINE__, 0,
 				   MSG_FI_FAIL,
 				   tmp);
-	      if (sh.flag.checkSum != SH_CHECK_INIT)
+		}
+#ifndef REPLACE_OLD
+	      /* this will tell that we have seen the file, and thus prevent
+	       * deletion from the database, resulting in an incomplete
+	       * message when the file reappears
+	       */
+	      if (sh.flag.checkSum != SH_CHECK_INIT) 
 		sh_hash_set_visited_true(ptr->name);
+#else
+	      if (sh.flag.checkSum != SH_CHECK_INIT) 
+		sh_hash_set_missing(ptr->name);
+#endif
 	      if (sh.flag.reportonce == S_TRUE)
 		ptr->reported = S_TRUE;
 	    }
-	  else
+	  else 
 	    {
-	      /* Was missing once, but is there again
+	      /* exists (status >= 0), but was missing (reported == TRUE)
 	       */
 	      if (status != SH_FILE_UNKNOWN && ptr->reported == S_TRUE)
 		{
 		  ptr->reported = S_FALSE;
-		  /* if (0 == sh_hash_have_it (ptr->name)) */
-		    sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
-				     ShDFLevel[ptr->class] : 
-				     ShDFLevel[SH_ERR_T_FILE],
-				     FIL__, __LINE__, 0, MSG_FI_ADD,
-				     tmp);
-		    /*
-		  else 
-		    sh_error_handle (SH_ERR_NOTICE, FIL__, __LINE__, 0,
-				     MSG_FI_ADD,
-				     tmp);
-		    */
 		}
+	      /* Catchall
+	       */
 	      else if (status == SH_FILE_UNKNOWN)
 		{
-		  sh_error_handle (SH_ERR_INFO, FIL__, __LINE__, 0,
-				   MSG_FI_FAIL,
-				   tmp);
-		  if (sh.flag.checkSum != SH_CHECK_INIT)
-		    sh_hash_set_visited_true(ptr->name);
+		  /* Thu Mar  7 15:09:40 CET 2002 Make sure missing file
+		   * is reported if ptr->reported == S_TRUE because the
+		   * file has been added.
+		   */
+		  if (sh_hash_have_it (ptr->name) >= 0)
+		    {
+		      if (S_FALSE == sh_ignore_chk_del(ptr->name))
+			{
+			  sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE)? 
+					   ShDFLevel[ptr->class] : 
+					   ShDFLevel[SH_ERR_T_FILE],
+					   FIL__, __LINE__, 0, MSG_FI_MISS,
+					   tmp);
+			}
+#ifndef REPLACE_OLD
+		      if (sh.flag.checkSum != SH_CHECK_INIT) 
+			sh_hash_set_visited_true(ptr->name);
+#else
+		      /* delete from database
+		       */
+		      if (sh.flag.checkSum != SH_CHECK_INIT) 
+			sh_hash_set_missing(ptr->name);
+#endif
+		    }
+		  else
+		    {
+		      sh_error_handle (SH_ERR_INFO, FIL__, __LINE__, 0,
+				       MSG_FI_FAIL,
+				       tmp);
+		      if (sh.flag.checkSum != SH_CHECK_INIT)
+			sh_hash_set_visited_true(ptr->name);
+		    }
 		}
 	      ++fcount;
 	    }
@@ -271,39 +312,46 @@ int sh_files_delfilestack ()
   SL_RETURN(0, _("sh_files_delfilestack"));
 }
   
-int sh_files_setrec ()
+int sh_files_setrec_int (dirstack_t * ptr)
 {
-  dirstack_t * ptr = dirList;
-
   SL_ENTER(_("sh_files_setrec"));
-
   while (ptr != NULL) {
-
     if (ptr->rdepth < (-1) || ptr->rdepth > 99)
       {
 	ptr->rdepth = MaxRecursionLevel;
       }
     if (ptr->rdepth == (-1) && sh.flag.checkSum != SH_CHECK_INIT)
       hash_remove_tree (ptr->name);
-
     ptr = ptr->next;
   }
   SL_RETURN(0, _("sh_files_setrec"));
 }
 
-int sh_files_deldirstack ()
+int sh_files_setrec ()
 {
-  dirstack_t * old_ptr = dirList;
+  sh_files_setrec_int(dirListOne);
+  return sh_files_setrec_int(dirListTwo);
+}
+
+dirstack_t * sh_files_deldirstack_int (dirstack_t * ptr)
+{
+  dirstack_t * old_ptr = ptr;
 
   SL_ENTER(_("sh_files_deldirstack"));
-
   while (old_ptr != NULL) {
-    dirList = old_ptr->next;
+    ptr = old_ptr->next;
     SH_FREE (old_ptr->name);
     SH_FREE (old_ptr);
-    old_ptr = dirList;
+    old_ptr = ptr;
   }
-  SL_RETURN(0, _("sh_files_deldirstack"));
+  SL_RETURN(ptr, _("sh_files_deldirstack"));
+}
+
+int sh_files_deldirstack ()
+{
+  dirListOne = sh_files_deldirstack_int(dirListOne);
+  dirListTwo = sh_files_deldirstack_int(dirListTwo);
+  return 0;
 }
 
 void sh_files_reset()
@@ -320,17 +368,34 @@ void sh_files_reset()
 
 void sh_dirs_reset()
 {
-  dirstack_t * ptr = dirList;
+  dirstack_t * ptr = dirListOne;
 
   SL_ENTER(_("sh_dirs_reset"));
   while (ptr != NULL) {
     ptr->checked = 0;
     ptr = ptr->next;
   }
+  ptr = dirListTwo;
+  while (ptr != NULL) {
+    ptr->checked = 0;
+    ptr = ptr->next;
+  }
   SL_RET0(_("sh_dirs_reset"));
 }
 
 
+int sh_files_pushfile_user0 (char * str_s)
+{
+  return (sh_files_pushfile (SH_LEVEL_USER0, str_s));
+}
+
+
+int sh_files_pushfile_user1 (char * str_s)
+{
+  return (sh_files_pushfile (SH_LEVEL_USER1, str_s));
+}
+
+
 int sh_files_pushfile_ro (char * str_s)
 {
   return (sh_files_pushfile (SH_LEVEL_READONLY, str_s));
@@ -365,15 +430,16 @@ int sh_files_pushfile_allig (char * str_s)
 static void sh_files_set_mask (unsigned long * mask, 
 			       unsigned long val, int act)
 {
-  /* fprintf (stderr, "FIXME IN : %6lo  %6lo  %2d\n", (*mask), val, act); */
+  SL_ENTER(_("sh_files_set_mask"));
+
   if       (act == 0)
     (*mask)  = val;
   else if  (act > 0)
     (*mask) |= val;
   else 
     (*mask) &= ~val;
-  /* fprintf (stderr, "FIXME OUT: %6lo  %6lo  %2d\n", (*mask), val, act); */
-  return;
+
+  SL_RET0(_("sh_files_set_mask"));;
 }
 
 /* set mask(class)
@@ -382,9 +448,13 @@ static int sh_files_parse_mask (unsigned long * mask, char * str)
 {
   int l, i = 0, act = 0, k = 0;
   char  myword[64];
+  
+  SL_ENTER(_("sh_files_parse_mask"));
 
   if (str == NULL)
-    return (-1);
+    {
+      SL_RETURN ( (-1), _("sh_files_parse_mask"));
+    }
   else
     l = sl_strlen(str);
 
@@ -423,6 +493,9 @@ static int sh_files_parse_mask (unsigned long * mask, char * str)
 /* link         */
 	if (0 == strncmp(myword, _("LNK"), 3))
 	  sh_files_set_mask (mask, MODI_LNK, act);
+/* inode        */
+	if (0 == strncmp(myword, _("RDEV"), 3))
+	  sh_files_set_mask (mask, MODI_RDEV, act);
 /* inode        */
 	if (0 == strncmp(myword, _("INO"), 3))
 	  sh_files_set_mask (mask, MODI_INO, act);
@@ -453,9 +526,17 @@ static int sh_files_parse_mask (unsigned long * mask, char * str)
 	
       }
   }
-  return 0;
+  SL_RETURN ( (0), _("sh_files_parse_mask"));
 }
 
+int sh_files_redef_user0(char * str)
+{
+  return (sh_files_parse_mask(&mask_USER0, str));
+} 
+int sh_files_redef_user1(char * str)
+{
+  return (sh_files_parse_mask(&mask_USER1, str));
+} 
 int sh_files_redef_readonly(char * str)
 {
   return (sh_files_parse_mask(&mask_READONLY, str));
@@ -481,22 +562,26 @@ int sh_files_redef_allignore(char * str)
   return (sh_files_parse_mask(&mask_ALLIGNORE, str));
 } 
 
-static unsigned long sh_files_maskof (int class)
+unsigned long sh_files_maskof (int class)
 {
   switch (class)
     {
-    case SH_LEVEL_ALLIGNORE:
-      return (unsigned long) mask_ALLIGNORE;
-    case SH_LEVEL_NOIGNORE:
-      return (unsigned long) mask_NOIGNORE;
+    case SH_LEVEL_READONLY:
+      return (unsigned long) mask_READONLY;
     case SH_LEVEL_ATTRIBUTES:
       return (unsigned long) mask_ATTRIBUTES;
     case SH_LEVEL_LOGFILES:
       return (unsigned long) mask_LOGFILES;
     case SH_LEVEL_LOGGROW:
       return (unsigned long) mask_LOGGROW;
-    case SH_LEVEL_READONLY:
-      return (unsigned long) mask_READONLY;
+    case SH_LEVEL_ALLIGNORE:
+      return (unsigned long) mask_ALLIGNORE;
+    case SH_LEVEL_NOIGNORE:
+      return (unsigned long) mask_NOIGNORE;
+    case SH_LEVEL_USER0:
+      return (unsigned long) mask_USER0;
+    case SH_LEVEL_USER1:
+      return (unsigned long) mask_USER1;
     default:
       return (unsigned long) 0;
     }
@@ -516,6 +601,7 @@ int sh_files_has_metachar (const char * str)
     SL_RETURN(0, _("sh_files_has_metachar"));
 }
 
+
 int sh_files_globerr (const char * epath, int errnum)
 {
   char * p;
@@ -560,7 +646,7 @@ int sh_files_push_file_int (int class, const char * str_s, int len)
 }
 
 
-int sh_files_pushfile (int class, char * str_s)
+static int sh_files_pushfile (int class, char * str_s)
 {
   char  * tmp;
   int     len;
@@ -592,7 +678,7 @@ int sh_files_pushfile (int class, char * str_s)
 
   len = sl_strlen(str_s);
 
-  if (len > PATH_MAX) 
+  if (len >= PATH_MAX) 
     {
       tmp = sh_util_safe_name (str_s);
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_2LONG,
@@ -650,23 +736,23 @@ int sh_files_pushfile (int class, char * str_s)
 	      )
 	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, 
 			     globstatus, MSG_FI_GLOB,
-			     _("No matches found."), tmp);
+			     _("No matches found"), tmp);
 #ifdef GLOB_NOSPACE
 	  else if (globstatus == GLOB_NOSPACE)
 	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__,
 			     globstatus, MSG_FI_GLOB,
-			     _("Out of memory."), tmp);
+			     _("Out of memory"), tmp);
 #endif
 #ifdef GLOB_ABORTED
 	  else if (globstatus == GLOB_ABORTED)
 	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__,
 			     globstatus, MSG_FI_GLOB,
-			     _("Read error."), tmp);
+			     _("Read error"), tmp);
 #endif
 	  else 
 	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__,
 			     globstatus, MSG_FI_GLOB,
-			     _("Unknown error."), tmp);
+			     _("Unknown error"), tmp);
 
 	  SH_FREE(tmp);
 
@@ -685,43 +771,114 @@ int sh_files_pushfile (int class, char * str_s)
 
 /* ------ directories ----- */
 
-unsigned long sh_dirs_chk ()
+int sh_files_is_allignore_int (char * str, dirstack_t * ptr)
+{
+  SL_ENTER(_("sh_files_is_allignore"));
+
+  while (ptr)
+    {
+      if (0 == strcmp(str, ptr->name))
+	{
+	  if (ptr->class == SH_LEVEL_ALLIGNORE)
+	    SL_RETURN( 1, _("sh_files_is_allignore"));
+	  else
+	    SL_RETURN( 0, _("sh_files_is_allignore"));
+	}
+      ptr = ptr->next;
+    }
+  SL_RETURN( 0, _("sh_files_is_allignore"));
+}
+
+int sh_files_is_allignore (char * str)
+{
+  if (1 == sh_files_is_allignore_int(str, dirListOne))
+    return 1;
+  if (NULL == dirListTwo)
+    return 0;
+  return sh_files_is_allignore_int(str, dirListTwo);
+}
+
+unsigned long sh_dirs_chk (int which)
 {
   dirstack_t * ptr;
-  int         status;
+  dirstack_t * dst_ptr;
+  int          status;
   unsigned long dcount = 0;
   char       * tmp;
   
   SL_ENTER(_("sh_dirs_chk"));
 
-  ptr = dirList;
+  if (which == 1)
+    ptr = dirListOne;
+  else
+    ptr = dirListTwo;
 
   while (ptr) 
     {
-      if (sig_terminate == 1)
+      if (sig_urgent == 1) {
 	SL_RETURN(dcount, _("sh_dirs_chk"));
+      }
 
       if (ptr->checked == S_FALSE)
 	{
+	  /* 28 Aug 2001 check the top level directory
+	   */
+	  status        = S_FALSE;
+	  dst_ptr       = fileList;
+	  while (dst_ptr) 
+	    {
+	      if (0 == sl_strcmp(dst_ptr->name,  ptr->name))
+		{
+		  if (dst_ptr->checked == S_FALSE)
+		    {
+		      BREAKEXIT(sh_files_filecheck);
+		      sh_files_filecheck (dst_ptr->class,  ptr->name,  
+					  NULL,  &status);
+		      dst_ptr->checked = S_TRUE;
+		      status           = S_TRUE;
+		      break;
+		    }
+		  else
+		    {
+		      status           = S_TRUE;
+		      break;
+		    }
+		}
+	      dst_ptr = dst_ptr->next;
+	    }
+	  if (status == S_FALSE)
+	    sh_files_filecheck (ptr->class,  ptr->name,  NULL,  &status);
+
+	  BREAKEXIT(sh_files_checkdir);
 	  status = sh_files_checkdir (ptr->class, ptr->rdepth, ptr->name);
+
 	  if (status < 0 && ptr->reported == S_FALSE) 
 	    {
-	      tmp = sh_util_safe_name (ptr->name);
-	      sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
-			       ShDFLevel[ptr->class] : 
-			       ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__,
-			       0, MSG_FI_MISS, tmp);
-	      SH_FREE(tmp);
+	      /* directory is missing
+	       */
+	      if (S_FALSE == sh_ignore_chk_del(ptr->name))
+		{
+		  tmp = sh_util_safe_name (ptr->name);
+		  sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
+				   ShDFLevel[ptr->class] : 
+				   ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__,
+				   0, MSG_FI_MISS, tmp);
+		  SH_FREE(tmp);
+		}
 	      if (sh.flag.reportonce == S_TRUE)
 		ptr->reported = S_TRUE;
 	    } 
 	  else 
 	    {
-	      /* It is there again
+	      /* exists (status >= 0), but was missing (reported == TRUE)
 	       */
 	      if (status >= 0 && ptr->reported == S_TRUE)
 		{
 		  ptr->reported = S_FALSE;
+#if 0
+		  /* obsoleted (really?) by the mandatory sh_files_filecheck()
+		   * above, which will catch missing directories anyway
+		   */
 		  tmp = sh_util_safe_name (ptr->name);
 		  sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
 				   ShDFLevel[ptr->class] : 
@@ -729,9 +886,12 @@ unsigned long sh_dirs_chk ()
 				   FIL__, __LINE__, 0, MSG_FI_ADD,
 				   tmp);
 		  SH_FREE(tmp);
+#endif
 		}
 	      else if (status == SH_FILE_UNKNOWN)
 		{
+		  /* catchall
+		   */
 		  tmp = sh_util_safe_name (ptr->name);
 		  sh_error_handle (SH_ERR_INFO, FIL__, __LINE__, 0,
 				   MSG_FI_FAIL,
@@ -746,14 +906,25 @@ unsigned long sh_dirs_chk ()
 	  ptr->checked = S_TRUE;
 	}
 
-      if (sig_terminate == 1)
+      if (sig_urgent == 1) {
 	SL_RETURN(dcount, _("sh_dirs_chk"));
+      }
 
       ptr = ptr->next;
     }
   SL_RETURN(dcount, _("sh_dirs_chk"));
 }
 
+int sh_files_pushdir_user0 (char * str_s)
+{
+  return (sh_files_pushdir (SH_LEVEL_USER0, str_s));
+}
+
+int sh_files_pushdir_user1 (char * str_s)
+{
+  return (sh_files_pushdir (SH_LEVEL_USER1, str_s));
+}
+
 int sh_files_pushdir_attr (char * str_s)
 {
   return (sh_files_pushdir (SH_LEVEL_ATTRIBUTES, str_s));
@@ -784,6 +955,16 @@ int sh_files_pushdir_allig (char * str_s)
   return (sh_files_pushdir (SH_LEVEL_ALLIGNORE, str_s));
 }
 
+static int which_dirList = 1;
+
+int set_dirList (int which)
+{
+  if (which == 2)
+    which_dirList = 2;
+  else
+    which_dirList = 1;
+  return 0;
+}
 
 int sh_files_push_dir_int (int class, char * tail, int len, int rdepth)
 {
@@ -804,14 +985,21 @@ int sh_files_push_dir_int (int class, char * tail, int len, int rdepth)
   new_item_ptr->checked        = S_FALSE;
   new_item_ptr->reported       = S_FALSE;
   new_item_ptr->childs_checked = S_FALSE;
-  new_item_ptr->next           = dirList;
-
-  dirList = new_item_ptr;
 
+  if (which_dirList == 1)
+    {
+      new_item_ptr->next           = dirListOne;
+      dirListOne                   = new_item_ptr;
+    }
+  else
+    {
+      new_item_ptr->next           = dirListTwo;
+      dirListTwo                   = new_item_ptr;
+    }
   SL_RETURN(0, _("sh_files_push_dir_int"));
 }
 
-int sh_files_pushdir (int class, char * str_s)
+static int sh_files_pushdir (int class, char * str_s)
 {
   char  * tmp;
   int     len;
@@ -850,7 +1038,7 @@ int sh_files_pushdir (int class, char * str_s)
 
   len = sl_strlen(tail);
 
-  if (len > PATH_MAX) 
+  if (len >= PATH_MAX) 
     {
       tmp = sh_util_safe_name (tail);
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_2LONG,
@@ -910,23 +1098,23 @@ int sh_files_pushdir (int class, char * str_s)
 	      )
 	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, 
 			     globstatus, MSG_FI_GLOB,
-			     _("No matches found."), tmp);
+			     _("No matches found"), tmp);
 #ifdef GLOB_NOSPACE
 	  else if (globstatus == GLOB_NOSPACE)
 	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__,
 			     globstatus, MSG_FI_GLOB,
-			     _("Out of memory."), tmp);
+			     _("Out of memory"), tmp);
 #endif
 #ifdef GLOB_ABORTED
 	  else if (globstatus == GLOB_ABORTED)
 	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__,
 			     globstatus, MSG_FI_GLOB,
-			     _("Read error."), tmp);
+			     _("Read error"), tmp);
 #endif
 	  else 
 	    sh_error_handle (SH_ERR_ERR, FIL__, __LINE__,
 			     globstatus, MSG_FI_GLOB,
-			     _("Unknown error."), tmp);
+			     _("Unknown error"), tmp);
 	  SH_FREE(tmp);
 	}
 
@@ -939,9 +1127,59 @@ int sh_files_pushdir (int class, char * str_s)
   SL_RETURN((0), _("sh_files_pushdir"));
 }  
 
+struct sh_dirent {
+  char             * sh_d_name;
+  struct sh_dirent * next;
+};
 
-int sh_files_checkdir (int iclass, int idepth, char * iname)
+static void kill_sh_dirlist (struct sh_dirent * dirlist)
 {
+  struct sh_dirent * this;
+
+  while (dirlist)
+    {
+      this    = dirlist->next;
+      SH_FREE(dirlist->sh_d_name);
+      SH_FREE(dirlist);
+      dirlist = this;
+    }
+  return;
+}
+  
+/* -- add an entry to a directory listing
+ */
+static struct sh_dirent * addto_sh_dirlist (struct dirent * thisEntry, 
+					    struct sh_dirent * dirlist)
+{
+  struct sh_dirent * this;
+  int i;
+
+  if (thisEntry == NULL)
+    return dirlist;
+
+  i = sl_strlen(thisEntry->d_name);
+  if (i == 0)
+    return dirlist;
+  ++i;
+
+  this = SH_ALLOC(sizeof(struct sh_dirent));
+  if (!this)
+    return dirlist;
+
+  this->sh_d_name = SH_ALLOC(i);
+  sl_strlcpy(this->sh_d_name, thisEntry->d_name, i);
+
+  this->next = dirlist;
+  return this;
+}
+
+/* -- check a single directory and its content
+ */
+static int sh_files_checkdir (int iclass, int idepth, char * iname)
+{
+  struct sh_dirent * dirlist = NULL;
+  struct sh_dirent * dirlist_orig = NULL;
+
   DIR *           thisDir = NULL;
   struct dirent * thisEntry;
   int             status;
@@ -958,6 +1196,7 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
   int             class  = 0;
   int             rdepth_next;
   int             class_next;
+  int             file_class_next;
 
   int             checked_flag  = S_FALSE;
   int             cchecked_flag = S_FALSE;
@@ -966,15 +1205,15 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 
   SL_ENTER(_("sh_files_checkdir"));
 
-  if (sig_terminate == 1)
+  if (sig_urgent == 1) {
     SL_RETURN((0), _("sh_files_checkdir"));
+  }
 
   if (iname == NULL || idepth < (-1))
     SL_RETURN((-1), _("sh_files_checkdir"));
   
   if (idepth < 0)
     {
-      /* fprintf(stderr, "FIXME: remove %s\n", iname); */
       /* hash_remove_tree (iname); */
       SL_RETURN((0), _("sh_files_checkdir"));
     }
@@ -986,14 +1225,14 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 
   /* ---- check for obscure name ----
    */
-  sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], iname);
+  sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], iname, S_TRUE);
   
   sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_CHK,
 		   tmpname);
   
   /* ---- check input ----
    */
-  if ( sl_strlen(iname) > PATH_MAX) 
+  if ( sl_strlen(iname) >= PATH_MAX) 
     {
       sh_error_handle (ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__, 0, 
 		       MSG_FI_2LONG,
@@ -1015,10 +1254,17 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
     
   /* ---- stat the directory ----
    */
-  sl_strlcpy (theFile.fullpath, iname, PATH_MAX+1);
+  sl_strlcpy (theFile.fullpath, iname, PATH_MAX);
   status = sh_unix_getinfo (ShDFLevel[SH_ERR_T_DIR], 
 			    iname,
 			    &theFile, NULL);
+
+  if (status == -1)
+    {
+      SH_FREE(tmpname); 
+      SL_RETURN((-1), _("sh_files_checkdir"));
+    }
+
   if (theFile.c_mode[0] != 'd') 
     { 
       sh_error_handle (ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__, 0,
@@ -1056,141 +1302,195 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
   theDir.NumBDev     = 0;
   theDir.NumAll      = 0;
   theDir.TotalBytes  = 0;
-  sl_strlcpy (theDir.DirPath, iname, PATH_MAX+1); 
+  sl_strlcpy (theDir.DirPath, iname, PATH_MAX); 
 
-  if (0 == (rand() % 5))
-    (void) sh_derr();
 
   /* ---- read ----
    */
   do {
-    thisEntry = readdir (thisDir);
-
-    if (thisEntry != NULL) {
-      ++theDir.NumAll;
-
-      if (sl_strcmp (thisEntry->d_name, ".") == 0)
-	{ 
-	  ++theDir.NumDirs;
-	  continue;
-	}
-
-      if (sl_strcmp (thisEntry->d_name, "..") == 0)
+      thisEntry = readdir (thisDir);
+      if (thisEntry != NULL) 
 	{
-	  ++theDir.NumDirs;
-	  continue;
-	}
-
-      /* ---- Check the file. ---- 
-       */
-      tmpcat = SH_ALLOC(PATH_MAX+2);
-      sl_strlcpy(tmpcat, iname,                 PATH_MAX+2);
-      if (sl_strlen(tmpcat) > 1 || tmpcat[0] != '/')
-	sl_strlcat(tmpcat, "/",                   PATH_MAX+2);
-      sl_strlcat(tmpcat, thisEntry->d_name,     PATH_MAX+2);
-
-      dst_ptr       = fileList;
-      rdepth_next   = rdepth - 1;
-      class_next    = class;
-      checked_flag  = -1;
-      cchecked_flag = -1;
-
-      while (dst_ptr) 
-	{
-	  if (0 == sl_strcmp(dst_ptr->name,  tmpcat))
-	    {
-	      class_next    = dst_ptr->class;
-	      checked_flag  = dst_ptr->checked;
-	      cchecked_flag = dst_ptr->childs_checked;
-	      break;
+	  ++theDir.NumAll;
+	  if (sl_strcmp (thisEntry->d_name, ".") == 0)
+	    { 
+	      ++theDir.NumDirs;
+	      continue;
 	    }
-	  dst_ptr = dst_ptr->next;
-	}
-
-      /* ---- Has been checked already. ----
-       */
-      if (checked_flag == S_TRUE && cchecked_flag == S_TRUE)
-	continue;
-	  
-      /* --- May be true, false, or not found. --- 
-       */
-      if (checked_flag == S_TRUE)
-	{
-	  /* -- need only the file type --
-	   */
-	  checkit = sh_unix_get_ftype(tmpcat);
-	}
-      else
-	{
-	  /* -- need to check the file itself --
-	   */
-	  checkit = sh_files_filecheck (class_next, 
-					iname, 
-					thisEntry->d_name,
-					&dummy);
-	  if (checked_flag == S_FALSE)
-	    dst_ptr->checked = S_TRUE;
-	}
-	  
-      if      (checkit == SH_FILE_REGULAR)   
-	++theDir.NumRegular;
-
-      else if (checkit == SH_FILE_DIRECTORY) 
-	{
-	  ++theDir.NumDirs;
-	  if (rdepth_next >= 0 && cchecked_flag != S_TRUE) 
+	  if (sl_strcmp (thisEntry->d_name, "..") == 0)
 	    {
-	      rdepth_next = rdepth - 1;
-
-	      /* check whether the new directory is in the
-	       * list with a recursion depth already defined
-	       */
-	      checked_flag  = -1;
-	      cchecked_flag = -1;
-
-	      dst_ptr     = dirList;
-	      while (dst_ptr) 
-		{
-		  if (0 == sl_strcmp(dst_ptr->name, tmpcat))
-		    {
-		      TPT((0, FIL__, __LINE__, 
-			   _("msg=<%s -> recursion depth %d\n>"),
-			   dst_ptr->name, dst_ptr->rdepth));
-		      rdepth_next   = dst_ptr->rdepth;
-		      class_next    = dst_ptr->class;
-		      checked_flag  = dst_ptr->childs_checked;
-		      cchecked_flag = dst_ptr->checked;
-		      break;
-		    }
-		  dst_ptr = dst_ptr->next;
-		}
-
-	      if (cchecked_flag == S_FALSE)
-		{
-		  sh_files_checkdir (class_next, rdepth_next, tmpcat);
-		  dst_ptr->childs_checked = S_TRUE;
-		}
-	      else if (checked_flag == -1)
-		sh_files_checkdir (class_next, rdepth_next, tmpcat);
-		    
+	      ++theDir.NumDirs;
+	      continue;
 	    }
+	  dirlist = addto_sh_dirlist (thisEntry, dirlist);
 	}
-
-      else if (checkit == SH_FILE_SYMLINK)   ++theDir.NumSymlinks;
-      else if (checkit == SH_FILE_FIFO)      ++theDir.NumFifos;
-      else if (checkit == SH_FILE_SOCKET)    ++theDir.NumSockets;
-      else if (checkit == SH_FILE_CDEV)      ++theDir.NumCDev;
-      else if (checkit == SH_FILE_BDEV)      ++theDir.NumBDev;
-
-      SH_FREE(tmpcat);
-
-    }
-
-    if (sig_terminate == 1)
-      SL_RETURN((0), _("sh_files_checkdir"));
-
   } while (thisEntry != NULL);
 
+  closedir (thisDir);
+
+  dirlist_orig = dirlist;
+
+  do {
+
+    /* If the directory is empty, dirlist = NULL
+     */
+    if (!dirlist)
+      break;
+
+    if (sig_termfast == 1) {
+      SL_RETURN((0), _("sh_files_checkdir"));
+    }
+
+    BREAKEXIT(sh_derr);
+    if (0 == (rand() % 5))
+      (void) sh_derr();
+    
+    /* ---- Check the file. ---- 
+     */
+    tmpcat = SH_ALLOC(PATH_MAX);
+    sl_strlcpy(tmpcat, iname,                   PATH_MAX);
+    if (sl_strlen(tmpcat) > 1 || tmpcat[0] != '/')
+      sl_strlcat(tmpcat, "/",                   PATH_MAX);
+    sl_strlcat(tmpcat, dirlist->sh_d_name,      PATH_MAX);
+    
+    dst_ptr         = fileList;
+    rdepth_next     = rdepth - 1;
+    class_next      = class;
+    file_class_next = class;
+    checked_flag    = -1;
+    cchecked_flag   = -1;
+    
+    while (dst_ptr) 
+      {
+	if (0 == sl_strcmp(dst_ptr->name,  tmpcat))
+	  {
+	    /* Tue Aug  6 22:13:27 CEST 2002 introduce file_class_next
+	     * this fixes the problem that a policy for the directory
+	     * inode erroneously becomes a policy for the directory itself.
+	     */
+	    file_class_next    = dst_ptr->class;
+	    /* class_next         = dst_ptr->class; */
+	    checked_flag       = dst_ptr->checked;
+	    cchecked_flag      = dst_ptr->childs_checked;
+	    break;
+	  }
+	dst_ptr = dst_ptr->next;
+      }
+    
+    /* ---- Has been checked already. ----
+     */
+    if (checked_flag == S_TRUE && cchecked_flag == S_TRUE)
+      continue;
+    
+    /* --- May be true, false, or not found. --- 
+     */
+    if (checked_flag == S_TRUE)
+      {
+	/* -- need only the file type --
+	 */
+	checkit = sh_unix_get_ftype(tmpcat);
+      }
+    else
+      {
+	/* -- need to check the file itself --
+	 */
+	if (dst_ptr && sh.flag.reportonce == S_TRUE)
+	  dummy = dst_ptr->reported;
+	checkit = sh_files_filecheck (file_class_next, 
+				      iname, 
+				      dirlist->sh_d_name,
+				      &dummy);
+	if (dst_ptr && checked_flag == S_FALSE)
+	  dst_ptr->checked = S_TRUE;
+	/* Thu Mar  7 15:09:40 CET 2002 Propagate the 'reported' flag
+	 */
+	if (dst_ptr && sh.flag.reportonce == S_TRUE)
+	  dst_ptr->reported = dummy;
+      }
+    
+    if      (checkit == SH_FILE_REGULAR)   
+      ++theDir.NumRegular;
+    
+    else if (checkit == SH_FILE_DIRECTORY) 
+      {
+	++theDir.NumDirs;
+	if (rdepth_next >= 0 && cchecked_flag != S_TRUE) 
+	  {
+	    rdepth_next = rdepth - 1;
+	    
+	    /* check whether the new directory is in the
+	     * list with a recursion depth already defined
+	     */
+	    checked_flag  = -1;
+	    cchecked_flag = -1;
+	    
+	    dst_ptr     = dirListOne;
+	    while (dst_ptr) 
+	      {
+		if (0 == sl_strcmp(dst_ptr->name, tmpcat))
+		  {
+		    TPT((0, FIL__, __LINE__, 
+			 _("msg=<%s -> recursion depth %d\n>"),
+			 dst_ptr->name, dst_ptr->rdepth));
+		    rdepth_next   = dst_ptr->rdepth;
+		    class_next    = dst_ptr->class;
+		    /* 28. Aug 2001 reversed
+		     */
+		    cchecked_flag = dst_ptr->childs_checked;
+		    checked_flag  = dst_ptr->checked;
+		    break;
+		  }
+		dst_ptr = dst_ptr->next;
+	      }
+	    
+	    if (checked_flag == -1)
+	      {
+		dst_ptr     = dirListTwo;
+		while (dst_ptr) 
+		  {
+		    if (0 == sl_strcmp(dst_ptr->name, tmpcat))
+		      {
+			TPT((0, FIL__, __LINE__, 
+			     _("msg=<%s -> recursion depth %d\n>"),
+			     dst_ptr->name, dst_ptr->rdepth));
+			rdepth_next   = dst_ptr->rdepth;
+			class_next    = dst_ptr->class;
+			/* 28. Aug 2001 reversed
+			 */
+			cchecked_flag = dst_ptr->childs_checked;
+			checked_flag  = dst_ptr->checked;
+			break;
+		      }
+		    dst_ptr = dst_ptr->next;
+		  }
+	      }
+	    
+	    if (cchecked_flag == S_FALSE)
+	      {
+		sh_files_checkdir (class_next, rdepth_next, tmpcat);
+		dst_ptr->childs_checked = S_TRUE;
+	      }
+	    else if (checked_flag == -1)
+	      sh_files_checkdir (class_next, rdepth_next, tmpcat);
+	    
+	  }
+      }
+    
+    else if (checkit == SH_FILE_SYMLINK)   ++theDir.NumSymlinks;
+    else if (checkit == SH_FILE_FIFO)      ++theDir.NumFifos;
+    else if (checkit == SH_FILE_SOCKET)    ++theDir.NumSockets;
+    else if (checkit == SH_FILE_CDEV)      ++theDir.NumCDev;
+    else if (checkit == SH_FILE_BDEV)      ++theDir.NumBDev;
+    
+    SH_FREE(tmpcat);
+    
+    if ((sig_termfast == 1) || (sig_terminate == 1)) {
+      SL_RETURN((0), _("sh_files_checkdir"));
+    }
+    
+    dirlist = dirlist->next;
+    
+  } while (dirlist != NULL);
 
   sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_DSUM,
 		   theDir.NumDirs,
@@ -1200,9 +1500,9 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 		   theDir.NumSockets,
 		   theDir.NumCDev,
 		   theDir.NumBDev);
-
   
-  closedir (thisDir);
+  kill_sh_dirlist (dirlist_orig);
+
   SH_FREE(tmpname);
   SL_RETURN((0), _("sh_files_checkdir"));
 }
@@ -1210,10 +1510,13 @@ int sh_files_checkdir (int iclass, int idepth, char * iname)
 int get_the_fd (SL_TICKET ticket);
 
 
-ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
-			       int * reported)
+static ShFileType sh_files_filecheck (int class, char * dirName, 
+				      char * fileName,
+				      int * reported)
 {
-  char            fullpath[PATH_MAX+2];
+  /* 28 Aug 2001 allow NULL fileName
+   */
+  char            fullpath[PATH_MAX];
   char            fileHash[KEY_LEN + 1];
   int             status;
   file_type       theFile;
@@ -1223,18 +1526,20 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
 
   SL_ENTER(_("sh_files_filecheck"));
 
+  BREAKEXIT(sh_derr);
   if (0 == (rand() % 2))
     (void) sh_derr();
 
-  if (dirName == NULL || fileName == NULL)
+  /* fileName may be NULL if this is a directory
+   */
+  if (dirName == NULL /* || fileName == NULL */)
     {
-     
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_NULL);
       SL_RETURN(SH_FILE_UNKNOWN, _("sh_files_filecheck"));
     }
 
-  if (0 != sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], 
-				  fileName)) 
+  if (fileName != NULL && 0 != sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], 
+						    fileName, S_FALSE)) 
     {
       tmpdir  = sh_util_safe_name (dirName);
       tmpname = sh_util_safe_name (fileName);
@@ -1246,6 +1551,8 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
       SH_FREE(tmpdir);
     }    
 
+  /* sh_files_fullpath accepts NULL fileName
+   */
   if (0 != sh_files_fullpath (dirName, fileName, fullpath)) 
     { 
       tmpdir  = sh_util_safe_name (dirName);
@@ -1261,10 +1568,12 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
 
   /* stat the file and determine checksum (if a regular file)
    */
-  sl_strlcpy (theFile.fullpath, fullpath, PATH_MAX+1);
+  sl_strlcpy (theFile.fullpath, fullpath, PATH_MAX);
   theFile.check_mask = sh_files_maskof(class);
   theFile.reported   = (*reported);
 
+  TPT(( 0, FIL__, __LINE__, _("msg=<checking file: %s>\n"),  fullpath));
+
   status = sh_unix_getinfo ( (class == SH_LEVEL_ALLIGNORE) ? 
 			     ShDFLevel[class] : ShDFLevel[SH_ERR_T_FILE], 
 			     fileName,
@@ -1279,12 +1588,18 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
       SL_RETURN(SH_FILE_UNKNOWN, _("sh_files_filecheck"));
     }
   
+  if (sig_termfast == 1) {
+    goto ret_point;
+  }
+
   /* report
    */
   if (theFile.c_mode[0] == '-')
     {
+      tmpname = sh_util_safe_name (fullpath); /* fixed in 1.5.4 */
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_CSUM,
-		       fileHash, fullpath);
+		       fileHash, tmpname);
+      SH_FREE(tmpname);
     } 
       
   if      ( sh.flag.checkSum == SH_CHECK_INIT && sh.flag.update == S_FALSE ) 
@@ -1293,8 +1608,8 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
     }
   else if (sh.flag.checkSum == SH_CHECK_INIT && sh.flag.update == S_TRUE )
     {
-      sh_hash_compdata (class, &theFile, fileHash);
-      sh_hash_pushdata (&theFile, fileHash);
+      if (0 == sh_hash_compdata (class, &theFile, fileHash))
+	sh_hash_pushdata (&theFile, fileHash);
     }
   else if (sh.flag.checkSum == SH_CHECK_CHECK 
 	   /* && theFile.c_mode[0] == '-' */
@@ -1308,7 +1623,7 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
 
   /* reset the access time 
    */
-  if (class == SH_LEVEL_NOIGNORE)
+  if (class == SH_LEVEL_NOIGNORE && (theFile.check_mask & MODI_ATM) != 0)
     {
       utime_buf.actime   = (time_t) theFile.atime;
       utime_buf.modtime  = (time_t) theFile.mtime;
@@ -1319,6 +1634,8 @@ ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
 #endif
     }
   
+ ret_point:
+
   /* yes I know, the 'break' is not needed
    */
   switch (theFile.c_mode[0]) 
@@ -1348,15 +1665,15 @@ static int sh_files_fullpath (char * testdir, char * d_name, char * statpath)
     {
       if ( (llen = sl_strlen(testdir)) > (PATH_MAX-2) ) 
 	SL_RETURN((-1),_("sh_files_fullpath"));
-      sl_strlcpy(statpath, testdir,    PATH_MAX - 2);
+      sl_strlcpy(statpath, testdir,    PATH_MAX - 1);
     }
   if (d_name != NULL) 
     {
       if (llen > 1 || statpath[0] != '/')
-	sl_strlcat(statpath, "/",   PATH_MAX+1);
-      if (sl_strlen(d_name) >= ( PATH_MAX - sl_strlen(statpath) ) ) 
+	sl_strlcat(statpath, "/",   PATH_MAX);
+      if ((sl_strlen(d_name) + sl_strlen(statpath)) >= PATH_MAX)
 	SL_RETURN((-1),_("sh_files_fullpath"));
-      sl_strlcat(statpath, d_name,   PATH_MAX+1);
+      sl_strlcat(statpath, d_name,   PATH_MAX);
     }
   if (statpath == NULL) 
     SL_RETURN((-1),_("sh_files_fullpath"));
@@ -1393,11 +1710,11 @@ static int check_file(char * name)
   SL_RETURN((-1), _("check_file"));
 }
   
-int sh_files_test_setup ()
+int sh_files_test_setup_int (dirstack_t * pdirL)
 {
   int dlen, flen;
 
-  dirstack_t * pdirL = dirList;  
+  /* dirstack_t * pdirL = dirList; */ 
   dirstack_t * pfilL = fileList;
 
   SL_ENTER(_("sh_files_test_setup"));
@@ -1444,6 +1761,52 @@ int sh_files_test_setup ()
   SL_RETURN((0), _("sh_files_test_setup"));
 }
       
+int sh_files_test_double (dirstack_t * firstList, dirstack_t * secondList)
+{
+  int          count;
 
+  dirstack_t * saveList = firstList;
+  dirstack_t * tempList;
+
+  while (firstList)
+    {
+      tempList = saveList;
+      count    = 0;
+
+      while (tempList)
+	{
+	  if (0 == sl_strcmp(firstList->name, tempList->name))
+	    ++count;
+	  tempList = tempList->next;
+	}
+      tempList = secondList;
+      while (tempList)
+	{
+	  if (0 == sl_strcmp(firstList->name, tempList->name))
+	    ++count;
+	  tempList = tempList->next;
+	}
+      if (count > 1)
+	sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_DOUBLE,
+			 firstList->name);
+      firstList = firstList->next;
+    }
+  return 0;
+}
+      
+int sh_files_test_setup ()
+{
+  /* Test for modifications allowed in ReadOnly directory
+   */  
+  sh_files_test_setup_int (dirListOne);
+  sh_files_test_setup_int (dirListTwo);
+
+  /* Test for files/dirz defined twice
+   */  
+  sh_files_test_double (dirListOne, dirListTwo);
+  sh_files_test_double (dirListTwo, dirListOne);
+  sh_files_test_double (fileList,   NULL);
+  return 0;
+}
 
 #endif
