@@ -28,6 +28,9 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <signal.h>
+#if defined(SH_WITH_SERVER)
+#include <pwd.h>
+#endif
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <errno.h>
@@ -53,6 +56,8 @@
 #include "sh_error.h"
 #include "sh_tiger.h"
 
+extern struct passwd * sh_getpwnam(const char *name);
+
 static struct {
   char     conf_id[SH_MINIBUF+1];
   char     conf_fp[SH_MINIBUF+1];
@@ -69,6 +74,40 @@ typedef struct {
 #define SH_GPG_BAD     1
 #define SH_GPG_BADSIGN 2
 
+/* replace #if 0 by #if 1 and set an appropriate path in front of '/pdbg.'
+ * for debugging
+ */
+#if 0
+#define PDGBFILE "/pdbg."
+#endif
+
+#if defined(PDGBFILE)
+FILE * pdbg;
+FILE * pdbgc;
+#define PDBG_OPEN    pdbg = fopen(PDGBFILE"main",  "a")  
+#define PDBG_CLOSE   fclose (pdbg)
+#define PDBG(arg)    fprintf(pdbg,  "PDBG: step %d\n", arg); fflush(pdbg)
+#define PDBG_D(arg)  fprintf(pdbg,  "PDBG: %d\n", arg); fflush(pdbg)
+#define PDBG_S(arg)  fprintf(pdbg,  "PDBG: %s\n", arg); fflush(pdbg)
+
+#define PDBGC_OPEN   pdbgc = fopen(PDGBFILE"child", "a")  
+#define PDBGC_CLOSE  fclose (pdbgc)
+#define PDBGC(arg)   fprintf(pdbgc, "PDBG: step %d\n", arg); fflush(pdbgc)
+#define PDBGC_D(arg) fprintf(pdbgc, "PDBG: %d\n", arg); fflush(pdbgc)
+#define PDBGC_S(arg) fprintf(pdbgc, "PDBG: %s\n", arg); fflush(pdbgc)
+#else
+#define PDBG_OPEN    
+#define PDBG_CLOSE   
+#define PDBG(arg)    
+#define PDBG_D(arg)  
+#define PDBG_S(arg)  
+#define PDBGC_OPEN    
+#define PDBGC_CLOSE   
+#define PDBGC(arg)    
+#define PDBGC_D(arg)  
+#define PDBGC_S(arg)  
+#endif
+
 #undef  FIL__
 #define FIL__  _("sh_gpg.c")
 
@@ -78,6 +117,11 @@ static int sh_gpg_checksum (SL_TICKET checkfd, int flag)
   char * test_gpg;
   char * test_ptr1 = NULL;
   char * test_ptr2 = NULL;
+  char   wstrip1[128];
+  char   wstrip2[128];
+  int    i, k;
+
+  SL_ENTER(_("sh_gpg_checksum"));
 
   tiger_fd = checkfd;
 #if defined(WITH_PGP)
@@ -98,24 +142,54 @@ static int sh_gpg_checksum (SL_TICKET checkfd, int flag)
     test_ptr1 += 2;
   else
     test_ptr1 = GPG_HASH;
-  
-  if (0 != sl_strncmp(test_ptr1, test_ptr2, KEY_LEN + 6))
+
+  /* Tue Jun 24 23:11:54 CEST 2003 (1.7.9) -- strip whitespace
+   */
+  k = 0;
+  for (i = 0; i < 127; ++i)
+    {
+      if (test_ptr1[i] == '\0')
+	break;
+      if (test_ptr1[i] != ' ')
+	{
+	  wstrip1[k] = test_ptr1[i];
+	  ++k;
+	}
+    }
+  wstrip1[k] = '\0';
+
+  k = 0;
+  for (i = 0; i < 127; ++i)
+    {
+      if (test_ptr2[i] == '\0')
+	break;
+      if (test_ptr2[i] != ' ')
+	{
+	  wstrip2[k] = test_ptr2[i];
+	  ++k;
+	}
+    }
+  wstrip2[k] = '\0';
+
+  if (0 != sl_strncmp(wstrip1, wstrip2, 127))
     {
       TPT(((0), FIL__, __LINE__, _("msg=<pgp checksum: %s>\n"), test_gpg));
       TPT(((0), FIL__, __LINE__, _("msg=<Compiled-in : %s>\n"), GPG_HASH));
+      TPT(((0), FIL__, __LINE__, _("msg=<wstrip1     : %s>\n"), wstrip1));
+      TPT(((0), FIL__, __LINE__, _("msg=<wstrip2     : %s>\n"), wstrip2));
       if (flag == 1)
 	sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_GPG, 
 			GPG_HASH, test_gpg);
       SH_FREE(test_gpg);
-      return(-1);
+      SL_RETURN((-1), _("sh_gpg_checksum"));
     }
   SH_FREE(test_gpg);
-  return (0);
+  SL_RETURN( (0), _("sh_gpg_checksum"));
 }
 #endif
 
 static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd, 
-			    int mode, char * id)
+			    int mode, char * id, char * homedir)
 {
   int pipedes[2];
   FILE * outf = NULL;
@@ -129,9 +203,12 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 #endif
 #if defined(WITH_GPG)
   char   cc0[2] = "-";
+  char   cc3[32];
+  char   cc4[SH_PATHBUF+32];
+  char   cc5[32] = "--no-tty";
 #endif
 
-  char * arg[6];
+  char * arg[9];
 
 #if defined(HAVE_GPG_CHECKSUM)
   SL_TICKET   checkfd;
@@ -144,20 +221,104 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 #endif
 #endif
 
+  SL_ENTER(_("sh_gpg_popen"));
+
 #if defined(WITH_GPG)
   /* -- GnuPG -- */
-  sl_strlcpy (path,  DEFAULT_GPG_PATH, 256);
+  sl_strlcpy (path,  DEFAULT_GPG_PATH,  256);
   sl_strlcpy (cc1,   _("--status-fd"),  32);
   sl_strlcpy (cc2,   _("--verify"),     32);
+  sl_strlcpy (cc3,   _("--homedir"),    32);
+  /* sl_strlcpy (cc4,   sh.effective.home, SH_PATHBUF+32); */
+  sl_strlcpy (cc4,   homedir,           SH_PATHBUF+32);
+  sl_strlcat (cc4,   _("/.gnupg"),      SH_PATHBUF+32); 
 
+  /* fprintf(stderr, "YULE: homedir=%s\n", homedir); */
+
+#if defined(SH_WITH_SERVER)
+  if (0 == sl_ret_euid())   /* privileges not dropped yet */
+    {
+      struct stat lbuf;
+      int         status_stat = 0;
+      struct passwd * tempres = getpwnam(DEFAULT_IDENT);
+      if (!tempres)
+	tempres = sh_getpwnam(DEFAULT_IDENT);
+      if (!tempres)
+	{
+	  dlog(1, FIL__, __LINE__, 
+	       _("User %s does not exist. Please add the user to your system.\n"), 
+	       DEFAULT_IDENT);
+	  status_stat = -1;
+	}
+      if (!tempres->pw_dir || tempres->pw_dir[0] == '\0')
+	{
+	  dlog(1, FIL__, __LINE__, 
+	       _("User %s does not have a home directory.\nPlease add the home directory for this user to your system.\n"), 
+	       DEFAULT_IDENT);
+	  status_stat = -2;
+	}
+      if (status_stat == 0)
+	{
+	  sl_strlcpy (cc4, tempres->pw_dir, SH_PATHBUF+32); 
+	  sl_strlcat (cc4,   _("/.gnupg"),      SH_PATHBUF+32); 
+	  status_stat =  retry_lstat(FIL__, __LINE__, cc4, &lbuf);
+	  if (status_stat == -1)
+	    {
+	      dlog(1, FIL__, __LINE__, 
+		   _("Gnupg directory %s for user %s\ndoes not exist or is not accessible.\nPlease add the directory and put the keyring (pubring.gpg) there\nto verify the configuration file.\n"),
+		   cc4, DEFAULT_IDENT);
+	      status_stat = -3;
+	    }
+	}
+      if (status_stat == 0 && lbuf.st_uid != tempres->pw_uid)
+	{
+	  dlog(1, FIL__, __LINE__, 
+	       _("Gnupg directory %s\nis not owned by user %s.\n"), 
+	       cc4, DEFAULT_IDENT);
+	  status_stat = -4;
+	}
+      if (status_stat == 0)
+	{
+	  sl_strlcat (cc4,   _("/pubring.gpg"),      SH_PATHBUF+32); 
+	  status_stat =  retry_lstat(FIL__, __LINE__, cc4, &lbuf);
+	  if (status_stat == -1)
+	    {
+	      dlog(1, FIL__, __LINE__, 
+		   _("Gnupg public keyring %s for user %s\ndoes not exist or is not accessible.\nPlease add the directory and put the keyring (pubring.gpg) there\nto verify the configuration file.\n"),
+		   cc4, DEFAULT_IDENT);
+	      status_stat = -5;
+	    }
+	}
+      if (status_stat == 0 && lbuf.st_uid != tempres->pw_uid)
+	{
+	  dlog(1, FIL__, __LINE__, 
+	       _("Gnupg public keyring %s\nis not owned by user %s.\n"), 
+	       cc4, DEFAULT_IDENT);
+	  status_stat = -6;
+	}
+      if (status_stat != 0)
+	{
+	  sh_error_handle((-1), FIL__, __LINE__, status_stat, MSG_EXIT_ABORT1, 
+			  sh.prg_name);
+	  aud_exit (FIL__, __LINE__, EXIT_FAILURE);
+	}
+      sl_strlcpy (cc4, tempres->pw_dir, SH_PATHBUF+32); 
+      sl_strlcat (cc4,   _("/.gnupg"),      SH_PATHBUF+32); 
+    }
+#endif
 
   arg[0] = path; 
   arg[1] = cc1;
   arg[2] = "1";
   arg[3] = cc2;
-  arg[4] = cc0;
-  arg[5] = NULL;
+  arg[4] = cc3;
+  arg[5] = cc4;
+  arg[6] = cc5;
+  arg[7] = cc0;
+  arg[8] = NULL;
 
+  /* catch 'unused parameter' compiler warning
+   */
   (void) mode;
   (void) id;
 #elif defined(WITH_PGP)
@@ -194,9 +355,10 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
    */
   if (sh.effective.home != NULL)
     {
-      envp[0] = my_malloc (sl_strlen(sh.effective.home) + 6);
-      sprintf (envp[0], "HOME=%s",                      /* known to fit  */
-		    sh.effective.home); 
+      envp[0] = malloc (sl_strlen(sh.effective.home) + 6);
+      if (envp[0] != NULL)
+	sprintf (envp[0], "HOME=%s",                     /* known to fit  */
+		 sh.effective.home); 
       envp[1] = NULL;
     }
   else
@@ -209,8 +371,8 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
   if (aud_pipe(FIL__, __LINE__, pipedes) < 0) 
     {
       if (envp[0] != NULL) 
-	my_free(envp[0]);
-      return (NULL);
+	free(envp[0]);
+      SL_RETURN( (NULL), _("sh_gpg_popen"));
     }
   
   source->pid = aud_fork(FIL__, __LINE__);
@@ -222,8 +384,8 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
       close(pipedes[0]);
       close(pipedes[1]);
       if (envp[0] != NULL) 
-	my_free(envp[0]);
-      return (NULL);
+	free(envp[0]);
+      SL_RETURN( (NULL), _("sh_gpg_popen"));
     }
 
   if (source->pid == (pid_t) 0) 
@@ -277,6 +439,20 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
       sl_set_suid(); 
       aud_setuid(FIL__, __LINE__, geteuid());
       
+      PDBGC_OPEN;
+      PDBGC_D((int)getuid());
+      PDBGC_D((int)geteuid());
+
+      {
+	int i = 0;
+	while (arg[i] != NULL)
+	  {
+	    PDBGC_S(arg[i]);
+	    ++i;
+	  }
+      }
+      PDBGC_CLOSE;
+
       /* exec the program */
 
 #if defined(__linux__) && defined(HAVE_GPG_CHECKSUM)
@@ -293,7 +469,7 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
 	aud__exit(FIL__, __LINE__, EXIT_FAILURE);
 
       pfd = get_the_fd(checkfd);
-      sprintf(pname, _("/proc/self/fd/%d"),       /*  known to fit */
+      sprintf(pname, _("/proc/self/fd/%d"),             /* known to fit  */
 		   pfd);
       if (0 == access(pname, R_OK|X_OK))
 	{
@@ -350,11 +526,11 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
    */
 
   if (envp[0] != NULL) 
-    my_free(envp[0]);
+    free(envp[0]);
 
   close (pipedes[STDOUT_FILENO]);
-  retry_fcntl (pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
-  retry_fcntl (pipedes[STDIN_FILENO], F_SETFL,  O_NONBLOCK);
+  retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);
+  retry_fcntl (FIL__, __LINE__, pipedes[STDIN_FILENO], F_SETFL,  O_NONBLOCK);
 
   MBLK( outf = fdopen (pipedes[STDIN_FILENO], "r"); )
   
@@ -364,10 +540,10 @@ static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, int fd,
       close (pipedes[STDOUT_FILENO]);
       waitpid (source->pid, NULL, 0);
       source->pid = 0;
-      return (NULL);
+      SL_RETURN( (NULL), _("sh_gpg_popen"));
     }
   
-  return (outf);
+  SL_RETURN( (outf), _("sh_gpg_popen"));
 }
 
 
@@ -375,20 +551,23 @@ static int sh_gpg_pclose (sh_gpg_popen_t *source)
 {
   int status = 0;
   
+  SL_ENTER(_("sh_gpg_pclose"));
+
   MBLK( status = fclose(source->pipe); )
   if (status)
-    return (-1);
+    SL_RETURN( (-1), _("sh_gpg_pclose"));
   
   if (waitpid(source->pid, NULL, 0) != source->pid)
     status = -1;
   
   source->pipe = NULL;
   source->pid = 0;
-  return (status);
+  SL_RETURN( (status), _("sh_gpg_pclose"));
 }
  
 static
-int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
+int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp, 
+			   char * homedir)
 {
   struct stat buf;
   char line[256];
@@ -402,6 +581,8 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
   SL_TICKET checkfd;
 #endif
 
+  SL_ENTER(_("sh_gpg_check_file_sign"));
+
   /* check whether GnuPG exists and has the correct checksum
    */
 #if defined(WITH_GPG)
@@ -409,15 +590,15 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
   TPT(((0), FIL__, __LINE__, _("msg=<Check signature>\n")));
   TPT(((0), FIL__, __LINE__, _("msg=<gpg is %s>\n"), DEFAULT_GPG_PATH));
 
-  if (0 != retry_lstat(DEFAULT_GPG_PATH, &buf))
+  if (0 != retry_lstat(FIL__, __LINE__, DEFAULT_GPG_PATH, &buf))
     {
       status = errno;
-      sh_error_handle((-1), FIL__, __LINE__, status, MSG_E_LSTAT,
+      sh_error_handle(SH_ERR_ERR, FIL__, __LINE__, status, MSG_ERR_LSTAT,
 		      sh_error_message(status), DEFAULT_GPG_PATH);
-      return SH_GPG_BAD;
+      SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
     }
   if (0 != tf_trust_check (DEFAULT_GPG_PATH, SL_YESPRIV))
-    return SH_GPG_BAD;
+    SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
 
 #ifdef HAVE_GPG_CHECKSUM
   checkfd = sl_open_read(DEFAULT_GPG_PATH, SL_YESPRIV);
@@ -427,8 +608,10 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
       sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN, 
 		      _("Checksum mismatch"), 
 		      _("gpg_check_file_sign"));
-      return SH_GPG_BAD;
+      sl_close(checkfd);
+      SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
     }
+  sl_close(checkfd);
 #endif
 
 #elif defined(WITH_PGP)
@@ -436,15 +619,15 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
   TPT(((0), FIL__, __LINE__, _("msg=<Check signature>\n")));
   TPT(((0), FIL__, __LINE__, _("msg=<pgp is %s>\n"), DEFAULT_PGP_PATH));
 
-  if (0 != retry_lstat(DEFAULT_PGP_PATH, &buf))
+  if (0 != retry_lstat(FIL__, __LINE__, DEFAULT_PGP_PATH, &buf))
     {
       status = errno;
-      sh_error_handle((-1), FIL__, __LINE__, status, MSG_E_LSTAT,
+      sh_error_handle(SH_ERR_ERR, FIL__, __LINE__, status, MSG_ERR_LSTAT,
 		      sh_error_message(status), DEFAULT_PGP_PATH);
-      return SH_GPG_BAD;
+      SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
     }
   if (0 != tf_trust_check (DEFAULT_PGP_PATH, SL_YESPRIV))
-    return SH_GPG_BAD;
+    SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
 
 #ifdef HAVE_GPG_CHECKSUM
   checkfd = sl_open_read(DEFAULT_PGP_PATH, SL_YESPRIV);
@@ -454,8 +637,10 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
       sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN, 
 		      _("Checksum mismatch"), 
 		      _("gpg_check_file_sign"));
-      return SH_GPG_BAD;
+      sl_close(checkfd);
+      SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
     }
+  sl_close(checkfd);
 #endif
 
 #endif
@@ -464,14 +649,14 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
 
   fflush(NULL);
  
-  source.pipe   = sh_gpg_popen  ( &source, fd, 0, NULL );
+  source.pipe   = sh_gpg_popen  ( &source, fd, 0, NULL, homedir );
 
   if (NULL == source.pipe)
     {
       sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN, 
 		      _("Could not open pipe"), 
 		      _("gpg_check_file_sign"));
-      return SH_GPG_BAD;
+      SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
     }
 
   TPT(((0), FIL__, __LINE__, _("msg=<Open pipe success>\n")));
@@ -483,6 +668,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
   while (NULL != fgets(line, 255, source.pipe))
     {
 
+      TPT(((0), FIL__, __LINE__, _("msg=<gpg out: %s>\n"), line));
       if (line[strlen(line)-1] == '\n')
 	line[strlen(line)-1] = ' ';
       sh_error_handle(SH_ERR_ALL, FIL__, __LINE__, 0, MSG_E_SUBGEN, 
@@ -549,13 +735,13 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
 #ifdef WITH_PGP
   /* get the fingerprint */
 
-  source.pipe   = sh_gpg_popen  ( &source, fd, 1,  sign_id);
+  source.pipe   = sh_gpg_popen  ( &source, fd, 1,  sign_id, homedir);
   if (NULL == source.pipe)
     {
       sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN, 
 		      _("Could not open pipe for fp"), 
 		      _("gpg_check_file_sign"));
-      return SH_GPG_BAD;
+      SL_RETURN( SH_GPG_BAD, _("sh_gpg_check_file_sign"));
     }
 
   TPT(((0), FIL__, __LINE__, _("msg=<Open pipe success>\n")));
@@ -616,7 +802,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
     }
 
   if (have_id == GOOD && have_fp == GOOD)
-    return SH_GPG_OK;
+    SL_RETURN( SH_GPG_OK, _("sh_gpg_check_file_sign"));
   else
     {
       if (have_id == BAD)
@@ -627,7 +813,7 @@ int sh_gpg_check_file_sign(int fd, char * sign_id, char * sign_fp)
 	sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN, 
 			_("No fingerprint for key"), 
 			_("gpg_check_file_sign"));
-      return SH_GPG_BADSIGN;
+      SL_RETURN( SH_GPG_BADSIGN, _("sh_gpg_check_file_sign"));
     }
 }
 
@@ -635,11 +821,21 @@ int get_the_fd(SL_TICKET file_1);
 
 int sh_gpg_check_sign (long file_1, long file_2, int what)
 {
-  int status = SH_GPG_BAD, fd1 = 0, fd2 = 0;
+  int status = SH_GPG_BAD;
+  int fd1 = 0;
+  int fd2 = 0;
   static int smsg = S_FALSE;
+  char  * tmp;
+  char  * tmp2;
+
+  char  * homedir = sh.effective.home;
+#if defined(SH_WITH_SERVER)
+  struct passwd * tempres;
+#endif
 
   SL_ENTER(_("sh_gpg_check_sign"));
 
+
   if (what == 0 || what == 1)
     fd1 = get_the_fd(file_1);
   if (what == 0 || what == 2)
@@ -649,15 +845,27 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
     {
       TPT(((0), FIL__, __LINE__, _("msg=<GPG_CHECK: FD1 = %d>\n"), fd1));
       TPT(((0), FIL__, __LINE__, _("msg=<GPG_CHECK: FD2 = %d>\n"), fd2));
+      dlog(1, FIL__, __LINE__, 
+	   _("This looks like an unexpected internal error.\n"));
       sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1, sh.prg_name);
       aud_exit (FIL__, __LINE__, EXIT_FAILURE);
-      return (-1);
+      SL_RETURN( (-1), _("sh_gpg_check_sign"));
     }
   
   if (what == 0 || what == 1)
     {
       TPT(((0), FIL__, __LINE__, _("msg=<GPG_CHECK: FD1 = %d>\n"), fd1));
-      status = sh_gpg_check_file_sign (fd1, gp.conf_id, gp.conf_fp);
+#if defined(SH_WITH_SERVER)
+      tempres = getpwnam(DEFAULT_IDENT);
+      if (!tempres)
+	tempres = sh_getpwnam(DEFAULT_IDENT);
+      if ((tempres != NULL) && (0 == sl_ret_euid()))
+	{
+	  /* privileges not dropped yet*/
+	  homedir = tempres->pw_dir;
+	}
+#endif
+      status = sh_gpg_check_file_sign (fd1, gp.conf_id, gp.conf_fp, homedir);
       TPT(((0), FIL__, __LINE__, _("msg=<CONF SIGUSR: |%s|>\n"), gp.conf_id));
       TPT(((0), FIL__, __LINE__, _("msg=<CONF SIGFP:  |%s|>\n"), gp.conf_fp));
     }
@@ -665,7 +873,17 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
   if ((what == 0 && SH_GPG_OK == status) || what == 2)
     {
       TPT(((0), FIL__, __LINE__, _("msg=<GPG_CHECK: FD2 = %d>\n"), fd2));
-      status = sh_gpg_check_file_sign (fd2, gp.data_id, gp.data_fp);
+#if defined(SH_WITH_SERVER)
+      tempres = getpwnam(DEFAULT_IDENT);
+      if (!tempres)
+	tempres = sh_getpwnam(DEFAULT_IDENT);
+      if ((tempres != NULL) && (0 == sl_ret_euid()))
+	{
+	  /* privileges not dropped yet*/
+	  homedir = tempres->pw_dir;
+	}
+#endif
+      status = sh_gpg_check_file_sign (fd2, gp.data_id, gp.data_fp, homedir);
       TPT(((0), FIL__, __LINE__, _("msg=<DATA SIGUSR: |%s|>\n"), gp.data_id));
       TPT(((0), FIL__, __LINE__, _("msg=<DATA SIGFP:  |%s|>\n"), gp.data_fp));
     }
@@ -676,11 +894,16 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
       if ((sl_strcmp(SH_GPG_FP, gp.conf_fp) == 0))
 	{
 	  if (smsg == S_FALSE)
-	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_GH,
-			     sh.prg_name, sh.real.uid,
-			     (sh.flag.hidefile == S_TRUE) ? 
-			     _("(hidden)") : file_path('C', 'R'), 
-			     gp.conf_id, gp.conf_fp);
+	    {
+	      tmp  = sh_util_safe_name(gp.conf_id);
+	      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_GH,
+			       sh.prg_name, sh.real.uid,
+			       (sh.flag.hidefile == S_TRUE) ? 
+			       _("(hidden)") : file_path('C', 'R'), 
+			       tmp, 
+			       gp.conf_fp);
+	      SH_FREE(tmp);
+	    }
 	  smsg = S_TRUE;
 	  SL_RETURN(0, _("sh_gpg_check_sign"));
 	}
@@ -688,6 +911,9 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 	{
 	  /* fp mismatch
 	   */
+	  dlog(1, FIL__, __LINE__, 
+	       _("The fingerprint of the signing key: %s\ndoes not match the compiled-in fingerprint: %s.\nTherefore the signature could not be verified.\n"), 
+	       gp.conf_fp, SH_GPG_FP);
 	  sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN, 
 		      _("Fingerprint mismatch"), 
 		      _("gpg_check_sign"));
@@ -695,11 +921,16 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 	}
 #else
       if (smsg == S_FALSE)
-	sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_GH,
-			 sh.prg_name, sh.real.uid,
-			 (sh.flag.hidefile == S_TRUE) ? 
-			 _("(hidden)") : file_path('C', 'R'), 
-			 gp.conf_id, gp.conf_fp);
+	{
+	  tmp = sh_util_safe_name(gp.conf_id);
+	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_GH,
+			   sh.prg_name, sh.real.uid,
+			   (sh.flag.hidefile == S_TRUE) ? 
+			   _("(hidden)") : file_path('C', 'R'), 
+			   tmp, 
+			   gp.conf_fp);
+	  SH_FREE(tmp);
+	}
       smsg = S_TRUE;
       SL_RETURN(0, _("sh_gpg_check_sign"));
 #endif
@@ -716,18 +947,45 @@ int sh_gpg_check_sign (long file_1, long file_2, int what)
 	{
 	  /* ID or fp not equal 
 	   */
+	  dlog(1, FIL__, __LINE__, 
+	       _("The fingerprint or ID of the signing key is not the same for the\nconfiguration file and the file signature database.\nTherefore the signature could not be verified.\n"));
+	  tmp  = sh_util_safe_name (gp.conf_id);
+	  tmp2 = sh_util_safe_name (gp.data_id);
 	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_START_GH2,
 			   sh.prg_name, sh.real.uid,
-			   (sh.flag.hidefile == S_TRUE) ? _("(hidden)") : file_path('C', 'R'), gp.conf_id, gp.conf_fp,
-			   (sh.flag.hidefile == S_TRUE) ? _("(hidden)") : file_path('D', 'R'), gp.data_id, gp.data_fp);
-	  
+			   (sh.flag.hidefile == S_TRUE) ? _("(hidden)") : file_path('C', 'R'),
+			   tmp,  gp.conf_fp,
+			   (sh.flag.hidefile == S_TRUE) ? _("(hidden)") : file_path('D', 'R'),
+			   tmp2, gp.data_fp);
+	  SH_FREE(tmp);
+	  SH_FREE(tmp2);
 	}
     }
 
+  if (status != SH_GPG_OK) 
+    {
+      uid_t   e_uid  = sl_ret_euid();
+      char  * e_home = sh.effective.home;
+
+#if defined(SH_WITH_SERVER)
+      struct passwd * e_tempres = getpwnam(DEFAULT_IDENT);
+      if (!e_tempres)
+	e_tempres = sh_getpwnam(DEFAULT_IDENT);
+      if ((e_tempres != NULL) && (0 == sl_ret_euid()))   
+	{
+	  /* privileges not dropped yet */
+	  e_uid  = e_tempres->pw_uid;
+	  e_home = e_tempres->pw_dir;
+	}
+#endif
+      dlog(1, FIL__, __LINE__, 
+	   _("The signature of the configuration file or the file signature database\ncould not be verified. Possible reasons are:\n - the signature key is not in the private keyring of UID %d,\n - there is no keyring in %s/.gnupg, or\n - the file is not signed - did you move /filename.asc to /filename ?\nTo create a signed file, use:\n   gpg -a --clearsign --not-dash-escaped FILE\n   mv FILE.asc FILE\n"), 
+	   (int) e_uid, e_home);
+    }
 
   TPT(((0), FIL__, __LINE__, _("msg=<Status = %d>\n"), status));
 
-  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1, sh.prg_name);
+  sh_error_handle((-1), FIL__, __LINE__, status, MSG_EXIT_ABORT1, sh.prg_name);
   aud_exit (FIL__, __LINE__, EXIT_FAILURE);
 
   return (-1); /* make compiler happy */
