@@ -60,10 +60,25 @@
 
 #undef  FIL__
 #define FIL__  _("sh_mail.c")
+#undef  GOOD
+#undef  BAD
 
-static int failedMail = GOOD;
+static int failedMail = SL_FALSE;
 
+/* MX Resolver Struct
+ */
+typedef struct mx_ {
+  int    pref;
+  char * address;
+} mx;
 
+typedef struct dnsrep_ {
+  int    count;
+  mx   * reply;
+} dnsrep;
+
+static int free_mx (dnsrep * answers);
+static dnsrep * return_mx (char *domain);
 
 /*********************************************
  *  utility function for verifying mails
@@ -103,6 +118,11 @@ int sh_mail_sigverify (char * s)
 
   /* open the file, then check it 
    */
+  if (sl_is_suid())
+    {
+      fprintf(stderr, _("Cannot open file %s in suid mode\n"), s);
+      _exit (EXIT_FAILURE);
+    }
   if ( SL_ISERROR(fd = sl_open_read (s, SL_NOPRIV)))
     {
       fprintf(stderr, _("Could not open file %s\n"), s);
@@ -194,7 +214,7 @@ int sh_mail_sigverify (char * s)
 
       if (numsig == 0)
 	{
-	  sh_util_encode(key, bufc, 1);
+	  sh_util_encode(key, bufc, 1, 'A');
 	  sl_strlcpy (mail_trail_ptr->trail_key,      key,    KEY_LEN+1);
 	  fprintf (stderr, _("(unchecked)\n"));
 	}
@@ -251,7 +271,6 @@ void reset_count_dev_mail(void)
 
 int sh_mail_setaddress (char * address)
 {
-  unsigned long  numcop;
   char     *     p;
 
   SL_ENTER(_("sh_mail_setaddress"));
@@ -266,19 +285,21 @@ int sh_mail_setaddress (char * address)
       else
 	p = &sh.srvmail.alt[address_num*64];
 
-      numcop = sl_strlcpy (p, address, 64);
+      sl_strlcpy (p, address, 64);
       
-      if (SL_ISERROR(numcop))
+      if ((p == NULL) || ( sl_strlen(address) != sl_strlen(p)))
 	{
 	  memset(p, '\0', 64);
 	  SL_RETURN ( (-1), _("sh_mail_setaddress"));
 	}
       address_list[address_num] = p;
+#if 0
       if (!sl_is_suid())
 	{
 	  TPT(( 0, FIL__, __LINE__, _("msg=<address_list[%d] = %s>\n"), 
 		address_num, address_list[address_num]));
 	}
+#endif
       if (setaddress_compiled == S_TRUE)
 	{
 	  ++address_num;
@@ -301,23 +322,24 @@ int sh_mail_setaddress (char * address)
 int sh_mail_setaddress_int (char * address)
 {
   int i;
+  SL_ENTER(_("sh_mail_setaddress_int"));
   setaddress_compiled = S_TRUE;
   i = sh_mail_setaddress(address);
   setaddress_compiled = S_FALSE;
-  return i;
+  SL_RETURN(i, _("sh_mail_setaddress_int"));
 }
-  
-
 
 int sh_mail_setNum (char * str)
 {
   int i = atoi (str);
 
+  SL_ENTER(_("sh_mail_setNum"));
+
   if (i >= 0 && i < SH_FIFO_MAX) 
     sh.mailNum.alarm_interval = i;
   else 
-    return (-1);
-  return (0);
+    SL_RETURN ((-1), _("sh_mail_setNum"));
+  SL_RETURN( (0), _("sh_mail_setNum"));
 }
 
 
@@ -325,10 +347,32 @@ static int all_in_one = S_FALSE;
 
 int sh_mail_setFlag (char * str)
 {
-  return sh_util_flagval(str, &all_in_one);
+  int i;
+  SL_ENTER(_("sh_mail_setFlag"));
+  i = sh_util_flagval(str, &all_in_one);
+  SL_RETURN(i, _("sh_mail_setFlag"));
 }
 
+static char * mail_subject = NULL;
 
+int set_mail_subject (char * str)
+{
+  SL_ENTER(_("set_mail_subject"));
+  if (!str)
+    SL_RETURN( (-1), _("set_mail_subject"));
+
+  if (mail_subject != NULL)
+    SH_FREE(mail_subject);
+
+  if (0 == sl_strncmp(str, _("NULL"), 4))
+    {
+      mail_subject = NULL;
+      SL_RETURN( 0, _("set_mail_subject"));
+    }
+
+  mail_subject = sh_util_strdup(str);
+  SL_RETURN( (0), _("set_mail_subject"));
+}
 
 
 static SH_FIFO * fifo_mail = NULL;
@@ -338,8 +382,10 @@ void sh_mail_emptystack (void)
   char * msg;
   int    len;
 
+  SL_ENTER(_("sh_mail_emptystack"));
+
   if (fifo_mail == NULL)
-    return;
+    SL_RET0(_("sh_mail_emptystack"));
 
   while (NULL != (msg = pop_list(fifo_mail)))
     {
@@ -348,13 +394,28 @@ void sh_mail_emptystack (void)
       SH_FREE(msg);
     }
 
-  return;
+  SL_RET0(_("sh_mail_emptystack"));
 }
 
+/* insert "\r\n" after each 998 char
+ */
+static char * split_string(char * str);
+
 void sh_mail_pushstack (char * msg)
 {
-  if (msg == NULL || failedMail == BAD || sh.srvmail.name[0] == '\0') 
-    return;
+  char * p;
+
+  SL_ENTER(_("sh_mail_pushstack"));
+
+  if (msg == NULL || failedMail == SL_TRUE || sh.srvmail.name[0] == '\0') 
+    SL_RET0(_("sh_mail_pushstack"));
+
+#if 0
+  if (msg != NULL && sl_strlen(msg) > 998)  /* RFC 2822 */
+    msg[998] = '\0';
+#endif
+
+  p = split_string(msg);
 
   if (fifo_mail == NULL)
     {
@@ -362,25 +423,128 @@ void sh_mail_pushstack (char * msg)
       fifo_init(fifo_mail);
     }
 
-  push_list (fifo_mail, msg);
-  ++sh.mailNum.alarm_last;
+  if (0 == push_list (fifo_mail, p))
+    ++sh.mailNum.alarm_last;
+
+  SH_FREE(p);
 
   if (sh.mailNum.alarm_last >= sh.mailNum.alarm_interval)
-    sh_mail_msg (NULL);
+    {
+      BREAKEXIT(sh_mail_msg);
+      sh_mail_msg (NULL);
+    }
 
-  return;
+  SL_RET0(_("sh_mail_pushstack"));
 }
 
+
 /* The mailer.
  */
 static int sh_mail_end_conn (FILE * connfile);
 static FILE * sh_mail_start_conn (int aFlag);
 
+void sh_mail_get_subject(char * message,
+			 char * mheader, int len)
+{
+  st_format rep_serv_tab[] = {
+    { 'T', S_FMT_TIME,    0, 0, NULL},
+    { 'H', S_FMT_STRING,  0, 0, NULL},
+    { 'M', S_FMT_STRING,  0, 0, NULL},
+    { 'S', S_FMT_STRING,  0, 0, NULL},
+    {'\0', S_FMT_ULONG,   0, 0, NULL},
+  };
+
+  char * p;
+  char * mptr;
+  char   sev[8];
+
+  SL_ENTER(_("sh_mail_get_subject"));
+
+  sl_strlcpy(mheader, _("Subject: "), len);
+  if (NULL == strchr(mail_subject, '%'))
+    {
+      sl_strlcat(mheader, mail_subject, len);
+      SL_RET0(_("sh_mail_get_subject"));
+    }
+
+
+  rep_serv_tab[0].data_ulong = (unsigned long) time(NULL);
+  rep_serv_tab[1].data_str   = sh.host.name;
+
+  /* fast forward to the important part
+   */
+  mptr = sl_strstr(message, _("msg="));
+  if (mptr)
+    {
+      mptr += 4;
+      rep_serv_tab[2].data_str   = mptr;
+    }
+  else
+    rep_serv_tab[2].data_str   = message;
+
+  mptr = sl_strstr(message, _("sev="));
+  if (mptr)
+    {
+      mptr += 5;
+      sev[0] = *mptr; ++mptr;
+      sev[1] = *mptr; ++mptr;
+      sev[2] = *mptr; ++mptr;
+      sev[3] = *mptr; ++mptr;
+      sev[4] = '\0';
+    }
+  else
+    {
+      mptr = message;
+      sev[0] = *mptr; ++mptr;
+      sev[1] = *mptr; ++mptr;
+      sev[2] = *mptr; ++mptr;
+      sev[3] = *mptr; ++mptr;
+      if (*mptr == ' ') {
+	sev[4] = '\0';
+      } else {
+	sev[4] = *mptr; ++mptr;
+	if (*mptr == ' ') {
+	  sev[5] = '\0';
+	} else {
+	  sev[5] = *mptr;
+	  sev[6] = '\0';
+	}
+      }
+    }
+  rep_serv_tab[3].data_str   = sev;
+
+
+  p = sh_util_formatted(mail_subject, rep_serv_tab);
+  sl_strlcat(mheader, p, len);
+  SH_FREE(p);
+  SL_RET0(_("sh_mail_get_subject"));
+}
+
 
 #if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
 #include <sys/mman.h>
 #endif
 
+static char * sh_mail_realloc (char * inbuf, size_t * insize, int increase)
+{
+  size_t newsize = (*insize) + increase + 1;
+  char * outbuf;
+
+  SL_ENTER(_("sh_mail_realloc"));
+
+  outbuf = SH_ALLOC(newsize);
+  MLOCK(outbuf, newsize);
+  sl_strlcpy(outbuf, inbuf, newsize);
+
+  memset (inbuf, '\0', (*insize));
+  MUNLOCK(inbuf, (*insize));
+  SH_FREE(inbuf);
+
+  *insize = newsize;
+
+  SL_RETURN( (outbuf), _("sh_mail_realloc"));
+}
+
 int sh_mail_msg (char * message)
 {
     char         subject[32+32+SH_MINIBUF+2+3+SH_PATHBUF];
@@ -390,6 +554,7 @@ int sh_mail_msg (char * message)
     char       * popMsg;
     int          status = 0, errcount;
     int          i;
+    int          num_popped = 0;
 
     char       * bufcompress;
     static int   failcount = 0;
@@ -402,50 +567,75 @@ int sh_mail_msg (char * message)
 
     static  time_t id_audit  = 0;
     static  time_t fail_time = 0;
+    static  time_t success_time = 0;
 
     static  int m_block = 0;
 
     SH_FIFO * fifo_temp = NULL;
 
     char    * theSig;
+    char    * theMsg = NULL;
 
-    if (m_block == 1)
-      return (0);
+    /* #define SH_MAILBUF (256)    */
+#define SH_MAILBUF (8*4096)
+
+    size_t    msgbufsize = SH_MAILBUF;
+    size_t    combufsize = SH_MAILBUF;
 
     SL_ENTER(_("sh_mail_msg"));
 
+    if (m_block == 1)
+      SL_RETURN( (0), _("sh_mail_msg"));
+
     /* Return if we cannot mail.
      */
-    if (failedMail == BAD) 
+    if (failedMail == SL_TRUE) 
       SL_RETURN((-1), _("sh_mail_msg"));
 
-    if (failedMail == GOOD && address_list[0] == NULL)
+    if (failedMail == SL_FALSE && address_list[0] == NULL)
       {
 	TPT((0, FIL__, __LINE__, 
 	     _("msg=<Mail error: no recipient address.>\n")));
-	failedMail = BAD;
+	failedMail = SL_TRUE;
+	SL_RETURN((-1), _("sh_mail_msg"));
+      }
+
+#if 0
+    if (message != NULL && sl_strlen(message) > 998)  /* RFC 2822 */
+      message[998] = '\0';
+#endif
+
+    if ( (success_time > 0) && (fail_time > 0) &&
+	 (time(NULL) - success_time) > 3600*SH_MAX_FAIL)
+      {
+	m_block = 1;
+	sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_SRV_FAIL,
+			 _("mail"), address_list[0]);
+	m_block = 0;
+	sh_mail_emptystack();
+	sh.mailNum.alarm_last = 0;
+	failedMail = SL_TRUE;
 	SL_RETURN((-1), _("sh_mail_msg"));
       }
 
     /* Try at most each hour.
      */
-    if (fail_time > 0 &&  (time(NULL) - fail_time) < 3600)
+    if ((fail_time > 0) && (time(NULL) - fail_time) < 3600)
       {
-	/* -- Save for later. -- 
-	 */
-	sh_mail_pushstack (message);
-	++failcount;
-
-	if (failcount > SH_MAX_FAIL)
+	if (failcount > 3)
 	  {
-	    m_block = 1;
-	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_SRV_FAIL,
-			     _("mail"), address_list[0]);
-	    m_block = 0;
-	    sh_mail_emptystack();
-	    failedMail = BAD;
+	    /* -- Save for later. -- 
+	     */
+	    sh_mail_pushstack (message);
+	    ++failcount;
+	    
+	    SL_RETURN((-1), _("sh_mail_msg"));
+	  }
+	else
+	  {
+	    sleep(2);
+	    ++failcount;
 	  }
-	SL_RETURN((-1), _("sh_mail_msg"));
       }
 
     /* -- Reset time of last failure. --
@@ -461,13 +651,22 @@ int sh_mail_msg (char * message)
 
     /* ---------  Build complete message. ------------------------ */
 
+    theMsg = split_string(message);
 
     /* ---------- Header  ---------------------------------------- */
 
-    sl_strlcpy(mheader, _("Subject: "),            sizeof(mheader));
-    sl_strlcat(mheader, sh_unix_time (0),          sizeof(mheader));
-    sl_strlcat(mheader, " ",                       sizeof(mheader));
-    sl_strlcat(mheader, sh.host.name,              sizeof(mheader));
+    if (mail_subject == NULL)
+      {
+	sl_strlcpy(mheader, _("Subject: "),            sizeof(mheader));
+	sl_strlcat(mheader, sh_unix_time (0),          sizeof(mheader));
+	sl_strlcat(mheader, " ",                       sizeof(mheader));
+	sl_strlcat(mheader, sh.host.name,              sizeof(mheader));
+      }
+    else
+      {
+	sh_mail_get_subject(message, mheader, sizeof(mheader));
+      }
+
     /* RFC 821: Header is terminated by an empty line
      */
     sl_strlcat(mheader, "\015\012\015\012",        sizeof(mheader));
@@ -479,26 +678,33 @@ int sh_mail_msg (char * message)
     sl_strlcat(subject, sh.host.name,              sizeof(subject));
     sl_strlcat(subject, "\r\n",                    sizeof(subject));
 
-#define SH_MAILBUF (4*4096)
 
-    mailMsg     = (char *) SH_ALLOC (SH_MAILBUF);
-    bufcompress = (char *) SH_ALLOC (SH_MAILBUF);
+    mailMsg     = (char *) SH_ALLOC (msgbufsize);
+    bufcompress = (char *) SH_ALLOC (combufsize);
 
-    MLOCK(mailMsg     , SH_MAILBUF);
-    MLOCK(bufcompress , SH_MAILBUF);
+    MLOCK(mailMsg     , msgbufsize);
+    MLOCK(bufcompress , combufsize);
 
-    sl_strlcpy(mailMsg, mheader, SH_MAILBUF);
+    sl_strlcpy(mailMsg, mheader, msgbufsize);
 
 
-    sl_strlcat(mailMsg, _("-----BEGIN MESSAGE-----\r\n"), SH_MAILBUF);
-    sl_strlcat(mailMsg, subject, SH_MAILBUF);
-    sh_util_compress (bufcompress, subject, SH_MAILBUF-KEY_LEN-1);
+    sl_strlcat(mailMsg, _("-----BEGIN MESSAGE-----\r\n"), msgbufsize);
+    sl_strlcat(mailMsg, subject, msgbufsize);
+    sh_util_compress (bufcompress, subject, (combufsize - KEY_LEN - 1));
     if (message != NULL)
       {
-	sl_strlcat(mailMsg, message, SH_MAILBUF);
-	sl_strlcat(mailMsg,  "\r\n", SH_MAILBUF);
+	if ((sl_strlen(theMsg) + sl_strlen(mailMsg) + 1) > 
+	    (msgbufsize-(4*KEY_LEN)))
+	  {
+	    mailMsg     = sh_mail_realloc(mailMsg,     
+					  &msgbufsize, sl_strlen(theMsg)+2);
+	    bufcompress = sh_mail_realloc(bufcompress, 
+					  &combufsize, sl_strlen(theMsg));
+	  }
+	sl_strlcat(mailMsg,  theMsg, msgbufsize-(4*KEY_LEN));
+	sl_strlcat(mailMsg,  "\r\n", msgbufsize-(4*KEY_LEN));
 
-	sh_util_compress (bufcompress, message, SH_MAILBUF-KEY_LEN-1);
+	sh_util_compress (bufcompress,  theMsg, combufsize-KEY_LEN-1);
       }
 
     if (sh.mailNum.alarm_last > 0)
@@ -509,23 +715,39 @@ int sh_mail_msg (char * message)
 	while ( NULL != (popMsg = pop_list(fifo_mail)) )
 	  {
 	    push_list (fifo_temp, popMsg);
-	    sl_strlcat(mailMsg, popMsg, SH_MAILBUF);
-	    sl_strlcat(mailMsg, "\r\n", SH_MAILBUF);
-	    sh_util_compress (bufcompress, popMsg, SH_MAILBUF-KEY_LEN-1);
+
+	    if ((sl_strlen(popMsg) + sl_strlen(mailMsg) + 1) > 
+		(msgbufsize-(4*KEY_LEN)))
+	      {
+		mailMsg     = sh_mail_realloc(mailMsg,     
+					      &msgbufsize, 
+					      sl_strlen(popMsg)+2);
+		bufcompress = sh_mail_realloc(bufcompress, 
+					      &combufsize, 
+					      sl_strlen(popMsg));
+	      }
+
+	    sl_strlcat(mailMsg, popMsg, msgbufsize-(4*KEY_LEN));
+	    sl_strlcat(mailMsg, "\r\n", msgbufsize-(4*KEY_LEN));
+	    sh_util_compress (bufcompress, popMsg, combufsize-KEY_LEN-1);
 
 	    SH_FREE(popMsg);
 	    --sh.mailNum.alarm_last;
+	    ++num_popped;
 	  }
       }
 
     /* ------ signature block ------------------------------------ */
     
-    sl_strlcat(mailMsg, _("-----BEGIN SIGNATURE-----\r\n"), SH_MAILBUF);
+    sl_strlcat(mailMsg, _("-----BEGIN SIGNATURE-----\r\n"), msgbufsize);
 
     /* Generate new signature key.
      */
     if (isfirst == 1)
-      sh_util_keyinit (skey->mailkey_old, KEY_LEN+1);
+      {
+	BREAKEXIT(sh_util_keyinit);
+	sh_util_keyinit (skey->mailkey_old, KEY_LEN+1);
+      }
 
     /* iterate the key
      */
@@ -541,7 +763,7 @@ int sh_mail_msg (char * message)
 				 bufcompress, sl_strlen(bufcompress));
 	sl_strlcat (mailMsg, 
 		    theSig,
-		    SH_MAILBUF);
+		    msgbufsize);
       }
     else
       {
@@ -551,13 +773,14 @@ int sh_mail_msg (char * message)
          */
         sl_strlcpy(skey->crypt, skey->mailkey_new, KEY_LEN+1);
 
-        sh_util_encode(skey->crypt, bufcompress, 0);
+	BREAKEXIT(sh_util_encode);
+        sh_util_encode(skey->crypt, bufcompress, 0, 'A');
 
-        sl_strlcat (mailMsg, skey->crypt, SH_MAILBUF);
+        sl_strlcat (mailMsg, skey->crypt, msgbufsize);
         memset (skey->crypt, '\0', KEY_LEN);
         isfirst = 0;
       }
-    sl_strlcat (mailMsg, "\r\n", SH_MAILBUF);
+    sl_strlcat (mailMsg, "\r\n", msgbufsize);
 
     /* X(n) -> X(n-1)
      */
@@ -566,10 +789,10 @@ int sh_mail_msg (char * message)
     sprintf(subject, _("%06d %010ld::%s\r\n"),         /* known to fit  */
             mailcount, (long) id_audit, sh.host.name);
 
-    sl_strlcat (mailMsg, subject, SH_MAILBUF);
+    sl_strlcat (mailMsg, subject, msgbufsize);
     ++mailcount;
 
-    sl_strlcat (mailMsg, _("-----END MESSAGE-----"), SH_MAILBUF);
+    sl_strlcat (mailMsg, _("-----END MESSAGE-----"), msgbufsize);
 
 
 
@@ -635,12 +858,12 @@ int sh_mail_msg (char * message)
       }
 
     
-    memset (bufcompress, '\0', SH_MAILBUF);
-    MUNLOCK(bufcompress , SH_MAILBUF);
+    memset (bufcompress, '\0', combufsize);
+    MUNLOCK(bufcompress , combufsize);
     SH_FREE(bufcompress);
 
-    memset (mailMsg, '\0', SH_MAILBUF);
-    MUNLOCK(mailMsg , SH_MAILBUF);
+    memset (mailMsg, '\0', msgbufsize);
+    MUNLOCK(mailMsg , msgbufsize);
     SH_FREE(mailMsg);
 
     /* --- Stay responsible for delivery in case of failure --- */
@@ -649,14 +872,17 @@ int sh_mail_msg (char * message)
       {
         while ( (NULL != (popMsg = pop_list(fifo_temp))) )
           {
-            push_list (fifo_mail, popMsg);
-	    ++sh.mailNum.alarm_last;
+            if (0 == push_list (fifo_mail, popMsg))
+	      ++sh.mailNum.alarm_last;
             SH_FREE(popMsg);
           }
 	if (message != NULL)
 	  {
+#if 0
 	    push_list (fifo_mail, message);
-	    ++sh.mailNum.alarm_last;
+#endif
+	    if (0 == push_list (fifo_mail,  theMsg))
+	      ++sh.mailNum.alarm_last;
 	  }
       }
     else if (fifo_temp != NULL)
@@ -671,6 +897,9 @@ int sh_mail_msg (char * message)
 
     if (connfile != NULL) fclose (connfile);
 
+    if (theMsg != NULL)
+      SH_FREE(theMsg);
+
     /* --- Reset signal. ---
      */
     sigaction (SIGPIPE, &old_act, NULL);
@@ -680,6 +909,8 @@ int sh_mail_msg (char * message)
 	fail_time = time(NULL);
 	SL_RETURN((-1), _("sh_mail_msg"));
       }
+    success_time = time(NULL);
+    failcount = 0;
 
     SL_RETURN((0), _("sh_mail_msg"));
 }
@@ -712,19 +943,27 @@ int sh_mail_set_relay (char * str_s)
 {
   int i = 0;
 
+  SL_ENTER(_("sh_mail_set_relay"));
+
   if (str_s == NULL)
-    return -1;
+    SL_RETURN( -1, _("sh_mail_set_relay"));
 
   if (relay_host != NULL)
-    my_free (relay_host);
+    SH_FREE (relay_host);
+
+  if (0 == sl_strncmp(str_s, _("NULL"), 4))
+    {
+      relay_host = NULL;
+      SL_RETURN( 0, _("sh_mail_set_relay"));
+    }
 
   i = sl_strlen(str_s) + 1;
-  relay_host = my_malloc (i);
+  relay_host = SH_ALLOC (i);
   if (relay_host != NULL)
     sl_strlcpy(relay_host, str_s, i);
   else
     fprintf(stderr, _("ERROR:  sh_mail_set_relay: Out of memory"));
-  return 0;
+  SL_RETURN( 0, _("sh_mail_set_relay"));
 }
 
 
@@ -756,6 +995,9 @@ static FILE * sh_mail_start_conn (int aFlag)
 
   int                  fd;
 
+  dnsrep     * answers;
+  mx         * result;
+
   SL_ENTER(_("sh_mail_start_conn"));
 
   time_wait = 300;
@@ -802,18 +1044,49 @@ static FILE * sh_mail_start_conn (int aFlag)
 
 
   if (relay_host != NULL) 
-    sl_strlcpy (m_machine, relay_host, sizeof(m_machine));
- 
-  TPT(( 0, FIL__, __LINE__, _("msg=<user %s machine %s>\n"), 
-	m_user, m_machine)); 
+    {
+      sl_strlcpy (m_machine, relay_host, sizeof(m_machine));
+      TPT((0, FIL__, __LINE__, _("msg=<user %s machine %s>\n"), 
+	   m_user, m_machine)); 
+      fd = connect_port (m_machine, IPPORT_SMTP, 
+			 error_call, &error_num, error_msg, 256);
+    }
+  else
+    {
+      answers = return_mx (m_machine);
+      if (answers)
+	{
+	  result = answers->reply;
+	  fd     = -1;
+ 	  for (i = 0; i < answers->count; ++i)
+	    {
+	      sl_strlcpy(m_machine, result[i].address, sizeof(m_machine));
+	      TPT((0, FIL__, __LINE__, 
+		   _("msg=<user %s mx %s pref %d>\n"), 
+		   m_user, m_machine, result[i].pref));
+	      fd = connect_port (m_machine, IPPORT_SMTP, 
+				 error_call, &error_num, error_msg, 256);
+	      if (fd >= 0)
+		break;
+	    }
+	  free_mx(answers);
+	}
+      else
+	{
+	  sl_strlcpy(error_call, _("return_mx"), SH_MINIBUF);
+	  sl_strlcpy(error_msg, _("The specified host is unknown: "), 256);
+	  sl_strlcat(error_msg, m_machine, 256); 
 
-  fd = connect_port (m_machine, IPPORT_SMTP, 
-		     error_call, &error_num, error_msg, 256);
+	  fd = -1;
+	}
+    }
 
+  
   if (fd < 0)
     {
-      sh_error_handle((-1), FIL__, __LINE__, error_num, MSG_TCP_NETRP, 
-		      error_msg, (long) IPPORT_SMTP, error_call);
+      sh_error_handle ((-1), FIL__, __LINE__, error_num, 
+		       MSG_E_NET, error_msg, error_call,
+		       _("email"), m_machine);
       SL_RETURN( NULL, _("sh_mail_start_conn"));
     }
 
@@ -832,6 +1105,10 @@ static FILE * sh_mail_start_conn (int aFlag)
    */
   if (!sh_mail_wait (220, connFile)) 
     {
+      sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_NET, 
+		      _("Timeout on SMTP session init"), 
+		      _("sh_mail_start_conn"), 
+		      _("mail"), sh.host.name);
       TPT(( 0, FIL__, __LINE__, _("msg=<Timeout>\n")));
       MBLK( fclose(connFile); )
       SL_RETURN( NULL, _("sh_mail_start_conn"));
@@ -870,8 +1147,11 @@ static FILE * sh_mail_start_conn (int aFlag)
   /* tell them who we are
    */
   sl_strlcpy (this_address, DEFAULT_SENDER, 256);
-  sl_strlcat (this_address, "@", 256);
-  sl_strlcat (this_address, sh.host.name, 256);
+  if (NULL == strchr(this_address, '@'))
+    {
+      sl_strlcat (this_address, "@", 256);
+      sl_strlcat (this_address, sh.host.name, 256);
+    }
 
   TPT(( 0, FIL__, __LINE__,  _("msg=<MAIL FROM:<%s>>%c%c"), 
 	this_address, 13, 10));
@@ -916,6 +1196,8 @@ static FILE * sh_mail_start_conn (int aFlag)
       ecount = 0;
       for (i = 0; i < address_num; ++i)
 	{
+	  if (address_list[i] == NULL)  /* paranoia */
+	    break;
 	  TPT(( 0, FIL__, __LINE__,  _("msg=<RCPT TO:<%s>>%c%c"), 
 		address_list[i], 13, 10)); 
 	  
@@ -1027,13 +1309,9 @@ static jmp_buf wait_timeout;
 
 static void sh_mail_alarmhandle (int mysignal)
 {
-  struct  sigaction  def_act;
-  def_act.sa_handler = SIG_IGN;
-
   if (mysignal == SIGALRM)   /* use 'signal' to avoid compiler warning */
     {  
       alarm(0);
-      sigaction (SIGALRM, &def_act, NULL);
       longjmp(wait_timeout, 1);
     }
 }
@@ -1042,6 +1320,8 @@ static int sh_mail_wait(int code, FILE * m_socket)
 {
   int rcode, g;
 
+  char errmsg[128];
+
   enum { 
     WAIT_CODE_START, 
     WAIT_CODE, 
@@ -1053,8 +1333,16 @@ static int sh_mail_wait(int code, FILE * m_socket)
 
   struct  sigaction  def_act;
   struct  sigaction  new_act;
+  sigset_t           unblock;
+
+  sigemptyset(&unblock);
+  sigaddset  (&unblock, SIGALRM);
+
   def_act.sa_handler = SIG_IGN;
+
   new_act.sa_handler = sh_mail_alarmhandle;
+  sigemptyset( &new_act.sa_mask );       /* set an empty mask       */
+  new_act.sa_flags = 0;                  /* init sa_flags           */
 
   SL_ENTER(_("mail_wait"));
   
@@ -1062,6 +1350,9 @@ static int sh_mail_wait(int code, FILE * m_socket)
    */
   if (setjmp(wait_timeout) != 0)
     {
+      alarm(0);
+      sigaction (SIGALRM, &def_act, NULL);
+      sigprocmask(SIG_UNBLOCK, &unblock, NULL);
       TPT((0, FIL__, __LINE__, _("msg=<mail_wait: timeout>\n"))); 
       SL_RETURN( 0, _("mail_wait"));
     }
@@ -1071,7 +1362,7 @@ static int sh_mail_wait(int code, FILE * m_socket)
   /* timeout after 5 minutes
    */
   alarm(0);
-  sigaction (SIGALRM, &new_act, NULL);
+  sigaction (SIGALRM, &new_act, &def_act);
   alarm(time_wait);
 
   rcode = 0;
@@ -1121,6 +1412,14 @@ static int sh_mail_wait(int code, FILE * m_socket)
 	   _("msg=<mail_wait: OK got %d (%d) need %d (%d)>\n"),
 	   rcode, (int)(rcode/100), code, (int)(code/100) ));
       g = ((int)(rcode/100) == (int)(code/100)) ? 1 : 0;
+      if (g != 1)
+	{
+          sprintf(errmsg,                              /* known to fit  */
+		  _("Bad response (%d), expected %d"), rcode, code);
+	  sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_NET, 
+			  errmsg, _("sh_mail_wait"), 
+			  _("mail"), _("SMTP server"));
+	}
       waited_time = time(NULL) - waited_time;
       time_wait -= waited_time;
       TPT((0, FIL__, __LINE__, 
@@ -1153,6 +1452,419 @@ static int sh_mail_wait(int code, FILE * m_socket)
   SL_RETURN( 0, _("mail_wait")); 
 }
 
+/* -- function to insert "\r\n" after each 998 chars --
+ */
+
+#define SPLIT_AT 998
+
+static char * split_string(char * str)
+{
+  int size;
+  int blocks;
+  int count = 0;
+
+  char * p, * p0;
+  char * q;
+
+  if (!str)
+    return NULL;
+
+  size   = strlen(str);
+  /* fprintf(stderr, "orig = %d\n", size); */
+  blocks = 1 + (size / SPLIT_AT);
+  
+  size   = size + (2*blocks) + 1;
+  p = SH_ALLOC(size);
+  memset(p, '\0', size);
+  /* fprintf(stderr, "alloc = %d\n", size); */
+  p0 = p;
+
+  q = str;
+  while (*q) {
+    *p = *q;
+    ++p;
+    ++q;
+    ++count;
+    if (0 == (count % SPLIT_AT)) {
+      count = 0;
+      *p = '\r';
+      ++p;
+      *p = '\n';
+      ++p;
+    }
+  }
+  /* fprintf(stderr, "used = %d\n", strlen(p0)); */
+
+  return p0;
+}
+
+
+
+/*****************************************************************
+ *
+ * MX Resolver Routines
+ *
+ *****************************************************************/
+
+#include <netinet/in.h>
+#include <arpa/nameser.h>
+#include <resolv.h>
+#include <netdb.h>
+#include <sys/socket.h>
+#include <arpa/inet.h>
+
+#include "sh_tools.h"
+
+#ifndef HFIXEDSZ
+#define HFIXEDSZ 12
+#endif
+#ifndef QFIXEDSZ
+#define QFIXEDSZ  4
+#endif
+
+
+static unsigned int get_short (unsigned char * loc)
+{
+  unsigned int retval = 0;
+  if (loc)
+    {
+      /* byte order: MSB first
+       */
+      retval = (((unsigned char) * loc) * 256) | ((unsigned char) * (loc + 1));
+    }
+  return (retval);
+}
+
+static dnsrep * get_mx (char *hostname)
+{
+  int  ret, length, status;
+  mx * result;
+
+  typedef union
+  {
+    HEADER head;
+    unsigned char buffer[4096];
+  } querybuf;
+
+  querybuf reply;
+  char expanded[1024];
+  unsigned char * comp_dn, * eom;
+  HEADER * header;
+  int      count, index, type, rdlength, pref;
+  dnsrep * retval;
+
+  SL_ENTER(_("get_mx"));
+
+  if (0 != res_init ())
+    SL_RETURN (NULL, _("get_mx"));
+
+  errno = 0;
+  length = res_query (hostname, C_IN, T_MX, 
+		      (unsigned char *) &reply, 4095);
+  if (length < 1)
+    {
+      /* error handling
+       */
+      if (length == -1)
+	{
+	  if (errno == ECONNREFUSED)
+	    status = ECONNREFUSED;
+	  else
+	    status = h_errno;
+
+#ifdef FIL__
+	  sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, status, MSG_E_SUBGEN,
+			   (errno == ECONNREFUSED) ? 
+			   sh_error_message (status) : 
+			   sh_tools_errmessage(status),
+			   _("res_query"));
+#else
+	  if (errno == ECONNREFUSED)
+	    fprintf(stderr, " ERROR: %s: \n", strerror(errno));
+	  else
+	    fprintf(stderr, "HERROR: %s\n", hstrerror(h_errno));
+#endif
+	}
+      SL_RETURN (NULL, _("get_mx"));
+    }
+
+  ret = 0;
+  header  = (HEADER *) &reply;
+
+  /* start of data section
+   */
+  comp_dn = (unsigned char *) &reply + HFIXEDSZ;
+
+  /* end-of-message
+   */
+  eom     = (unsigned char *) &reply + length;
+
+  /* HEADER NAME  -- must be skipped or decompressed
+   * TYPE         -- type of data we got back, 16 bit integer
+   * CLASS        -- class we got back, also a 16 bit integer 
+   * TTL          -- 32 bit time-to-live. just skip this 
+   * RDLENGTH     -- length of the data to follow 
+   * RDATA        -- the data:
+   *                 PREF  -- 16 bit preference 
+   *                 MX    -- name of mail exchanger, must be decompressed
+   */
+
+  /* Skip the query data. 
+   * QDCOUNT is the number of entries (unsigned 16 bit int). 
+   */
+  count = ntohs (header->qdcount); 
+  for (index = 0; index < count; ++index)
+    {
+      ret = dn_skipname (comp_dn, eom);
+      comp_dn += ret + QFIXEDSZ;
+      if (ret < 1 || comp_dn >= eom)
+	SL_RETURN (NULL, _("get_mx"));
+    }
+  count         = ntohs (header->ancount);
+  if (count < 1)
+    SL_RETURN (NULL, _("get_mx"));
+
+  retval        = SH_ALLOC (sizeof (dnsrep));
+  if (!retval)
+    SL_RETURN (NULL, _("get_mx"));
+  retval->count = count;
+
+  /* allocate space for the results */
+
+  result        = SH_ALLOC (count * sizeof (mx));
+  if (!result)
+    {
+      SH_FREE   (retval);
+      SL_RETURN (NULL, _("get_mx"));
+    }
+  retval->reply = result;
+
+  do
+    {
+      /* HEADER NAME 
+       */
+      ret = dn_expand ((unsigned char *) &reply, eom, comp_dn, 
+		       (char *) expanded, 1023);
+      comp_dn += ret;
+      if (ret < 1 || comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* TYPE
+       */
+      type = get_short (comp_dn);
+      comp_dn += 2;
+      if (type != T_MX || comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* CLASS (re-use 'type' var)
+       */
+      type = get_short (comp_dn);
+      comp_dn += 2;
+      if (comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* TTL
+       */
+      comp_dn += 4;
+      if (comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* RDLENGTH
+       */
+      rdlength = get_short (comp_dn);
+      comp_dn += 2;
+      if (rdlength < 1 || comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      /* RDATA
+       */
+      pref = get_short (comp_dn);
+      comp_dn += 2;
+      if (comp_dn >= eom)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+
+      ret = dn_expand ((unsigned char *) &reply, eom, comp_dn, 
+		       (char *) expanded, 1023);
+      comp_dn += ret;
+      if (ret < 1)
+	{
+	  SH_FREE (result);
+	  SH_FREE (retval);
+	  SL_RETURN (NULL, _("get_mx"));
+	}
+      count--;
+
+      /* fill in the struct 
+       */
+      result[count].pref = pref;
+      result[count].address = SH_ALLOC (strlen (expanded) + 1);
+      strcpy (result[count].address, expanded);        /* known to fit  */
+    }
+  while (ret > 0 && comp_dn < eom && count);
+
+  SL_RETURN (retval, _("get_mx"));
+}
+
+static int comp_mx_pref (const void * a, const void * b)
+{
+  const mx * ax = (const mx *) a;
+  const mx * bx = (const mx *) b;
+  
+  if      (ax->pref > bx->pref)
+    return 1;
+  else if (ax->pref < bx->pref)
+    return -1;
+  else
+    return 0;
+}
+
+/*
+ * return_mx returns a list of valid mail exchangers for domain
+ */
+static dnsrep * return_mx (char *domain)
+{
+  struct hostent *host = NULL;
+  dnsrep * answers = NULL;
+  mx     * result;
+  dnsrep * retval;
+  char     errmsg[128]; 
+
+  SL_ENTER(_("return_mx"));
+
+  if (domain != NULL)
+    answers = get_mx (domain);
+
+  if (answers != NULL && answers->count > 0)
+    {
+      qsort(answers->reply, answers->count, sizeof(mx),
+            comp_mx_pref);
+      SL_RETURN (answers, _("return_mx"));
+    }
+  else
+    {
+#ifdef FIL__
+      if (domain != NULL)
+	{
+	  sl_strlcpy (errmsg, _("No MX record for domain "), 127);
+	  sl_strlcat (errmsg, domain, 127);
+	  sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, 0, MSG_E_SUBGEN,
+			   errmsg,
+			   _("get_mx"));
+#else
+	  strcpy  (errmsg,                               /* known to fit  */
+		   _("No MX record for domain "));
+	  strncat (errmsg, domain, 100);
+	  errmsg[122] = '\0';
+	  fprintf(stderr, "Warning: %s\n", errmsg);
+#endif
+	}
+      if (domain != NULL)
+	host = gethostbyname (domain);
+      if (!host)
+	{
+#ifdef FIL__
+	  sl_strlcpy (errmsg, _("Unknown host "), 127);
+	  sl_strlcat (errmsg, domain, 127);
+	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN,
+			   errmsg,
+			   _("return_mx"));
+#endif
+	  SL_RETURN (NULL, _("return_mx"));
+	}
+      result       = SH_ALLOC (sizeof (mx));
+      retval       = SH_ALLOC (sizeof (dnsrep));
+      retval->reply = result;
+      retval->count = 1;
+      result->pref  = 0;
+      result->address = SH_ALLOC (strlen (host->h_name) + 1);
+      strcpy (result->address, host->h_name);          /* known to fit  */
+      SL_RETURN (retval, _("return_mx"));
+    }
+}
+
+static int free_mx (dnsrep * answers)
+{
+  mx     * result;
+  int      i;
+
+  SL_ENTER(_("free_mx"));
+  if (!answers)
+    SL_RETURN (0, _("return_mx"));
+
+  result = answers->reply;  
+  for (i = 0;  i < answers->count; ++i)
+    {
+      SH_FREE (result[i].address);
+    }
+  SH_FREE(result);
+  SH_FREE(answers);
+  SL_RETURN (0, _("return_mx"));
+}
+
+#ifdef TEST_ONLY
+int main(int argc, char * argv[])
+{
+  int      i;
+  dnsrep * answers;
+  mx     * result;
+
+  if (argc < 2)
+    {
+      fprintf(stderr, "Usage: dns <hostname>\n");
+      return -1;
+    }
+  answers = return_mx(argv[1]);
+
+  if (!answers)
+    {
+      fprintf(stderr, "No answer\n");
+      return -1;
+    }
+
+  if (answers->count > 0)
+    {
+      result = answers->reply;
+      for (i = 0; i < answers->count; ++i)
+	{
+	  fprintf(stderr, "Record %3d: [%3d] %s\n", i, 
+		  result[i].pref, result[i].address);
+	}	  
+    }
+  else
+    {
+      fprintf(stderr, "No answer\n");
+      free_mx(answers);
+      return -1;
+    }
+  free_mx(answers);
+  return (0);
+}
+#endif
+
   
 
 /* if defined(SH_WITH_MAIL) */
