@@ -28,6 +28,7 @@
 
 #include "samhain.h"
 #include "sh_error.h"
+#include "sh_database.h"
 #include "sh_unix.h"
 #include "sh_utils.h"
 #include "sh_files.h"
@@ -37,9 +38,16 @@
 #include "sh_forward.h"
 #include "sh_modules.h"
 #include "sh_gpg.h"
+#include "sh_hash.h"
+#include "sh_ignore.h"
 #ifdef WITH_EXTERNAL
 #include "sh_extern.h"
 #endif
+#ifdef WITH_DATABASE
+#include "sh_database.h"
+#endif
+
+extern int set_reverse_lookup (char * c);
 
 #undef  FIL__
 #define FIL__  _("sh_readconf.c")
@@ -54,6 +62,8 @@ typedef enum {
   SH_SECTION_LOGGROW,
   SH_SECTION_NOIGNORE,
   SH_SECTION_ALLIGNORE,
+  SH_SECTION_USER0,
+  SH_SECTION_USER1,
 #if defined (SH_WITH_MAIL) 
   SH_SECTION_MAIL,
 #endif
@@ -63,6 +73,9 @@ typedef enum {
 #ifdef WITH_EXTERNAL
   SH_SECTION_EXTERNAL,
 #endif
+#ifdef WITH_DATABASE
+  SH_SECTION_DATABASE,
+#endif
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
   SH_SECTION_OTHER,
 #endif
@@ -87,8 +100,13 @@ struct str_ListSections tab_ListSections[] = {
   { N_("[GrowingLogFiles]"),  SH_SECTION_LOGGROW},
   { N_("[IgnoreAll]"),        SH_SECTION_ALLIGNORE},
   { N_("[IgnoreNone]"),       SH_SECTION_NOIGNORE},
+  { N_("[User0]"),            SH_SECTION_USER0},
+  { N_("[User1]"),            SH_SECTION_USER1},
 #ifdef WITH_EXTERNAL
   { N_("[External]"),         SH_SECTION_EXTERNAL}, 
+#endif
+#ifdef WITH_DATABASE
+  { N_("[Database]"),         SH_SECTION_DATABASE}, 
 #endif
   { N_("[EventSeverity]"),    SH_SECTION_THRESHOLD},
 #ifdef SH_WITH_SERVER
@@ -105,7 +123,7 @@ struct str_ListSections tab_ListSections[] = {
 };
 
    
-static void sh_readconfig_line (char * line);
+static int sh_readconfig_line (char * line);
 
 static ShSectionType read_mode = SH_SECTION_NONE;
 
@@ -154,8 +172,11 @@ int sh_readconf_read (void)
 
 #if defined(WITH_GPG) || defined(WITH_PGP)
   int    signed_content = S_FALSE;
+  int    true_content   = S_FALSE;
+#endif
+#if defined(SH_STEALTH) && !defined(SH_STEALTH_MICRO)
+  int    hidden_count = 0;
 #endif
-
   uid_t  euid;
 
   SL_ENTER(_("sh_readconf_read"));
@@ -186,13 +207,31 @@ int sh_readconf_read (void)
    */
   if (local_file == 1)
     {
-      if (0 != tf_trust_check (file_path('C', local_flag), SL_YESPRIV)
-	  ||
-	  SL_ISERROR(fd = sl_open_read(file_path('C',local_flag),SL_YESPRIV)))
+      if (0 != tf_trust_check (file_path('C', local_flag), SL_YESPRIV))
 	{
 	  sl_get_euid(&euid);
+	  dlog(1, FIL__, __LINE__, 
+	       _("The configuration file: %s is untrusted, i.e. an\nuntrusted user owns or can write to some directory in the path.\n"), 
+	       ( (NULL == file_path('C', local_flag)) 
+			     ? _("(null)") : file_path('C', local_flag) ));
+	  sh_error_handle ((-1), FIL__, __LINE__, EACCES, MSG_TRUST, 
+			   (long) euid, 
+			   ( (NULL == file_path('C', local_flag)) 
+			     ? _("(null)") : file_path('C', local_flag) )
+			   );
+	  aud_exit (FIL__, __LINE__, EXIT_FAILURE);
+	}
+      if (SL_ISERROR(fd = sl_open_read(file_path('C',local_flag),SL_YESPRIV)))
+	{
+	  sl_get_euid(&euid);
+	  dlog(1, FIL__, __LINE__, 
+	       _("Could not open the local configuration file for reading because\nof the following error: %s\nIf this is a permission problem, you need to change file permissions\nto make the file readable for the effective UID: %d\n"), 
+	       sl_get_errmsg(), (int) euid);
 	  sh_error_handle ((-1), FIL__, __LINE__, EACCES, MSG_NOACCESS, 
-			   (long) euid, file_path('C', local_flag));
+			   (long) euid, 
+			   ( (NULL == file_path('C', local_flag)) 
+			     ? _("(null)") : file_path('C', local_flag) )
+			   );
 	  aud_exit (FIL__, __LINE__, EXIT_FAILURE);
 	}
     }
@@ -208,8 +247,9 @@ int sh_readconf_read (void)
 #if defined(SH_STEALTH) && !defined(SH_STEALTH_MICRO)
     /* extract the data and copy to temporary file
      */
-  fdTmp = open_tmp();
+  fdTmp = open_tmp(); 
   while ( sh_unix_getline_stealth (fd, line_in, 512) > 0) {
+    hidden_count++;
     if (line_in[0] == '\n')
       {
 	sl_write(fdTmp, line_in, 1);
@@ -218,13 +258,21 @@ int sh_readconf_read (void)
       {
 	sl_write_line(fdTmp, line_in, sl_strlen(line_in));
       }
+#if defined(WITH_GPG) || defined(WITH_PGP)
+    if (0 == sl_strncmp(line_in, _("-----END PGP SIGNATURE-----"), 25))
+      break;
+#else
+    if (0 == sl_strncmp(line_in, _("[EOF]"), 5))
+      break;
+#endif
+    if (hidden_count > 4096)  /* arbitrary safeguard */
+      break;
   }
   sl_close(fd);
   fd = fdTmp;
   sl_rewind (fd);
 #endif
 
-
   /* The system type, release, and machine.
    */
   sprintf(myident, _("%s:%s:%s"),                  /* known to fit  */
@@ -233,6 +281,8 @@ int sh_readconf_read (void)
 
   /* ---  Start reading lines.  ---
    */
+  conf_line = 0;
+
   while ( sh_unix_getline (fd, line_in, 512) > 0) {
 
     ++conf_line;
@@ -251,15 +301,24 @@ int sh_readconf_read (void)
 	else 
 	  continue;
       }
+    else if (true_content == S_FALSE)
+      {
+	if (line[0] == '\n')
+	  true_content = S_TRUE;
+	else
+	  continue;
+      }
     else if (signed_content == S_TRUE)
       { 
-	if (0 == sl_strcmp(line, _("-----END PGP SIGNATURE-----")))
+	if (0 == sl_strcmp(line, _("-----BEGIN PGP SIGNATURE-----")))
 	  break;
 	else if (0 == sl_strcmp(line, _("-----BEGIN PGP SIGNED MESSAGE-----")))
 	  {
 	    sh_error_handle((-1), FIL__, __LINE__, 0, MSG_E_SUBGEN,
 			    _("second signed message in file"),
 			    _("sh_readconf_read"));
+	    dlog(1, FIL__, __LINE__, 
+		 _("There seems to be more than one signed message in the configuration\nfile. Please make sure there is only one signed message.\n"));
 	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EXIT_ABORT1,
 			     sh.prg_name);
 	    aud_exit (FIL__, __LINE__,EXIT_FAILURE);
@@ -279,7 +338,8 @@ int sh_readconf_read (void)
 
     /* Skip header etc. 
      */
-    if (line[0] == '#' || line[0] == '\0' || line[0] == ';' || line[0] == '/')
+    if (line[0] == '#' || line[0] == '\0' || line[0] == ';' || 
+	(line[0] == '/' && line[1] == '/'))
       continue; 
   
     /* Clip off trailing white space.                 
@@ -327,8 +387,10 @@ int sh_readconf_read (void)
 	  }
 #endif
 	if (read_mode == SH_SECTION_NONE)
-	  sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, 0, MSG_EINVALL,
-			   _("config file"), (long) conf_line);
+	  {
+	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALHEAD,
+			     (long) conf_line);
+	  }
       } 
 
     /* ---  an @host directive -------------- */
@@ -396,6 +458,20 @@ int sh_readconf_read (void)
 	  ++host_int;
       } 
 
+    /* ---  an %schedule directive ------------ */
+
+    else if (line[0] == '%' || (line[0] == '!' && line[1] == '%')) 
+      {
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
+	if (line[0] == '!' && 0 == sl_strcmp(&(line[2]), _("SCHEDULE_TWO")))
+	  set_dirList(1);
+	else if (0 == sl_strcmp(&(line[1]), _("SCHEDULE_TWO")))
+	  set_dirList(2);
+#else
+	;
+#endif
+      }
+	  
     /* ---  an $system directive -------------- */
 
     else if (line[0] == '$' || (line[0] == '!' && line[1] == '$')) 
@@ -461,7 +537,11 @@ int sh_readconf_read (void)
 	     some_other_system[seen_system] == 0      && 
 	     read_mode != SH_SECTION_NONE) 
       { 
-	sh_readconfig_line (line);
+	if (0 != sh_readconfig_line (line))
+	  {
+	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALCONF,
+			     (long) conf_line);
+	  }
       }
 
   }
@@ -482,6 +562,7 @@ int sh_readconf_read (void)
 #endif
 
   sl_close (fd);
+
   SL_RETURN( 0, _("sh_readconf_read"));
 }
 
@@ -589,6 +670,25 @@ int sh_readconf_setFiletime (char * c)
   return sh_readconf_setTime (c, SET_FILETIME);
 }
 
+int sh_readconf_set_nice (char * c)
+{
+  long val;
+
+  SL_ENTER(_("sh_readconf_set_nice"));
+
+  val = strtol (c, (char **)NULL, 10);
+  if (val < -20 || val > 20)
+    {
+      SL_RETURN((-1), _("sh_readconf_set_nice"));
+    }
+
+  val = (val < -19 ? -19 : val);
+  val = (val >  19 ?  19 : val);
+
+  sh.flag.nice =  val;
+  SL_RETURN((0), _("sh_readconf_set_nice"));
+}
+
 
 typedef struct _cfg_options {
   char * optname;
@@ -597,6 +697,15 @@ typedef struct _cfg_options {
   int (*func)(char * opt);
 } cfg_options;
 
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
+extern int sh_set_schedule_one(char * str);
+extern int sh_set_schedule_two(char * str);
+#endif
+#if defined (SH_WITH_SERVER)
+extern int sh_socket_use (char * c);
+extern int sh_socket_uid (char * c);
+#endif
+
 cfg_options ext_table[] = {
 #if defined(WITH_EXTERNAL)
   { N_("opencommand"),     SH_SECTION_EXTERNAL, SH_SECTION_NONE,  
@@ -621,12 +730,33 @@ cfg_options ext_table[] = {
     sh_ext_add_and },
   { N_("setfilteror"),     SH_SECTION_EXTERNAL, SH_SECTION_NONE,  
     sh_ext_add_or },
-  { N_("externalseverity"),SH_SECTION_LOG,      SH_SECTION_NONE,  
+  { N_("externalseverity"),SH_SECTION_LOG,      SH_SECTION_EXTERNAL,  
     sh_error_set_external },
-  { N_("externalclass"),   SH_SECTION_LOG,      SH_SECTION_NONE,  
+  { N_("externalclass"),   SH_SECTION_LOG,      SH_SECTION_EXTERNAL,  
     sh_error_external_mask },
 #endif
 
+#if defined(WITH_DATABASE)
+  { N_("setdbname"),       SH_SECTION_DATABASE, SH_SECTION_NONE,  
+    sh_database_set_database },
+  { N_("setdbtable"),      SH_SECTION_DATABASE, SH_SECTION_NONE,  
+    sh_database_set_table },
+  { N_("setdbhost"),       SH_SECTION_DATABASE, SH_SECTION_NONE,  
+    sh_database_set_host },
+  { N_("setdbuser"),       SH_SECTION_DATABASE, SH_SECTION_NONE,  
+    sh_database_set_user },
+  { N_("setdbpassword"),   SH_SECTION_DATABASE, SH_SECTION_NONE,  
+    sh_database_set_password },
+  { N_("addtodbhash"),     SH_SECTION_DATABASE, SH_SECTION_NONE,  
+    sh_database_add_to_hash },
+  { N_("databaseseverity"),SH_SECTION_LOG,      SH_SECTION_DATABASE,  
+    sh_error_set_database },
+  { N_("databaseclass"),   SH_SECTION_LOG,      SH_SECTION_DATABASE,  
+    sh_error_database_mask },
+  { N_("setdbservertstamp"), SH_SECTION_DATABASE,      SH_SECTION_NONE,  
+    set_enter_wrapper },
+#endif
+
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)
   { N_("dir"),            SH_SECTION_ATTRIBUTES, SH_SECTION_NONE, 
     sh_files_pushdir_attr },
@@ -647,11 +777,31 @@ cfg_options ext_table[] = {
   { N_("dir"),            SH_SECTION_NOIGNORE,   SH_SECTION_NONE, 
     sh_files_pushdir_noig },
   { N_("file"),           SH_SECTION_NOIGNORE,   SH_SECTION_NONE, 
-sh_files_pushfile_noig },
+    sh_files_pushfile_noig },
   { N_("dir"),            SH_SECTION_ALLIGNORE,  SH_SECTION_NONE, 
     sh_files_pushdir_allig },
   { N_("file"),           SH_SECTION_ALLIGNORE,  SH_SECTION_NONE, 
     sh_files_pushfile_allig },
+
+  { N_("dir"),            SH_SECTION_USER0,      SH_SECTION_NONE, 
+    sh_files_pushdir_user0 },
+  { N_("file"),           SH_SECTION_USER0,      SH_SECTION_NONE, 
+    sh_files_pushfile_user0 },
+  { N_("dir"),            SH_SECTION_USER1,      SH_SECTION_NONE, 
+    sh_files_pushdir_user1 },
+  { N_("file"),           SH_SECTION_USER1,      SH_SECTION_NONE, 
+    sh_files_pushfile_user1 },
+
+  { N_("ignoreadded"), SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_ignore_add_new },
+  { N_("ignoremissing"), SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_ignore_add_del },
+
+  { N_("filecheckscheduleone"), SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_set_schedule_one },
+  { N_("filecheckscheduletwo"), SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_set_schedule_two },
+
   { N_("setrecursionlevel"),  SH_SECTION_MISC,   SH_SECTION_NONE, 
     sh_files_setrecursion },
   { N_("checksumtest"),       SH_SECTION_MISC,   SH_SECTION_NONE, 
@@ -660,6 +810,22 @@ sh_files_pushfile_noig },
     sh_files_reportonce },
   { N_("reportfulldetail"),   SH_SECTION_MISC,   SH_SECTION_NONE, 
     sh_files_fulldetail },
+  { N_("uselocaltime"),       SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_unix_uselocaltime },
+
+  { N_("setnicelevel"),   SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_readconf_set_nice },
+
+#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
+  { N_("setiolimit"),   SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_unix_set_io_limit },
+#endif
+
+  { N_("versionstring"),        SH_SECTION_MISC,   SH_SECTION_NONE,
+    sh_hash_version_string },
+
+  { N_("digestalgo"),           SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_tiger_hashtype },
 
   { N_("redefreadonly"),        SH_SECTION_MISC,   SH_SECTION_NONE, 
     sh_files_redef_readonly },
@@ -679,9 +845,29 @@ sh_files_pushfile_noig },
   { N_("redefignoreall"),       SH_SECTION_MISC,   SH_SECTION_NONE, 
     sh_files_redef_allignore },
 
+  { N_("redefuser0"),           SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_files_redef_user0 },
+
+  { N_("redefuser1"),           SH_SECTION_MISC,   SH_SECTION_NONE, 
+    sh_files_redef_user1 },
+
 #endif
 
 #ifdef SH_WITH_SERVER
+#ifdef INET_SYSLOG
+  { N_("setudpactive"),        SH_SECTION_SRV,  SH_SECTION_MISC, 
+    set_syslog_active },
+#endif
+  { N_("setusesocket"),        SH_SECTION_SRV,  SH_SECTION_MISC, 
+    sh_socket_use },
+  { N_("setsocketallowuid"),   SH_SECTION_SRV,  SH_SECTION_MISC, 
+    sh_socket_uid },
+  { N_("setstripdomain"),      SH_SECTION_SRV,  SH_SECTION_MISC, 
+    sh_forward_set_strip },
+  { N_("useseparatelogs"),     SH_SECTION_SRV,  SH_SECTION_MISC, 
+    set_flag_sep_log },
+  { N_("setchrootdir"),        SH_SECTION_SRV,  SH_SECTION_MISC, 
+    sh_unix_set_chroot },
   { N_("setclienttimelimit"),  SH_SECTION_SRV,  SH_SECTION_MISC, 
     sh_forward_set_time_limit },
   { N_("useclientseverity"),   SH_SECTION_SRV,  SH_SECTION_MISC, 
@@ -690,6 +876,8 @@ sh_files_pushfile_noig },
   sh_forward_use_clt_class },
   { N_("severitylookup"),      SH_SECTION_SRV,  SH_SECTION_MISC, 
   sh_forward_lookup_level },
+  { N_("setclientfromaccept"), SH_SECTION_SRV,  SH_SECTION_MISC, 
+    set_socket_peer },
   { N_("client"),              SH_SECTION_CLIENTS,           SH_SECTION_NONE, 
     sh_forward_register_client },
 #endif
@@ -722,6 +910,8 @@ sh_files_pushfile_noig },
     sh_mail_set_relay },
   { N_("mailsingle"),        SH_SECTION_MAIL,  SH_SECTION_MISC, 
     sh_mail_setFlag },
+  { N_("mailsubject"),       SH_SECTION_MAIL,  SH_SECTION_MISC, 
+    set_mail_subject },
 #endif
 
   { N_("daemon"),            SH_SECTION_MISC,  SH_SECTION_NONE, 
@@ -747,6 +937,12 @@ sh_files_pushfile_noig },
     sh_error_set_syslog },
   { N_("syslogclass"),       SH_SECTION_LOG,   SH_SECTION_NONE, 
     sh_error_syslog_mask },
+#ifdef HAVE_LIBPRELUDE
+  { N_("preludeseverity"),   SH_SECTION_LOG,   SH_SECTION_NONE, 
+    sh_error_set_prelude },
+  { N_("preludeclass"),      SH_SECTION_LOG,   SH_SECTION_NONE, 
+    sh_error_prelude_mask },
+#endif
 
   { N_("logcalls"),          SH_SECTION_LOG,   SH_SECTION_NONE, 
     sh_aud_set_functions },
@@ -758,13 +954,13 @@ sh_files_pushfile_noig },
     sh_log_set_console },
 
 #ifdef WITH_MESSAGE_QUEUE
-  { N_("setconsole"),        SH_SECTION_MISC,  SH_SECTION_NONE, 
-    sh_log_set_console },
-
   { N_("messagequeueactive"),SH_SECTION_MISC,  SH_SECTION_NONE, 
     enable_msgq },
 #endif
 
+  { N_("setreverselookup"),    SH_SECTION_MISC,  SH_SECTION_NONE, 
+    set_reverse_lookup },
+
   { N_("setdatabasepath"),    SH_SECTION_MISC,  SH_SECTION_NONE, 
     sh_readconf_set_database_path },
 
@@ -789,11 +985,13 @@ sh_files_pushfile_noig },
 
 
 
-static void sh_readconfig_line (char * line)
+static int sh_readconfig_line (char * line)
 {    
   char * c;
   char * tmp;
   int    i;
+  int    good_opt = -1;
+
 #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
   int    modnum, modkey;
 #endif
@@ -808,6 +1006,8 @@ static void sh_readconfig_line (char * line)
     N_("severitydirs"),
     N_("severityfiles"),
     N_("severitynames"),
+    N_("severityuser0"),
+    N_("severityuser1"),
     NULL
   };
 
@@ -820,7 +1020,9 @@ static void sh_readconfig_line (char * line)
     SH_ERR_T_ATTR, 
     SH_ERR_T_DIR,   
     SH_ERR_T_FILE, 
-    SH_ERR_T_NAME        
+    SH_ERR_T_NAME,       
+    SH_ERR_T_USER0,       
+    SH_ERR_T_USER1,       
   };
     
   SL_ENTER(_("sh_readconf_line"));
@@ -835,7 +1037,7 @@ static void sh_readconfig_line (char * line)
 	  TPT(( 0, FIL__, __LINE__, _("msg=<ConfigFile: not key=value: %s>\n"),
 		line));
 	}
-      SL_RET0(_("sh_readconf_line"));
+      SL_RETURN(good_opt, _("sh_readconf_line"));
     }
   else
     ++c;
@@ -850,7 +1052,7 @@ static void sh_readconfig_line (char * line)
 	  TPT(( 0, FIL__, __LINE__, _("msg=<ConfigFile: not key=value: %s>\n"),
 		line));
 	}
-      SL_RET0(_("sh_readconf_line"));
+      SL_RETURN(good_opt, _("sh_readconf_line"));
     }
 
   /* convert to lowercase                              */
@@ -880,7 +1082,10 @@ static void sh_readconfig_line (char * line)
 			      _(modList[modnum].conf_table[modkey].the_opt),
 			      sl_strlen(modList[modnum].conf_table[modkey].the_opt)) == 0)
 		{
-		  modList[modnum].conf_table[modkey].func(c);
+		  good_opt = 0;
+		  if (0 != modList[modnum].conf_table[modkey].func(c))
+		    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALS,
+				     _(modList[modnum].conf_table[modkey].the_opt), c);
 		  if (!sl_is_suid())
 		    {
 		      TPT(( 0, FIL__, __LINE__, 
@@ -899,7 +1104,11 @@ static void sh_readconfig_line (char * line)
       i = 0;
       while (ident[i] != NULL) {
 	if (sl_strncmp (line, _(ident[i]), sl_strlen(ident[i])-1) == 0)
-	  sh_error_set_iv (identnum[i], c);
+	  {
+	    good_opt = 0;
+	    sh_error_set_iv (identnum[i], c);
+	    break;
+	  }
 	++i;
       }
     }
@@ -913,6 +1122,7 @@ static void sh_readconfig_line (char * line)
 	      sl_strncmp (line, _(ext_table[i].optname), 
 			  sl_strlen(ext_table[i].optname)) == 0)
 	    {
+	      good_opt = 0;
 	      if (0 != ext_table[i].func (c))
 		sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_EINVALS,
 				 _(ext_table[i].optname), c);
@@ -921,8 +1131,8 @@ static void sh_readconfig_line (char * line)
 	  ++i;
 	}
     }
-     
-  SL_RET0(_("sh_readconf_line"));
+
+  SL_RETURN(good_opt, _("sh_readconf_line"));
 }
   
     
