@@ -72,6 +72,9 @@
 #include "sh_modules.h"
 #include "sh_suidchk.h"
 #include "sh_hash.h"
+#include "sh_unix.h"
+#include "sh_files.h"
+#include "sh_schedule.h"
 
 
 sh_rconf sh_suidchk_table[] = {
@@ -87,6 +90,18 @@ sh_rconf sh_suidchk_table[] = {
     N_("suidcheckinterval"),
     sh_suidchk_set_timer
   },
+  {
+    N_("suidcheckschedule"),
+    sh_suidchk_set_schedule
+  },
+  {
+    N_("suidcheckexclude"),
+    sh_suidchk_set_exclude
+  },
+  {
+    N_("suidcheckfps"),
+    sh_suidchk_set_fps
+  },
   {
     NULL,
     NULL
@@ -96,17 +111,29 @@ sh_rconf sh_suidchk_table[] = {
 
 static time_t  lastcheck         = (time_t) 0;
 static int     ShSuidchkActive   = S_TRUE;
-static int     ShSuidchkInterval = 7200;
+static long    ShSuidchkInterval = 7200;
+static long    ShSuidchkFps      = 0;
 static int     ShSuidchkSeverity = SH_ERR_SEVERE;
+static char *  ShSuidchkExclude  = NULL;
+static int     ExcludeLen        = 0;
 
+static time_t  FileLimNow        = 0;
+static time_t  FileLimStart      = 0;
+static long    FileLimNum        = 0;
+static long    FileLimTotal      = 0;
+
+static sh_schedule_t * ShSuidchkSched = NULL;
 
 static char *
 filesystem_type (char * path, char * relpath, struct stat * statp);
 
 #ifndef PATH_MAX
-#define PATH_MAX 1023
+#define PATH_MAX 1024
 #endif
 
+extern unsigned long sh_files_maskof (int class);
+
+
 int sh_suidchk_check_internal (char * iname)
 {
   DIR *           thisDir = NULL;
@@ -116,14 +143,23 @@ int sh_suidchk_check_internal (char * iname)
   struct stat     buf;
   int             status;
   char          * fs;
-  long            sl_status;
+  long            sl_status = SL_ENONE;
+
+  file_type       theFile;
+  char            fileHash[KEY_LEN + 1];
+
+  SL_ENTER(_("sh_suidchk_check_internal"));
 
   if (iname == NULL)
     {
       TPT((0, FIL__, __LINE__ , _("msg=<directory name is NULL>\n")));
-      return (-1);
+      SL_RETURN( (-1), _("sh_suidchk_check_internal"));
     }
 
+  if (sig_urgent == 1) {
+    SL_RETURN( (0), _("sh_suidchk_check_internal"));
+  }
+
   sl_set_suid();
   thisDir = opendir (iname);
   sl_unset_suid();
@@ -136,10 +172,16 @@ int sh_suidchk_check_internal (char * iname)
 		       sh_error_message(status),
 		       tmp );
       SH_FREE(tmp);
-      return (-1);
+      SL_RETURN( (-1), _("sh_suidchk_check_internal"));
     }
 
   do {
+
+    if (sig_urgent == 1) {
+      SL_RETURN( (0), _("sh_suidchk_check_internal"));
+    }
+
+
     thisEntry = readdir (thisDir);
 
     if (thisEntry != NULL) {
@@ -150,48 +192,72 @@ int sh_suidchk_check_internal (char * iname)
       if (sl_strcmp (thisEntry->d_name, "..") == 0)
 	continue;
 
-      tmpcat = SH_ALLOC(PATH_MAX+1);
-      sl_status = sl_strlcpy(tmpcat, iname,                 PATH_MAX+1);
-      if (! SL_ISERROR(sl_status) && tmpcat[1] != '\0')
-	sl_status = sl_strlcat(tmpcat, "/",                 PATH_MAX+1);
+      tmpcat = SH_ALLOC(PATH_MAX);
+      sl_strlcpy(tmpcat, iname,                 PATH_MAX);
+
+      if ((sl_strlen(tmpcat) != sl_strlen(iname)) || (tmpcat[0] == '\0'))
+	sl_status = SL_ETRUNC;
+      else
+	{
+	  if (tmpcat[1] != '\0') 
+	    sl_status = sl_strlcat(tmpcat, "/",                 PATH_MAX);
+	}
       if (! SL_ISERROR(sl_status))
-	sl_status = sl_strlcat(tmpcat, thisEntry->d_name,   PATH_MAX+1);
+	sl_status = sl_strlcat(tmpcat, thisEntry->d_name,   PATH_MAX);
       if (SL_ISERROR(sl_status))
 	{
 	  tmp = sh_util_safe_name(tmpcat);
 	  sh_error_handle ((-1), FIL__, __LINE__, (int) sl_status, 
 			   MSG_E_SUBGPATH,
 			   _("path too long"),
-			   tmp,
-			   _("sh_suidchk_check_internal") );
+			   _("sh_suidchk_check_internal"), tmp );
 	  SH_FREE(tmp);
 	  continue;
 	}
+
+      ++FileLimNum;
+      ++FileLimTotal;
+
+      if (ShSuidchkFps > 0 && FileLimNum > ShSuidchkFps && FileLimTotal > 0)
+	{
+	  FileLimNum  = 0;
+	  FileLimNow  = time(NULL);
+ 
+	  if ( (FileLimNow  - FileLimStart) > 0 && 
+	       FileLimTotal/(FileLimNow  - FileLimStart) > ShSuidchkFps )
+	    sleep( (int)(FileLimTotal/(FileLimNow  - FileLimStart)) /
+		   ShSuidchkFps );
+	}
+	      
       sl_set_suid();
-      status = lstat(tmpcat, &buf);
+      status = retry_lstat(FIL__, __LINE__, tmpcat, &buf);
       sl_unset_suid();
       if (status != 0)
 	{
 	  status = errno;
 	  tmp = sh_util_safe_name(tmpcat);
-	  sh_error_handle ((-1), FIL__, __LINE__, status, MSG_E_LSTAT,
+	  sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, status, MSG_ERR_LSTAT,
 			   sh_error_message(status),
 			   tmpcat );
 	  SH_FREE(tmp);
 	}
       else
 	{
-	  if (S_ISDIR(buf.st_mode))
+	  if (S_ISDIR(buf.st_mode) &&
+	      (ShSuidchkExclude == NULL || 
+	      0 != strncmp(tmpcat, ShSuidchkExclude, ExcludeLen)))
 	    {
 	      /* fs is a STATIC string
 	       */
 	      fs = filesystem_type (tmpcat, tmpcat, &buf);
 	      if (fs != NULL && 
-		  0 != strcmp (_("nfs"),     fs) && 
-		  0 != strcmp (_("proc"),    fs) &&
-		  0 != strncmp(_("iso9660"), fs, 7) &&
-		  0 != strcmp (_("vfat"),    fs) &&
-		  0 != strcmp (_("msdos"),   fs)
+		  0 != strncmp (_("nfs"),     fs, 3) && 
+		  0 != strncmp (_("proc"),    fs, 4) &&
+		  0 != strncmp (_("iso9660"), fs, 7) &&
+		  0 != strncmp (_("vfat"),    fs, 4) &&
+		  0 != strncmp (_("msdos"),   fs, 5) &&
+		  0 != strncmp (_("devfs"),   fs, 5) &&
+		  0 != strncmp (_("nosuid"),  fs, 6) 
 		  )
 		{
 		  /* fprintf(stderr, "%s: %s\n", fs, tmpcat); */
@@ -201,15 +267,67 @@ int sh_suidchk_check_internal (char * iname)
 	  else if (S_ISREG(buf.st_mode) &&
 		   (0 !=(S_ISUID & buf.st_mode)||0 !=(S_ISGID & buf.st_mode)))
 	    {
-	      if (1 != sh_hash_have_it (tmpcat))
+	      
+	      sl_strlcpy (theFile.fullpath, tmpcat, PATH_MAX);
+	      theFile.check_mask = sh_files_maskof(SH_LEVEL_READONLY);
+	      theFile.reported   = S_FALSE;
+	      status = sh_unix_getinfo (ShDFLevel[SH_ERR_T_RO], 
+					thisEntry->d_name,
+					&theFile, fileHash);
+
+	      tmp = sh_util_safe_name(tmpcat);
+
+	      if (status != 0)
 		{
-		  tmp = sh_util_safe_name(tmpcat);
 		  sh_error_handle (ShSuidchkSeverity, FIL__, __LINE__, 
 				   0, MSG_SUID_POLICY,
-				   _("suid/sgid file not in database"),
+				   _("Could not check suid/sgid file"),
 				   tmp );
-		  SH_FREE(tmp);
 		}
+	      else
+		{
+		  if (sh.flag.checkSum == SH_CHECK_INIT  && 
+		      sh.flag.update == S_TRUE )
+		    {
+		      sh_hash_compdata (SH_LEVEL_READONLY, &theFile, fileHash);
+		      sh_hash_pushdata (&theFile, fileHash);
+		      sh_error_handle ((-1), FIL__, __LINE__, 
+				       0, MSG_SUID_FOUND,
+				       tmp );
+		    }
+		  else if (sh.flag.checkSum == SH_CHECK_INIT  && 
+			   sh.flag.update == S_FALSE )
+		    {
+		      sh_hash_pushdata (&theFile, fileHash);
+		      sh_error_handle ((-1), FIL__, __LINE__, 
+				       0, MSG_SUID_FOUND,
+				       tmp );
+		    }
+		  else if (sh.flag.checkSum == SH_CHECK_CHECK )
+		    {
+		      sh_error_handle (SH_ERR_ALL, FIL__, __LINE__, 
+				       0, MSG_SUID_FOUND,
+				       tmp );
+		      if (-1 == sh_hash_have_it (tmpcat))
+			{
+			  sh_error_handle (ShSuidchkSeverity, FIL__, __LINE__, 
+					   0, MSG_SUID_POLICY,
+					   _("suid/sgid file not in database"),
+					   tmp );
+			  /* 1.8.1 push file to in-memory database
+			   */
+			  sh_hash_compdata (SH_LEVEL_READONLY, 
+					    &theFile, fileHash);
+			}
+		      else
+			{
+			  sh_hash_compdata (SH_LEVEL_READONLY, 
+					    &theFile, fileHash);
+			}
+		    }
+		}
+	      SH_FREE(tmp);
+
 	    }
 	}
       SH_FREE(tmpcat);
@@ -217,7 +335,7 @@ int sh_suidchk_check_internal (char * iname)
   }  while (thisEntry != NULL);
 
   closedir (thisDir);
-  return 0;
+  SL_RETURN( (0), _("sh_suidchk_check_internal"));
 }
 
 /*************
@@ -233,6 +351,7 @@ int sh_suidchk_init ()
   return (0);
 }
 
+
 /*************
  *
  * module cleanup
@@ -251,7 +370,14 @@ int sh_suidchk_end ()
  *************/
 int sh_suidchk_timer (unsigned long tcurrent)
 {
-  if ((time_t) (tcurrent - lastcheck) > ShSuidchkInterval)
+  if (sh.flag.checkSum == SH_CHECK_INIT)
+    return -1;
+
+  if (ShSuidchkSched != NULL)
+    {
+      return test_sched(ShSuidchkSched);
+    }
+  if ((time_t) (tcurrent - lastcheck) >= ShSuidchkInterval)
     {
       lastcheck  = tcurrent;
       return (-1);
@@ -264,12 +390,29 @@ int sh_suidchk_timer (unsigned long tcurrent)
  * module check
  *
  *************/
+
 int sh_suidchk_check ()
 {
+  int status;
+
+  SL_ENTER(_("sh_suidchk_check"));
+
   sh_error_handle (SH_ERR_INFO, FIL__, __LINE__, EINVAL, MSG_E_SUBGEN,
 		   _("Checking for SUID programs"),
 		   _("suidchk_check") );
-  return (sh_suidchk_check_internal ("/"));
+
+  FileLimNow        = time(NULL);
+  FileLimStart      = FileLimNow;
+  FileLimNum        = 0;
+  FileLimTotal      = 0;
+
+  status = sh_suidchk_check_internal ("/");
+
+  sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_SUID_SUMMARY,
+		   FileLimTotal,
+		   (long) (time(NULL) - FileLimStart) );
+
+  SL_RETURN(status, _("sh_suidchk_check"));
 }
 
 /*************
@@ -281,16 +424,51 @@ int sh_suidchk_check ()
 int sh_suidchk_set_severity  (char * c)
 {
   char tmp[32];
+
+  SL_ENTER(_("sh_suidchk_set_severity"));
   tmp[0] = '='; tmp[1] = '\0';
   sl_strlcat (tmp, c, 32);
   sh_error_set_level (tmp, &ShSuidchkSeverity);
-  return 0;
+  SL_RETURN(0, _("sh_suidchk_set_severity"));
+}
+
+int sh_suidchk_set_exclude (char * c)
+{
+  SL_ENTER(_("sh_suidchk_set_exclude"));
+  if (c == NULL)
+    return -1;
+
+  if (0 == sl_strncmp(c, _("NULL"), 4))
+    {
+      if (ShSuidchkExclude != NULL)
+	SH_FREE(ShSuidchkExclude);
+      ShSuidchkExclude = NULL;
+      return 0;
+    }
+
+  if (ShSuidchkExclude != NULL)
+    SH_FREE(ShSuidchkExclude);
+
+  /* 1.8.1 add trailing slash
+   */
+  ExcludeLen       = sl_strlen(c);
+  if (c[ExcludeLen-1] != '/')
+    {
+      ExcludeLen++;
+    }
+  ShSuidchkExclude = SH_ALLOC(ExcludeLen + 1);
+  sl_strlcpy(ShSuidchkExclude, c, (ExcludeLen + 1));
+  ShSuidchkExclude[ExcludeLen-1] = '/';
+
+  SL_RETURN(0, _("sh_suidchk_set_exclude"));
 }
 
 int sh_suidchk_set_timer (char * c)
 {
   long val;
 
+  SL_ENTER(_("sh_suidchk_set_timer"));
+
   val = strtol (c, (char **)NULL, 10);
   if (val <= 0)
     sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS,
@@ -299,15 +477,141 @@ int sh_suidchk_set_timer (char * c)
   val = (val <= 0 ? 7200 : val);
 
   ShSuidchkInterval = (time_t) val;
+  SL_RETURN( 0, _("sh_suidchk_set_timer"));
+}
+
+
+int sh_suidchk_free_schedule ()
+{
+  sh_schedule_t * current = ShSuidchkSched;
+  sh_schedule_t * next    = NULL;
+
+  while (current != NULL)
+    {
+      next = current->next;
+      SH_FREE(current);
+      current = next;
+    }
+  ShSuidchkSched = NULL;
   return 0;
 }
 
+int sh_suidchk_set_schedule (char * str)
+{
+  int status;
+  sh_schedule_t * newSched = NULL;
+
+  SL_ENTER(_("sh_suidchk_set_schedule"));
+
+  /*
+  if (ShSuidchkSched != NULL)
+    {
+      SH_FREE(ShSuidchkSched);
+      ShSuidchkSched = NULL;
+    }
+  */
+
+  if (0 == sl_strncmp(str, _("NULL"), 4))
+    {
+      sh_suidchk_free_schedule ();
+      return 0;
+    }
+
+  newSched = SH_ALLOC(sizeof(sh_schedule_t));
+  status = create_sched(str, newSched);
+  if (status != 0)
+    {
+      SH_FREE(newSched);
+      newSched = NULL;
+    }
+  else
+    {
+      newSched->next = ShSuidchkSched;
+      ShSuidchkSched = newSched;
+    }
+  SL_RETURN( status, _("sh_suidchk_set_schedule"));
+}
+
+
+
+int sh_suidchk_set_fps (char * c)
+{
+  long val;
+
+  SL_ENTER(_("sh_suidchk_set_fps"));
+
+  val = strtol (c, (char **)NULL, 10);
+  if (val < 0)
+    sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS,
+                      _("suidchk fps"), c);
+
+  val = (val < 0 ? 0 : val);
+
+  ShSuidchkFps = val;
+  SL_RETURN( 0, _("sh_suidchk_set_fps"));
+}
+
 int sh_suidchk_set_activate (char * c)
 {
-  return sh_util_flagval(c, &ShSuidchkActive);
+  int i;
+  SL_ENTER(_("sh_suidchk_set_activate"));
+  i = sh_util_flagval(c, &ShSuidchkActive);
+  SL_RETURN(i, _("sh_suidchk_set_activate"));
 }
 
+#if defined(FSTYPE_STATFS) || defined(FSTYPE_AIX_STATFS)
+/* dirname.c -- return all but the last element in a path
+   Copyright (C) 1990 Free Software Foundation, Inc.
 
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation,
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+
+/* Return the leading directories part of PATH,
+   allocated with malloc.  If out of memory, return 0.
+   Assumes that trailing slashes have already been
+   removed.  */
+
+char * sh_dirname (const char * path)
+{
+  char *newpath;
+  char *slash;
+  int length;                   /* Length of result, not including NUL.  */
+
+  slash = strrchr (path, '/');
+  if (slash == NULL)
+    {
+      /* File is in the current directory.  */
+      path = ".";
+      length = 1;
+    }
+  else
+    {
+      /* Remove any trailing slashes from the result.  */
+      while (slash > path && *slash == '/')
+        --slash;
+
+      length = slash - path + 1;
+    }
+  newpath = (char *) SH_ALLOC (length + 1);
+  if (newpath == NULL)
+    return NULL;
+  strncpy (newpath, path, length);
+  newpath[length] = '\0';
+  return newpath;
+}
+/* #ifdef FSTYPE_STATFS */
+#endif
 
 /* fstype.c -- determine type of filesystems that files are on
    Copyright (C) 1990, 91, 92, 93, 94 Free Software Foundation, Inc.
@@ -332,6 +636,7 @@ int sh_suidchk_set_activate (char * c)
    - replaced error()   by sh_error_handle()
    - replaced xstrdup() by sl_strdup()
    - replaced strstr()  by sl_strstr()
+   - some additions to recognize nosuid fs
 */
 
 /* modetype.h -- file type bits definitions for POSIX systems
@@ -417,22 +722,6 @@ int sh_suidchk_set_activate (char * c)
 #endif
 
 
-static char * sl_strdup (char * str) 
-{
-  char * p = NULL;
-  int    len;
-
-  if (str != NULL)
-    {
-      len = sl_strlen(str);
-      p   = SH_ALLOC(len + 1);
-      sl_strlcpy (p, str, len+1);
-    }
-  return p;
-}
-
-
-
 static char *filesystem_type_uncached (char *path, char *relpath, 
 				       struct stat *statp);
 
@@ -540,15 +829,61 @@ fstype_to_string (t)
   switch (t)
     {
     case MNT_AIX:
-#if 0				/* NFS filesystems are actually MNT_AIX. */
-      return "aix";
+      return _("aix");	/* AIX 4.3: NFS filesystems are actually MNT_AIX. */
+#ifdef MNT_NAMEFS
+    case MNT_NAMEFS:
+      return _("namefs");
 #endif
     case MNT_NFS:
-      return "nfs";
+      return _("nfs");
     case MNT_JFS:
-      return "jfs";
+      return _("jfs");
     case MNT_CDROM:
-      return "cdrom";
+      return _("cdrom");
+#ifdef MNT_PROCFS
+    case MNT_PROCFS:
+      return _("procfs");
+#endif
+#ifdef MNT_SFS
+    case MNT_SFS:
+      return _("sfs");
+#endif
+#ifdef MNT_CACHEFS
+    case MNT_CACHEFS:
+      return _("cachefs");
+#endif
+#ifdef MNT_NFS3
+    case MNT_NFS3:
+      return _("nfs3");
+#endif
+#ifdef MNT_AUTOFS
+    case MNT_AUTOFS:
+      return _("autofs");
+#endif
+#ifdef MNT_VXFS
+    case MNT_VXFS:
+      return _("vxfs");
+#endif
+#ifdef MNT_VXODM
+    case MNT_VXODM:
+      return _("veritasfs");
+#endif
+#ifdef MNT_UDF
+    case MNT_UDF:
+      return _("udfs");
+#endif
+#ifdef MNT_NFS4
+    case MNT_NFS4:
+      return _("nfs4");
+#endif
+#ifdef MNT_RFS4
+    case MNT_RFS4:
+      return _("nfs4");
+#endif
+#ifdef MNT_CIFS
+    case MNT_CIFS:
+      return _("cifs");
+#endif
     default:
       return "?";
     }
@@ -622,8 +957,10 @@ filesystem_type_uncached (path, relpath, statp)
      char *relpath;
      struct stat *statp;
 {
-  char *type = NULL;
-
+  char * type = NULL;
+#ifdef MFSNAMELEN		/* NetBSD.  */
+  static char my_tmp_type[64];
+#endif
 
 #ifdef FSTYPE_MNTENT		/* 4.3BSD, SunOS, HP-UX, Dynix, Irix.  */
   char *table = MOUNTED;
@@ -686,7 +1023,14 @@ filesystem_type_uncached (path, relpath, statp)
 	}
 
       if (dev == statp->st_dev)
-	type = mnt->mnt_type;
+	{
+	  /* check for the "nosuid" option
+	   */
+	  if (NULL == hasmntopt(mnt, "nosuid"))
+	    type = mnt->mnt_type;
+	  else
+	    type = _("nosuid");
+	}
     }
 
   if (endmntent (mfp) == 0)
@@ -752,18 +1096,29 @@ filesystem_type_uncached (path, relpath, statp)
 	}
     }
   else
-    type = fss.f_basetype;
+    {
+       type = fss.f_basetype;
+
+       /* patch by Konstantin Khrooschev <nathoo@co.ru> 
+	*/
+       if( fss.f_flag & ST_NOSUID )
+         type = _("nosuid");
+    }
 #endif
 
 #ifdef FSTYPE_STATFS		/* 4.4BSD.  */
   struct statfs fss;
   char *p;
+#if defined(MNT_VISFLAGMASK) && defined(HAVE_STRUCT_STATFS_F_FLAGS)
+  int flags;
+#endif
+  /* char * sh_dirname(const char *path); */
 
   if (path == NULL || relpath == NULL)
     return NULL;
 
   if (S_ISLNK (statp->st_mode))
-    p = dirname (relpath);
+    p = sh_dirname (relpath);
   else
     p = relpath;
 
@@ -781,11 +1136,26 @@ filesystem_type_uncached (path, relpath, statp)
     }
   else
     {
+
 #ifdef MFSNAMELEN		/* NetBSD.  */
-      type = sl_strdup (fss.f_fstypename);
+      /* MEMORY LEAK !!!
+       *	 type = sh_util_strdup (fss.f_fstypename);
+       */
+      sl_strlcpy (my_tmp_type, fss.f_fstypename, 64);
+      type = my_tmp_type;
 #else
       type = fstype_to_string (fss.f_type);
 #endif
+
+#ifdef HAVE_STRUCT_STATFS_F_FLAGS
+#ifdef MNT_VISFLAGMASK
+      flags = fss.f_flags & MNT_VISFLAGMASK;
+      if (flags & MNT_NOSUID)
+#else 
+      if (fss.f_flags & MNT_NOSUID) 
+#endif
+         type = _("nosuid");
+#endif
     }
   if (p != relpath)
     SH_FREE (p);
@@ -800,7 +1170,7 @@ filesystem_type_uncached (path, relpath, statp)
      Don't cache those values.  */
   fstype_known = (type != NULL);
 
-  return sl_strdup (type ? type : "unknown");
+  return sh_util_strdup (type ? type : "unknown");
 }
 
 #ifdef FSTYPE_MNTENT		/* 4.3BSD etc.  */
