@@ -9,12 +9,18 @@ int sh_suidchk_init  (void);
 int sh_suidchk_timer (unsigned long tcurrent);
 int sh_suidchk_check (void);
 int sh_suidchk_end   (void);
+int sh_suidchk_free_schedule (void);
 
 int sh_suidchk_set_activate (char * c);
 int sh_suidchk_set_severity (char * c);
 int sh_suidchk_set_timer    (char * c);
+int sh_suidchk_set_schedule (char * c);
+int sh_suidchk_set_exclude  (char * c);
+int sh_suidchk_set_fps      (char * c);
 
 extern sh_rconf sh_suidchk_table[];
+
+
 #endif
 
 /* #ifndef SH_SUIDCHK_H */
