@@ -18,6 +18,12 @@ extern SL_TICKET tiger_fd;
  */
 char * sh_tiger_hash (char * filename, TigerType what, unsigned long Length);
 
+/* NEW Thu Oct 18 19:59:08 CEST 2001
+ */
+int sh_tiger_hashtype (char * c);
+char * sh_tiger_generic_hash (char * filename, TigerType what, 
+			      unsigned long Length);
+
 UINT32 * sh_tiger_hash_uint32 (char * filename, 
 			       TigerType what, 
 			       unsigned long Length);
