@@ -51,21 +51,71 @@ UINT32 ErrFlag[2];
 
 int sh_util_flagval(char * c, int * fval)
 {
+  SL_ENTER(_("sh_util_flagval"));
   if (c == NULL)
-    return (-1);
-  if ( c[0] == 1 || c[0] == 'y'  || c[0] == 'Y' ||
+    SL_RETURN( (-1), _("sh_util_flagval"));
+  if ( c[0] == '1'  || c[0] == 'y'  || c[0] == 'Y' ||
        c[0] == 't'  || c[0] == 'T')
     {
       *fval = S_TRUE;
-      return (0);
+      SL_RETURN( (0), _("sh_util_flagval"));
     }
-  if ( c[0] == 0 || c[0] == 'n'  || c[0] == 'N' ||
+  if ( c[0] == '0'  || c[0] == 'n'  || c[0] == 'N' ||
        c[0] == 'f'  || c[0] == 'F')
     {
       *fval = S_FALSE;
-      return (0);
+      SL_RETURN( (0), _("sh_util_flagval"));
     }
-  return (-1);
+  SL_RETURN( (-1), _("sh_util_flagval"));
+}
+
+static int sh_ask_update = S_FALSE;
+
+int sh_util_set_interactive(char * str)
+{
+  if (str == NULL)
+    sh_ask_update = S_TRUE;
+  else
+    sh_ask_update = S_TRUE;
+
+  return 0;
+}
+
+int sh_util_ask_update(char * path)
+{
+  int    inchar, c;
+  int    i = S_TRUE;
+  char * tmp = NULL;
+
+  SL_ENTER(_("sh_util_ask_update"));
+
+  if (sh_ask_update == S_TRUE)
+    {
+      tmp = sh_util_safe_name (path);
+      fprintf (stderr, _("Update %s [Y/n] ? "), tmp);
+      SH_FREE(tmp);
+      while (1 == 1)
+	{
+	  c = fgetc(stdin); inchar = c;
+	  while (c != '\n' && c != EOF)
+	    c = fgetc(stdin);
+	  /* fprintf(stderr, "CHAR (1): %c\n", inchar); */
+	  if (inchar == 'Y' || inchar == 'y' || inchar == '\n')
+	    {
+	      break;
+	    }
+	  else if (inchar == 'n' || inchar == 'N')
+	    {
+	      i = S_FALSE;
+	    }
+	  else
+	    {
+	      fprintf(stderr, _("Please answer y(es) or n(o)\n"));
+	    }
+	}
+    }
+
+  SL_RETURN(i, _("sh_util_ask_update"));
 }
 
 int sh_util_hidesetup(char * c)
@@ -77,8 +127,55 @@ int sh_util_hidesetup(char * c)
   SL_RETURN(i, _("sh_util_hidesetup"));
 }
     
+char * sh_util_strdup (const char * str) 
+{
+  char * p = NULL;
+  int    len;
 
-char * sh_util_formatted (char * formatt, st_format * ftab)
+  SL_ENTER(_("sh_util_strdup"));
+
+  if (str != NULL)
+    {
+      len = sl_strlen(str);
+      p   = SH_ALLOC (len + 1);
+      sl_strlcpy (p, str, len+1);
+    }
+  SL_RETURN( p, _("sh_util_strdup"));
+}
+
+/* by the eircom.net computer incident
+ * response team
+ */
+char * sh_util_strsep (char **str, const char *delim) 
+{
+  char *ret, *c, *d;
+
+  SL_ENTER(_("sh_util_strsep"));
+  ret = *str;
+
+  if (ret == NULL) {
+    SL_RETURN(ret, _("sh_util_strsep"));
+  }
+
+  for (c = *str; *c != '\0'; c++) {
+    for (d = (char *) delim; *d != '\0'; d++) {
+      if (*c == *d) {
+        *c = '\0';
+        *str = c + 1;
+        SL_RETURN(ret, _("sh_util_strsep"));
+      }
+    }
+  }
+
+  /* If we get to here, there's no delimiters in the string */
+  *str = NULL;
+  SL_RETURN(ret, _("sh_util_strsep"));
+}
+
+
+/* returned string must be free'd by caller
+ */
+char * sh_util_formatted (const char * formatt, st_format * ftab)
 {
   struct tm   * time_ptr;
   int    size;
@@ -281,7 +378,8 @@ char * sh_util_formatted (char * formatt, st_format * ftab)
   SL_RETURN(outstr, _("sh_util_formatted"));
 }
 
-
+/* can't inline (AIX)
+ */
 int sh_util_hexchar( char c )
 {
   if      ( c >= '0' && c <= '9' )
@@ -323,6 +421,7 @@ static void copy_four (unsigned char * dest, UINT32 in)
   UINT32 i, j;
   int    count;
 
+  SL_ENTER(_("copy_four"));
   for (count = 0; count < 4; ++count)
     {
       i  = in / 256;
@@ -330,7 +429,7 @@ static void copy_four (unsigned char * dest, UINT32 in)
       dest[count] = (unsigned char) j;
       in = i;
     }
-  return;
+  SL_RET0(_("copy_four"));
 }
 
 /* compute HMAC-TIGER
@@ -441,27 +540,33 @@ static int sigtype = TYPE_HMAC;
 
 int sh_util_sigtype (char * c)
 {
+  SL_ENTER(_("sh_util_sigtype"));
   if (c == NULL)
-    return -1;
+    SL_RETURN( -1, _("sh_util_sigtype"));
+
   if (0 == strcmp(_("HMAC-TIGER"), c))
     sigtype = TYPE_HMAC;
   else if  (0 == strcmp(_("HASH-TIGER"), c))
     sigtype = TYPE_HASH;
   else
-    return -1;
+    SL_RETURN( -1, _("sh_util_sigtype"));
 
-  return 0;
+  SL_RETURN( 0, _("sh_util_sigtype"));
 }
 
 char * sh_util_siggen (char * hexkey,  
 		       char * text, int textlen)  
 {
+  char * p;
+  
+  SL_ENTER(_("sh_util_siggen"));
   if (sigtype == TYPE_HMAC)
-    return sh_util_hmac_tiger (hexkey,  
-			       text, textlen);
+    p = sh_util_hmac_tiger (hexkey,  
+			    text, textlen);
   else
-    return sh_util_hash_tiger (hexkey,  
-			       text, textlen);
+    p = sh_util_hash_tiger (hexkey,  
+			    text, textlen);
+  SL_RETURN(p, _("sh_util_siggen"));
 }    
 
   
@@ -480,6 +585,9 @@ long sh_util_compress (char * dest, char * src, unsigned long dest_size)
   if (dest_size == 0)
     SL_RETURN((0), _("sh_util_compress"));
   
+  if ((dest == NULL) || (src == NULL))
+    SL_RETURN((0), _("sh_util_compress"));
+  
   dest_end = sl_strlen(dest);
   add      = &dest[dest_end];
   get      = src;
@@ -493,7 +601,8 @@ long sh_util_compress (char * dest, char * src, unsigned long dest_size)
 	  ++count;
 	}
       ++get; 
-      if (*get == '\0') /* end of src reached */
+      if (*get == '\0' && (count < (dest_size-dest_end))) 
+	/* end of src reached */
 	{
 	  *add = *get;  /* copy the '\0'      */
 	  break;        /* and stop copying   */
@@ -501,7 +610,7 @@ long sh_util_compress (char * dest, char * src, unsigned long dest_size)
     }
 
   dest[dest_size-1] = '\0'; /* paranoia       */
-  SL_RETURN((count), _("sh_util_compress")); /* no of chars copyed */    
+  SL_RETURN((count), _("sh_util_compress")); /* no of chars copied */    
 }
 
 
@@ -613,7 +722,7 @@ UINT32 taus_get (void *state1, void *state2, void *state3)
  */
 double taus_get_double (void *vstate)
 {
-  return taus_get_long (vstate) / (4294967296.0 - 1.0) ;
+  return taus_get_long (vstate) / (4294967296.0 + 1.0) ;
 }
 
 #define LCG(n) ((69069 * n) & 0xffffffffUL)
@@ -662,10 +771,12 @@ int taus_seed ()
   int                  status;
   static unsigned long seed_time = 0;
 
+  SL_ENTER(_("taus_seed"));
+
   if (skey->rngI == GOOD)
     {
       if ( (sh_unix_longtime () - seed_time) < 3600)
-	return (0);
+	SL_RETURN( (0), _("taus_seed"));
     }
   
   seed_time = sh_unix_longtime ();
@@ -680,11 +791,15 @@ int taus_seed ()
       memcpy (&skey->rng2[0], &bufx[4*sizeof(UINT32)],   2*sizeof(UINT32));
       memset (bufx, '\0', 9 * sizeof(UINT32) + 1);
 
+      skey->rng0[2] = 0;
+      skey->rng1[2] = 0;
+      skey->rng2[2] = 0;
+
       taus_set_from_state( &(skey->rng0[0]), &(skey->rng0[0]));
       taus_set_from_state( &(skey->rng1[0]), &(skey->rng1[0]));
       taus_set_from_state( &(skey->rng2[0]), &(skey->rng2[0]));
 
-      return (0);
+      SL_RETURN( (0), _("taus_seed"));
     }
 
   sh_error_handle ((-1), FIL__, __LINE__, status, MSG_ES_ENT,
@@ -698,22 +813,148 @@ int taus_seed ()
 #else
   taus_set_from_ulong ( &(skey->rng0[0]), LCG (seed_time)      );
 #endif
-  taus_set_from_ulong ( &(skey->rng0[0]), LCG (seed_time)      );
   taus_set_from_ulong ( &(skey->rng1[0]), LCG (skey->rng0[0])  );
   taus_set_from_ulong ( &(skey->rng2[0]), LCG (skey->rng1[0])  );
   skey->rngI = BAD;
 
-  return (-1);
+  SL_RETURN( (-1), _("taus_seed"));
 }
+
+static unsigned char new_key[] = { 0xA7,0xC3,0x12,0xAA,0xAA,0x12,0xC3,0xA7 };
+static void copy_four (unsigned char * dest, UINT32 in);
+
+int sh_util_set_newkey (char * new)
+{
+  int i, j = 0;
+  int len;
+  SL_TICKET fp;
+  SL_TICKET fout;
+  char * key;
+  char * path;
+  char * outpath;
+  unsigned char * image = NULL;
+  long s = 0;
+  long ilen = 0;
+  long k    = 0;
+  UINT32    * h1;
+
+  if (sl_is_suid())
+    {
+      fprintf(stderr, _("ERROR: insufficient privilege\n"));
+      _exit (EXIT_FAILURE);
+      return -1;  /* braindead MAC OSX compiler needs this */
+    }
+        
+  if (new == NULL || new[0] == '\0')
+    {
+      fprintf(stderr, 
+	      _("ERROR: no key given\n Argument must be 'key@path'\n"));
+      _exit (EXIT_FAILURE);
+      return -1;
+    }
+  key = new;
+  len = strlen(new);
+  for (i = 1; i < (len-2); ++i)
+    {
+      if (new[i] == '@' && new[i+1] == '/')
+	{
+	  j = i+1; new[i] = '\0'; break;
+	}
+    }
+  if (j == 0)
+    {
+      fprintf(stderr, 
+	      _("ERROR: no path to executable given\n Argument must be 'key@path'\n"));
+      _exit (EXIT_FAILURE);
+      return -1;
+    }
+  else
+    path = &new[j];
+
+  if (NULL == (outpath = malloc(strlen(path) + 1 + 4)))
+    goto bail_mem;
+  sprintf (outpath, _("%s.out"), path);               /* known to fit  */ 
+
+  fp = sl_open_read(path, SL_NOPRIV);
+  if (SL_ISERROR(fp))
+    {
+      fprintf(stderr, 
+	      _("ERROR: cannot open %s for read\n"), path);
+      _exit (EXIT_FAILURE);
+      return -1;
+    }
+  
+  fout = sl_open_write(outpath, SL_NOPRIV);
+  if (SL_ISERROR(fout))
+    {
+      fprintf(stderr, 
+	      _("ERROR: cannot open %s\n"), outpath);
+      _exit (EXIT_FAILURE);
+      return -1;
+    }
+
+
+  image = malloc (4096);
+  if (!image)
+    goto bail_mem;
+  while (0 < (i = sl_read (fp, &image[s], 4096)))
+    {
+      ilen += i;
+      s    += 4096;
+      image = realloc (image, 4096 + s);
+      if (!image)
+	goto bail_mem;
+    }
+
+  printf(_("%ld bytes read\n"), ilen);
+
+  
+  for (k = 0; k < (ilen - 8); ++k) 
+    {
+      if (image[k]   == new_key[0] &&
+	  image[k+1] == new_key[1] &&
+	  image[k+2] == new_key[2] &&
+	  image[k+3] == new_key[3] &&
+	  image[k+4] == new_key[4] &&
+	  image[k+5] == new_key[5] &&
+	  image[k+6] == new_key[6] &&
+	  image[k+7] == new_key[7])
+	{
+	  printf(_("old key found\n")); 
+	  h1 = sh_tiger_hash_uint32 (key, TIGER_DATA, strlen(key));
+	  copy_four( (unsigned char *) &(image[k]),   h1[0]);
+	  copy_four( (unsigned char *) &(image[k+4]), h1[1]);
+	  sl_write (fout, image, ilen);
+	  sl_close (fout);
+	  printf(_("new file %s written\n"), outpath);
+	  _exit (EXIT_SUCCESS);
+	  return 0;
+	}
+    }
+
+  fprintf(stderr, 
+	  _("ERROR: old key not found\n"));
+  _exit (EXIT_FAILURE);
+  return -1;
+
+
+ bail_mem:
+  fprintf(stderr, 
+	  _("ERROR: out of memory\n"));
+  _exit (EXIT_FAILURE);
+  return -1;
+}
+
   
 
-/* An even simpler en-/decoder, based on Vernam cipher. We use the
+	
+/* A simple en-/decoder, based on Vernam cipher. We use the
  * message as salt to hide the key by obtaining a different one-time 
  * pad each time.
  * Should be safe against a listener on the network, but not against someone
  * with read access to the binary.
  */
-void sh_util_encode (char * data, char * salt, int mode)
+void sh_util_encode (char * data, char * salt, int mode, char fill)
 {
   static char     cc1[17] = N_("0123456789ABCDEF");
   char            cc[17] = "\0";
@@ -726,22 +967,28 @@ void sh_util_encode (char * data, char * salt, int mode)
    */
   sl_strlcpy( cc, _(cc1), sizeof(cc));
 
-  /* 64 bits keyspace
+  /* max 128 bits keyspace
    */
-  memset (skey->vernam, 0, KEY_LEN+1);
+  memset (skey->vernam, fill, KEY_LEN+1);
 
   dez    = (char *) &(skey->ErrFlag[0]);
   sh_util_cpylong (skey->vernam,     dez, 4);
   dez    = (char *) &(skey->ErrFlag[1]);
   sh_util_cpylong (&skey->vernam[4], dez, 4);
 
-  skey->vernam[8] = '\0';
-
-  sl_strlcat(&skey->vernam[8], salt, KEY_LEN+1-8 );
   skey->vernam[KEY_LEN] = '\0';
+
   sl_strlcpy(skey->vernam, 
 	     sh_tiger_hash(skey->vernam, TIGER_DATA, KEY_LEN), KEY_LEN+1);
 
+  sl_strlcpy(skey->vernam, 
+	     sh_util_hmac_tiger (skey->vernam, salt,    strlen(salt)),
+	     KEY_LEN+1);
+
+  sl_strlcpy(skey->vernam, 
+	     sh_util_hmac_tiger (skey->vernam, (char*) new_key, 8),
+	     KEY_LEN+1);
+
   /* The following routine adds/subtracts  data[j] and vernam[j] mod 16.
    */
   j = 0;
@@ -897,14 +1144,17 @@ int sh_util_keyinit (char * buf, long size)
 		     6*sizeof(UINT32));
   p[size-1] = '\0';
 
-  i = sl_strlcpy(buf, 
+  /*
+  i = sl_strlcpy(buf,
 		 sh_tiger_hash ((char *) bufy, TIGER_DATA, 
 				6*sizeof(UINT32)), 
 		 size);
+  */
+  i = sl_strlcpy(buf, p, size);
 
   memset (bufy, '\0', 6*sizeof(UINT32));
 
-  if (status == 0 && (!SL_ISERROR(i) || (i == SL_ETRUNC)) )
+  if ((status == 0) && (!SL_ISERROR(i)) )
     SL_RETURN((0),_("sh_util_keyinit"));
 
   if (SL_ISERROR(i))
@@ -915,7 +1165,7 @@ int sh_util_keyinit (char * buf, long size)
 }
 
 #if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)
-int sh_util_obscurename (ShErrLevel level, char * name_orig)
+int sh_util_obscurename (ShErrLevel level, char * name_orig, int flag)
 {
   char * name = name_orig;
   char * safe;
@@ -928,15 +1178,19 @@ int sh_util_obscurename (ShErrLevel level, char * name_orig)
    */
   while (*name) 
     {
-      if ( (*name) == ' '  || (*name) == '"'  || 
+      if ( (*name) == '"'  || (*name) == '\t' ||
 	   (*name) == '\b' || (*name) == '\f' || 
-	   (*name) == '\n' || (*name) == '\r' || (*name) == '\t' ||
-	   (*name) == '\v' || iscntrl((int) *name) || !isgraph ((int) *name)) 
+	   (*name) == '\n' || (*name) == '\r' ||
+	   (*name) == '\v' || iscntrl((int) *name) || 
+	   ((*name) != ' ' && !isgraph ((int) *name)) ) 
 	{
-	  safe = sh_util_safe_name (name_orig);  
-	  sh_error_handle (level, FIL__, __LINE__, 0, MSG_FI_OBSC, 
-			   safe);
-	  SH_FREE(safe);
+	  if (flag == S_TRUE)
+	    {
+	      safe = sh_util_safe_name (name_orig);  
+	      sh_error_handle (level, FIL__, __LINE__, 0, MSG_FI_OBSC, 
+			       safe);
+	      SH_FREE(safe);
+	    }
 	  SL_RETURN((-1),_("sh_util_obscurename"));
 	}
       name++;
@@ -1012,9 +1266,24 @@ char * sh_util_safe_name (const char * name)
 
   SL_ENTER(_("sh_util_safe_name"));
 
-  ASSERT_RET ((name != NULL), _("name != NULL"), _("NULL"))
+  if (name == NULL)
+    {
+      /* return an allocated array
+       */
+      retval = SH_ALLOC(7);
+      sl_strlcpy(retval, _("(null)"), 7);
+      SL_RETURN(retval, _("sh_util_safe_name"));
+    }
 
-  retval = SH_ALLOC(4 * sl_strlen(name) + 2); 
+  /*
+  ASSERT_RET ((name != NULL), _("name != NULL"), _("NULL"))
+  */
+
+#ifdef SH_USE_XML
+  retval = SH_ALLOC(6 * sl_strlen(name) + 2);
+#else
+  retval = SH_ALLOC(4 * sl_strlen(name) + 2);
+#endif 
 
   p = name;
 
@@ -1040,12 +1309,43 @@ char * sh_util_safe_name (const char * name)
     } else if ( (*p) == '\f') {    /* form-feed        */
       retval[i] = '\\'; ++i; 
       retval[i] = 'f';
+#ifdef WITH_DATABASE
+    } else if ( (*p) == '\'') {    /* single quote     */
+      retval[i] = '\\'; ++i; 
+      retval[i] = '\'';
+#endif
     } else if ( (*p) == ' ') {     /* space            */
       retval[i] = '\\'; ++i; 
       retval[i] = ' ';
+#ifdef SH_USE_XML
+    } else if ( (*p) == '"') {     /* double quote     */
+      retval[i] = '&'; ++i; 
+      retval[i] = 'q'; ++i;
+      retval[i] = 'u'; ++i;
+      retval[i] = 'o'; ++i;
+      retval[i] = 't'; ++i;
+      retval[i] = ';';
+    } else if ( (*p) == '&') {     /* ampersand        */
+      retval[i] = '&'; ++i; 
+      retval[i] = 'a'; ++i;
+      retval[i] = 'm'; ++i;
+      retval[i] = 'p'; ++i;
+      retval[i] = ';';
+    } else if ( (*p) == '<') {     /* left angle       */
+      retval[i] = '&'; ++i; 
+      retval[i] = 'l'; ++i;
+      retval[i] = 't'; ++i;
+      retval[i] = ';';
+    } else if ( (*p) == '>') {     /* right angle      */
+      retval[i] = '&'; ++i; 
+      retval[i] = 'g'; ++i;
+      retval[i] = 't'; ++i;
+      retval[i] = ';';
+#else
     } else if ( (*p) == '"') {     /* double quote     */
       retval[i] = '\\'; ++i; 
       retval[i] = '\"';
+#endif
     } else if (!isgraph ((int) *p)) {    /* not printable    */
       sprintf(oct, _("%c%03o"), '\\',                 /* known to fit  */
 	      (unsigned char) *p);
@@ -1118,6 +1418,7 @@ char * sh_util_strconcat (const char * arg1, ...)
   SL_RETURN(strnew, _("sh_util_strconcat"));
 }
 
+
 #ifdef HAVE_REGEX_H
 
 #include <regex.h>
