@@ -27,6 +27,12 @@
 #include <fcntl.h>
 #include <errno.h>
 
+#ifdef HAVE_UTADDR
+#include <sys/socket.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
+#endif
+
 #ifdef SH_USE_UTMP
 
 #ifdef HAVE_UTMPX_H
@@ -98,13 +104,27 @@
 #endif 
 
 #ifndef UT_LINESIZE
-#define UT_LINESIZE           12
+#ifndef __UT_LINESIZE
+#define UT_LINESIZE            12
+#else
+#define UT_LINESIZE __UT_LINESIZE
+#endif 
 #endif
+
 #ifndef UT_NAMESIZE
-#define UT_NAMESIZE            8
+#ifndef __UT_NAMESIZE
+#define UT_NAMESIZE             8
+#else
+#define UT_NAMESIZE __UT_NAMESIZE
 #endif
+#endif
+
 #ifndef UT_HOSTSIZE
-#define UT_HOSTSIZE           16
+#ifndef __UT_HOSTSIZE
+#define UT_HOSTSIZE            16
+#else
+#define UT_HOSTSIZE __UT_HOSTSIZE
+#endif
 #endif
 
 #ifdef HAVE_UTMPX_H
@@ -146,8 +166,9 @@
 typedef struct log_user {
   char                ut_tty[UT_LINESIZE+1];    
   char                name[UT_NAMESIZE+1];
-  char                ut_host[UT_HOSTSIZE+1]; 
-  time_t              time; 
+  char                ut_host[UT_HOSTSIZE+1];
+  char                ut_ship[16]; /* IP address */
+  time_t              time;
   struct log_user   * next;
 } blah_utmp;
 
@@ -162,10 +183,10 @@ static void sh_utmp_login_morechecks(struct SH_UTMP_S * ut);
 static void sh_utmp_addlogin (struct SH_UTMP_S * ut);
 static void sh_utmp_check_internal(int mode);
 
-int    ShUtmpLoginSolo    = 7;
-int    ShUtmpLoginMulti   = 7;
-int    ShUtmpLogout       = 7;
-int    ShUtmpActive       = BAD;
+int    ShUtmpLoginSolo    = SH_ERR_INFO;
+int    ShUtmpLoginMulti   = SH_ERR_WARN;
+int    ShUtmpLogout       = SH_ERR_INFO;
+int    ShUtmpActive       = S_TRUE;
 time_t ShUtmpInterval     = 300;
 
 sh_rconf sh_utmp_table[] = {
@@ -195,6 +216,11 @@ sh_rconf sh_utmp_table[] = {
   },
 };
 
+int sh_utmp_null()
+{
+  return 0;
+}
+
 
 
 
@@ -236,6 +262,7 @@ static long   sh_utmp_feed_forward = 0;
 
 static void sh_utmp_utmpname(const char * str)
 {
+  SL_ENTER(_("sh_utmp_utmpname"));
   if (sh_utmpfile != NULL)
     {
       MBLK( fclose (sh_utmpfile); )
@@ -243,7 +270,7 @@ static void sh_utmp_utmpname(const char * str)
     }
 
   sl_strlcpy (sh_utmppath, str, 80);
-  return;
+  SL_RET0(_("sh_utmp_utmpname"));
 }
 
 static void sh_utmp_setutent(void)
@@ -256,7 +283,7 @@ static void sh_utmp_setutent(void)
   ASSERT((sh_utmppath != NULL), _("sh_utmppath != NULL"));
 
   if (sh_utmppath == NULL)
-    return;
+    SL_RET0(_("sh_utmp_setutent"));
 
   if (sh_utmpfile == NULL) 
     {
@@ -285,9 +312,10 @@ static void sh_utmp_setutent(void)
 
 static void sh_utmp_endutent(void)
 {
+  SL_ENTER(_("sh_utmp_endutent"));
   MBLK( fclose(sh_utmpfile); )
   sh_utmpfile = NULL;
-  return;
+  SL_RET0(_("sh_utmp_endutent"));
 }
 
 static struct SH_UTMP_S * sh_utmp_getutent(void)
@@ -376,6 +404,34 @@ static struct SH_UTMP_S * sh_utmp_getutid(struct SH_UTMP_S * ut)
 /* #ifdef HAVE_SETUTENT */
 #endif
 
+#ifdef HAVE_UTADDR
+#ifdef HAVE_INET_ATON
+char * my_inet_ntoa(struct in_addr in)
+{
+  return inet_ntoa(in);
+}
+#else
+char * my_inet_ntoa(struct in_addr in)
+{
+  unsigned char a, b, c, d;
+  static char   foo[16];
+  char          bar[4];
+  memcpy (bar, &(in.s_addr), 4); /* memory alignment (?) */
+  memcpy (&a, &bar[0], 1);
+  memcpy (&b, &bar[1], 1);
+  memcpy (&c, &bar[2], 1);
+  memcpy (&d, &bar[3], 1);
+  sprintf(foo, "%d.%d.%d.%d",                          /* known to fit  */
+	  (int) a, (int) b, (int) c, (int) d);
+  return foo;
+}
+#endif
+/* #ifdef HAVE_UTADDR */
+#endif
+
+#if defined(__linux__) && !defined(ut_addr)
+#define ut_addr         ut_addr_v6[0]
+#endif
 
 
 static struct log_user   * userlist   = NULL;
@@ -388,15 +444,24 @@ static time_t  lastcheck;
  *************/
 int sh_utmp_init ()
 {
-  if (ShUtmpActive == BAD)
-    return (-1);
+  static int done = 0;
 
+  SL_ENTER(_("sh_utmp_init"));
+  if (ShUtmpActive == BAD)
+    SL_RETURN( (-1), _("sh_utmp_init"));
+
+  /* do not re-initialize after a re-configuration
+   */
+  if (done == 1) {
+    SL_RETURN( (0), _("sh_utmp_init"));
+  }
   lastcheck  = time (NULL);
   userlist   = NULL;
   memset (&save_utmp, 0, sizeof(struct SH_UTMP_S));
   sh_utmp_check_internal (2); /* current logins */
   sh_utmp_check_internal (0);
-  return (0);
+  done = 1;
+  SL_RETURN( (0), _("sh_utmp_init"));
 }
 
 /*************
@@ -404,17 +469,27 @@ int sh_utmp_init ()
  * module cleanup
  *
  *************/
+#ifdef HAVE_UTTYPE
+static int sh_utmp_login_clean();
+#endif
+
 int sh_utmp_end ()
 {
   struct log_user * user    = userlist;
   struct log_user * userold;
+
+  SL_ENTER(_("sh_utmp_end"));
   while (user)
     {
       userold = user;
       user    = user->next;
       SH_FREE(userold);
     }
-  return (0);
+  userlist = NULL;
+#ifdef HAVE_UTTYPE
+  sh_utmp_login_clean();
+#endif
+  SL_RETURN( (0), _("sh_utmp_end"));
 }
 
 
@@ -425,7 +500,7 @@ int sh_utmp_end ()
  *************/
 int sh_utmp_timer (unsigned long tcurrent)
 {
-  if ((time_t) (tcurrent - lastcheck) > ShUtmpInterval)
+  if ((time_t) (tcurrent - lastcheck) >= ShUtmpInterval)
     {
       lastcheck  = tcurrent;
       return (-1);
@@ -440,10 +515,11 @@ int sh_utmp_timer (unsigned long tcurrent)
  *************/
 int sh_utmp_check ()
 {
+  SL_ENTER(_("sh_utmp_check"));
   sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_UT_CHECK);
   sh_utmp_check_internal (1);
 
-  return 0;
+  SL_RETURN(0, _("sh_utmp_check"));
 }
 
 /*************
@@ -455,34 +531,41 @@ int sh_utmp_check ()
 int sh_utmp_set_login_solo  (char * c)
 {
   char tmp[32];
+
+  SL_ENTER(_("sh_utmp_set_login_solo"));
   tmp[0] = '='; tmp[1] = '\0';
   sl_strlcat (tmp, c, 32);
   sh_error_set_level (tmp, &ShUtmpLoginSolo);
-  return 0;
+  SL_RETURN(0, _("sh_utmp_set_login_solo"));
 }
 
 int sh_utmp_set_login_multi (char * c)
 {
   char tmp[32];
+
+  SL_ENTER(_("sh_utmp_set_login_multi"));
   tmp[0] = '='; tmp[1] = '\0';
   sl_strlcat (tmp, c, 32);
   sh_error_set_level (tmp, &ShUtmpLoginMulti);
-  return 0;
+  SL_RETURN(0, _("sh_utmp_set_login_multi"));
 }
 
 int sh_utmp_set_logout_good (char * c)
 {
   char tmp[32];
+
+  SL_ENTER(_("sh_utmp_set_logout_good"));
   tmp[0] = '='; tmp[1] = '\0';
   sl_strlcat (tmp, c, 32);
   sh_error_set_level (tmp, &ShUtmpLogout);
-  return 0;
+  SL_RETURN(0, _("sh_utmp_set_logout_good"));
 }
 
 int sh_utmp_set_login_timer (char * c)
 {
   long val;
 
+  SL_ENTER(_("sh_utmp_set_login_timer"));
   val = strtol (c, (char **)NULL, 10);
   if (val <= 0)
     sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS,
@@ -491,32 +574,96 @@ int sh_utmp_set_login_timer (char * c)
   val = (val <= 0 ? 60 : val);
 
   ShUtmpInterval = (time_t) val;
-  return 0;
+  SL_RETURN(0, _("sh_utmp_set_login_timer"));
 }
 
 int sh_utmp_set_login_activate (char * c)
 {
-  if (c != NULL)
-    {
-      if (c[0] == '0' || c[0] == 'n' || c[0] == 'N')
-	{
-	  ShUtmpActive = BAD;
-	  return (0);
-	}
-      else if (c[0] == '1' || c[0] == 'y' || c[0] == 'Y')
-	{
-	  ShUtmpActive = GOOD;
-	  return (0);
-	}
-    }
-
-  sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS,
-                      _("utmp activate"), c);
-
-  ShUtmpActive = BAD;
-  return (-1);
+  int i;
+  SL_ENTER(_("sh_utmp_set_login_activate"));
+  i = sh_util_flagval(c, &ShUtmpActive);
+  SL_RETURN(i, _("sh_utmp_set_login_activate"));
 }
 
+#ifdef HAVE_UTTYPE
+struct login_ct {
+  char name[UT_NAMESIZE+1];
+  int  nlogin;
+  struct login_ct * next;
+};
+
+static struct login_ct * login_ct_list = NULL;
+
+static int sh_utmp_login_clean()
+{
+  struct login_ct * list = login_ct_list;
+  struct login_ct * old;
+
+  login_ct_list = NULL;
+
+  while (list)
+    {
+      old  = list;
+      list = list->next;
+      SH_FREE(old);
+    }
+  return 0;
+}
+
+static int sh_utmp_login_a(char * str)
+{
+  struct login_ct * list = login_ct_list;
+
+  while (list)
+    {
+      if (0 == sl_strcmp(list->name, str))
+	{
+	  ++(list->nlogin);
+	  return list->nlogin;
+	}
+      list = list->next;
+    }
+  list = SH_ALLOC(sizeof(struct login_ct));
+  sl_strlcpy(list->name, str, UT_NAMESIZE+1);
+  list->nlogin  = 1;
+  list->next    = login_ct_list;
+  login_ct_list = list;
+  return 1;
+}
+
+static int sh_utmp_login_r(char * str)
+{
+  struct login_ct * list = login_ct_list;
+  struct login_ct * old  = login_ct_list;
+
+  while (list)
+    {
+      if (0 == sl_strcmp(list->name, str))
+	{
+	  list->nlogin -= 1;
+	  if (list->nlogin > 0)
+	    {
+	      return list->nlogin;
+	    }
+	  if (old == list)
+	    {
+	      login_ct_list = NULL;
+	      SH_FREE(list);
+	    }
+	  else
+	    {
+	      old->next = list->next;
+	      SH_FREE(list);
+	    }
+	  return 0;
+	}
+      old  = list;
+      list = list->next;
+    }
+  return 0;
+}
+
+#endif
 
 
 /* for each login:
@@ -524,6 +671,27 @@ int sh_utmp_set_login_activate (char * c)
  *    - link device.ut_record -> log_record
  *    - link user.ut_record   -> log_record
  */
+#ifdef HAVE_UTTYPE  
+static int sh_utmp_is_virtual (char * in_utline, char * in_uthost)
+{
+  if (in_utline != NULL)
+    {
+      if (0 == sl_strncmp(in_utline, _("ttyp"), 4))
+	{ return 0; }
+      else if (0 == sl_strncmp(in_utline, _("ttyq"), 4))
+	{ return 0; }
+    }
+  if (in_uthost != NULL   && 
+      in_uthost[0] == ':' && 
+      in_uthost[1] == '0' && 
+      in_uthost[2] == '\0')
+    {
+      /* return 0; */
+      return 1;  /* actually does not seem portable, not even among linuxes */
+    }
+  return 1;
+}
+#endif
 
 static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 {
@@ -532,9 +700,14 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
   struct log_user   * userold  = userlist;
 
   char   ttt[TIM_MAX];
+#ifdef HAVE_UTTYPE
+  int    status;
+#endif
+
+  SL_ENTER(_("sh_utmp_addlogin"));
 
   if (ut->ut_line[0] == '\0')
-    return;
+    SL_RET0(_("sh_utmp_addlogin"));
 
   /* for some stupid reason, AIX repeats the wtmp entry for logouts
    * with ssh
@@ -542,7 +715,7 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
   if (memcmp (&save_utmp, ut, sizeof(struct SH_UTMP_S)) == 0)
     {
       memset(&save_utmp, '\0', sizeof(struct SH_UTMP_S));
-      return;
+      SL_RET0(_("sh_utmp_addlogin"));
     }
   memcpy (&save_utmp, ut, sizeof(struct SH_UTMP_S));
       
@@ -579,50 +752,67 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
       sl_strlcpy(user->name,    ut->ut_name, UT_NAMESIZE+1);
 #ifdef HAVE_UTHOST
       sl_strlcpy(user->ut_host, ut->ut_host, UT_HOSTSIZE+1);
+#else
+      user->ut_host[0] = '\0';
+#endif
+#ifdef HAVE_UTADDR
+      sl_strlcpy(user->ut_ship,my_inet_ntoa(*(struct in_addr*)&(ut->ut_addr)),16);
 #endif
       user->time = ut->ut_time;
 
-
       if (username == NULL                              /* not yet logged in */
-	  || 0 == sl_strncmp(ut->ut_line, _("ttyp"), 4) /* in virt. console  */
-	  || 0 == sl_strncmp(ut->ut_line, _("ttyq"), 4) /* in virt. console  */
+          || 0 == sl_strncmp(ut->ut_line, _("ttyp"), 4) /* in virt. console  */
+          || 0 == sl_strncmp(ut->ut_line, _("ttyq"), 4) /* in virt. console  */
 	  ) {
+	status = sh_utmp_login_a(user->name);
 	sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX);
 	sh_error_handle( ShUtmpLoginSolo, FIL__, __LINE__, 0,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+			 MSG_UT_LG1X,
+#elif defined(HAVE_UTHOST)
 			 MSG_UT_LG1A,
 #else
 			 MSG_UT_LG1B,
 #endif
 			 user->name,
 			 user->ut_tty,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+			 user->ut_host,
+			 user->ut_ship,
+#elif defined(HAVE_UTHOST)
 			 user->ut_host,
 #endif
-			 ttt
+			 ttt,
+			 status
 			 );
       } else
-	if (0 != sl_strncmp(ut->ut_line, _("ttyp"), 4) &&
-	    0 != sl_strncmp(ut->ut_line, _("ttyq"), 4))
+	if (0 != sh_utmp_is_virtual(ut->ut_line, user->ut_host))
 	{       
+	status = sh_utmp_login_a(user->name);
 	sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX);
 	sh_error_handle( ShUtmpLoginMulti, FIL__, __LINE__, 0,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+			 MSG_UT_LG2X,
+#elif defined(HAVE_UTHOST)
 			 MSG_UT_LG2A,
 #else
 			 MSG_UT_LG2B,
 #endif
 			 user->name,
 			 user->ut_tty,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+			 user->ut_host,
+			 user->ut_ship,
+#elif defined(HAVE_UTHOST)
 			 user->ut_host,
 #endif
-			 ttt
+			 ttt,
+			 status
 			 );
       }
       
       sh_utmp_login_morechecks(ut);
-      return;
+      SL_RET0(_("sh_utmp_addlogin"));
     }
 
 
@@ -633,19 +823,26 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
     {
       if (user != NULL)
 	{
+	  status = sh_utmp_login_r(user->name);
 	  sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX);
 	  sh_error_handle( ShUtmpLogout, FIL__, __LINE__, 0,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+			   MSG_UT_LG3X,
+#elif defined(HAVE_UTHOST)
 			   MSG_UT_LG3A,
 #else
 			   MSG_UT_LG3B,
 #endif
 			   user->name,
 			   user->ut_tty,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+			   user->ut_host,
+			   user->ut_ship,
+#elif defined(HAVE_UTHOST)
 			   user->ut_host,
 #endif
-			   ttt
+			   ttt,
+			   status
 			   );
 	  userold->next = user->next;
 	  if (user == userlist)
@@ -661,14 +858,14 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 	  sh_error_handle( ShUtmpLogout, FIL__, __LINE__, 0,
 			   MSG_UT_LG3C,
 			   terminated_line,
-			   ttt
+			   ttt, 0
 			   );
 	}
-      return;
+      SL_RET0(_("sh_utmp_addlogin"));
     }
 
   /* default */
-  return;
+  SL_RET0(_("sh_utmp_addlogin"));
 
   /* #ifdef HAVE_UTTYPE                   */
 #else
@@ -680,6 +877,9 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
       sl_strlcpy(user->name,    ut->ut_name, UT_NAMESIZE+1);
 #ifdef HAVE_UTHOST
       sl_strlcpy(user->ut_host, ut->ut_host, UT_HOSTSIZE+1);
+#endif
+#ifdef HAVE_UTADDR
+      sl_strlcpy(user->ut_ship,my_inet_ntoa((struct in_addr)ut->ut_addr),16);
 #endif
       user->time       = ut->ut_time;
       user->next       = userlist;
@@ -689,17 +889,23 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
 
 
       sh_error_handle( ShUtmpLoginSolo, FIL__, __LINE__, 0,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+		       MSG_UT_LG1X,
+#elif defined(HAVE_UTHOST)
 		       MSG_UT_LG1A,
 #else
 		       MSG_UT_LG1B,
 #endif
 		       user->name,
 		       user->ut_tty,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+		       user->ut_host,
+		       user->ut_ship,
+#elif defined(HAVE_UTHOST)
 		       user->ut_host,
 #endif
-		       ttt
+		       ttt,
+		       1
 		       );
       sh_utmp_login_morechecks(ut);
     }
@@ -708,17 +914,23 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
       sl_strlcpy(ttt, sh_unix_time (user->time), TIM_MAX);
 
       sh_error_handle( ShUtmpLogout, FIL__, __LINE__, 0,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+		       MSG_UT_LG2X,
+#elif defined(HAVE_UTHOST)
 		       MSG_UT_LG2A,
 #else
 		       MSG_UT_LG2B,
 #endif
 		       user->name,
 		       user->ut_tty,
-#ifdef HAVE_UTHOST
+#if defined(HAVE_UTHOST) && defined(HAVE_UTADDR)
+		       user->ut_host,
+		       user->ut_ship,
+#elif defined(HAVE_UTHOST)
 		       user->ut_host,
 #endif
-		       ttt
+		       ttt,
+		       1
 		       );
       sh_utmp_logout_morechecks(user);
       userold->next = user->next;
@@ -726,7 +938,7 @@ static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
       user = NULL;
     }
 
-  return;
+  SL_RET0(_("sh_utmp_addlogin"));
 #endif
 }
 
@@ -743,14 +955,16 @@ static void sh_utmp_check_internal (int mode)
   struct SH_UTMP_S * ut;
   unsigned long this_read = 0;
 
+  SL_ENTER(_("sh_utmp_check_internal"));
+
   /* error if no access
    */
-  if (0 != retry_lstat(mode_path[mode], &buf)) 
+  if (0 != retry_lstat(FIL__, __LINE__, mode_path[mode], &buf)) 
     {
       error = errno;
       sh_error_handle((-1), FIL__, __LINE__, error, MSG_E_ACCESS,
 		      (long) sh.real.uid, mode_path[mode]);
-      return;
+      SL_RET0(_("sh_utmp_check_internal"));
     }
 
   /* modification time
@@ -758,7 +972,7 @@ static void sh_utmp_check_internal (int mode)
   if (mode < 2)
     {
       if (buf.st_mtime <= lastmod) 
-	return;
+	SL_RET0(_("sh_utmp_check_internal"));
       else
 	lastmod = buf.st_mtime;
     }
@@ -779,7 +993,7 @@ static void sh_utmp_check_internal (int mode)
     lastsize = buf.st_size;
 
   if (buf.st_size == 0) 
-    return;
+    SL_RET0(_("sh_utmp_check_internal"));
 
   sh_utmp_utmpname(mode_path[mode]);
   sh_utmp_setutent();
@@ -811,7 +1025,7 @@ static void sh_utmp_check_internal (int mode)
 		      ))
       sh_utmp_addlogin (ut);
     /*************************************************
-    printf("%8s | %10s | %10s | %d %d | %16s | %ld\n", 
+    printf("%8s | %10s | %10s | %3d %5d | %16s | %ld\n", 
 	   ut->ut_name, ut->ut_id, ut->ut_line, 
 	   (int) ut->ut_type, (int) ut->ut_pid, 
 	   ut->ut_host, ut->ut_time);
@@ -830,7 +1044,7 @@ static void sh_utmp_check_internal (int mode)
 #endif
     }
 
-  return;
+  SL_RET0(_("sh_utmp_check_internal"));
 }
 
 
@@ -841,7 +1055,7 @@ static void sh_utmp_login_morechecks(struct SH_UTMP_S * ut)
   return;
 }
 
-static void sh_utmp_logout_morechecks(struct log_user   * user)
+static void sh_utmp_logout_morechecks(struct log_user * user)
 {
   if (user)
     return;
