@@ -3,6 +3,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
 #include <string.h>
 
 #include <unistd.h>
@@ -17,6 +18,37 @@
 #undef  FIL__
 #define FIL__  _("slib.c")
 
+const uid_t sh_uid_neg = ((uid_t) -1);
+const gid_t sh_gid_neg = ((gid_t) -1);
+ 
+#undef BREAKEXIT
+#if defined(SCREW_IT_UP) && defined(__linux__) && defined(__i386__)
+
+#ifdef SH_DEBUG
+#define BREAKEXIT(expr) \
+  do { \
+    int ixi; \
+    for (ixi = 0; ixi < 8; ++ixi) { \
+      if ((*(volatile unsigned *)((unsigned) expr + ixi) & 0xff) == 0xcc)  \
+        { dlog(0, FIL__, __LINE__, _("BREAKEXIT")); _exit(EXIT_FAILURE); } \
+      } \
+    } \
+  while (1 == 0)
+#else
+#define BREAKEXIT(expr) \
+  do { \
+    int ixi; \
+    for (ixi = 0; ixi < 8; ++ixi) { \
+      if ((*(volatile unsigned *)((unsigned) expr + ixi) & 0xff) == 0xcc) \
+        _exit(EXIT_FAILURE); \
+      } \
+    } \
+  while (1 == 0)
+#endif
+
+#else
+#define BREAKEXIT(expr)
+#endif
 
 /****************************************************************
  *
@@ -24,195 +56,269 @@
  *
  ****************************************************************/
 
-#if defined(WITH_TRACE) || defined(WITH_TPT)
+int slib_do_trace          = 0;
+int slib_trace_fd          = -1;
 
-#define ALLE_STEINE -1
+static char trace_log[256] = { '\0' };
+static int trace_level     = 0;
+static FILE * trace_fp     = NULL;
 
-#if !defined(WITH_TRACE)
-int stein_set = ALLE_STEINE;
-#else
-int stein_set = -99;
-#endif
-
-int stein     =  0;
-int ebene     =  0;
-int mindest   = -1;
-
-static int use_console = 0;
-void dbg_use_console()
+int  sl_trace_use (char * dummy)
 {
-  use_console = 1;
-  return;
+  if (dummy)
+    slib_do_trace = 1;
+  else
+    slib_do_trace = 1;
+  return 0;
 }
 
-void dbg(int mod_ebene, char * file, int line, char * fmt, ...)
+int  sl_trace_file (char * str)
 {
-  FILE * dfile;
-  va_list ap;
-  char indent_it[32];
-  int  i, imax;
+  if (!str)
+    return -1;
+  sl_strlcpy(trace_log, str, 256);
+  return 0;
+}
 
-  if (ebene < 1) 
-    ++stein;
+FILE * sl_tracefile_open(char * file, char * mode)
+{
+  FILE * xp = NULL;
+  slib_trace_fd = open(file, O_WRONLY|O_CREAT|O_APPEND, 0600);
+  if (slib_trace_fd >= 0)
+    xp = fdopen(slib_trace_fd, mode);
+  return xp;
+}
 
-  if (mod_ebene > 0)
-    ebene += mod_ebene;
-
-  if (ebene < 31)
-    imax = ebene;
-  else
-    imax = 31;
-
-  /*
-  fprintf (stderr, _("DBG: file=<%s>, line=<%d> %d %d \n"), 
-		 file, line, stein_set, mindest);
-  */
-
-  for (i = 0; i < imax; ++i)
-    indent_it[i] = ' ';
-  indent_it[imax] = '\0';
-
-  if ( (stein == stein_set || stein_set == ALLE_STEINE ) && ebene <= mindest)
+void sl_trace_in(char * str, char * file, int line)
+{
+  int    i;
+  if (trace_log[0] == '\0')
     {
-      if (ebene < 1)
-#if !defined(WITH_TRACE)
+      fprintf(stderr, "++ ");
+      for (i = 0; i < trace_level; ++i)
+	fprintf(stderr, ".  ");
+      fprintf(stderr, "[%2d] %s \t - File %c%s%c at line %d\n", 
+	     trace_level, str, 0x22, file, 0x22, line);
+    }
+  else if (!sl_is_suid())
+    {
+      if (!trace_fp)
+	trace_fp = sl_tracefile_open(trace_log, "a");
+      if (trace_fp)
 	{
-	  if (0 == use_console)
-	    {
-	      fprintf (stderr, _("DBG: file=<%s>, line=<%d>, "), 
-		       file, line);
-	      va_start(ap, fmt);
-	      vfprintf  (stderr, fmt, ap);
-	      va_end(ap);
-	    }
-	  else
-	    {
-	      dfile = fopen(_("/dev/console"), "w");
-	      fprintf (dfile, _("DBG: file=<%s>, line=<%d>, "), file, line);
-	      va_start(ap, fmt);
-	      vfprintf  (dfile, fmt, ap);
-	      va_end(ap);
-	      fclose (dfile);
-	    }
+	  fprintf(trace_fp, "++ ");
+	  for (i = 0; i < trace_level; ++i)
+	    fprintf(trace_fp, ".  ");
+	  fprintf(trace_fp, "[%2d] %s \t - File %c%s%c at line %d\n", 
+		 trace_level, str, 0x22, file, 0x22, line);
 	}
-#else
-      {
-	if (0 == use_console)
-	  fprintf (stderr, 
-		   _("MILEST: file=<%s>, line=<%d>, milestone=<%d>, "), 
-		   file, line, stein);
-	else
-	  {
-	    dfile = fopen(_("/dev/console"), "w");
-	    fprintf (dfile, 
-		     _("MILEST: file=<%s>, line=<%d>, milestone=<%d>, "), 
-		     file, line, stein);
-	    va_start(ap, fmt);
-	    vfprintf  (dfile, fmt, ap);
-	    va_end(ap);
-	    fclose (dfile);
-	  }
-      }
-#endif
       else
 	{
-	  if (0 == use_console)
+	  perror(_("sl_trace_in: fopen"));
+	  _exit(1);
+	}
+    }
+  ++trace_level;
+}
+
+void sl_trace_out(char * str, char * file, int line)
+{
+  int    i;
+
+  --trace_level; if (trace_level < 0) trace_level = 0;
+
+  if (trace_log[0] == '\0')
+    {
+      fprintf(stderr, "-- ");
+      for (i = 0; i < trace_level; ++i)
+	fprintf(stderr, ".  ");
+      fprintf(stderr, _("[%2d] %s \t - File %c%s%c at line %d\n"), 
+	     trace_level, str, 0x22, file, 0x22, line);
+    }
+  else if (!sl_is_suid())
+    {
+      if (!trace_fp)
+	trace_fp = sl_tracefile_open(trace_log, "a");
+      if (trace_fp)
+	{
+	  fprintf(trace_fp, "-- ");
+	  for (i = 0; i < trace_level; ++i)
+	    fprintf(trace_fp, ".  ");
+	  fprintf(trace_fp, _("[%2d] %s \t - File %c%s%c at line %d\n"), 
+		 trace_level, str, 0x22, file, 0x22, line);
+	}
+      else
+	{
+	  perror(_("sl_trace_out: fopen"));
+	  _exit(1);
+	}
+    }
+}
+
+extern int sh_log_console (char * msg);
+
+static int dlogActive = 0;
+
+/* this is called from sh_error_setprint()
+ */
+void dlog_set_active(int flag)
+{
+  dlogActive = flag;
+}
+
+/* flag = 0 debug messages
+ *      = 1 descriptive error messages
+ *      = 3 backtrace
+ */
+int dlog (int flag, char * file, int line,  const char *fmt, ...)
+{
+  va_list     ap;
+  char        val[81];
+  char        msg[512];
+  char        tmp[512];
+  int         retval = 0;
+  int         i;
+
+#ifdef SH_STEALTH
+  /* 
+   * do not even print descriptive failure messages in stealth mode
+   */
+  if (dlogActive == 0)
+    return 0;
+  if (dlogActive == 1 && flag == 0) /* debug requires debug level */
+    return 0;
+#else
+  if (dlogActive <= 1 && flag == 0) /* debug requires debug level */
+    return 0;
+#endif
+
+  if (flag == 1)
+    {
+      sprintf        (val, _("\n---------  %10s "), file);
+      sl_strlcpy     (msg,    val,   80);
+      sprintf        (val, _(" --- %6d ---------\n"), line);
+      sl_strlcat     (msg,     val,   80);
+      sh_log_console (msg);
+    }
+
+  va_start (ap, fmt);
+  if (flag == 1)
+    sl_strlcpy(tmp, fmt, 512);
+  else
+    sl_strlcpy(tmp, fmt, 256);
+  retval = sl_strlen(tmp);
+  if (retval > 0 && tmp[retval-1] == '\n')
+    tmp[retval-1] = '\0';
+  retval = 0;
+  if (flag == 1)
+    {
+      sl_vsnprintf (msg, 511, tmp, ap);
+    }
+  else
+    {
+      sl_strlcpy   (msg,    "## ", 256);
+      for (i = 0; i < trace_level; ++i)
+	sl_strlcat (msg, ".  ", 256);
+      sprintf      (val, _("[%2d] "), trace_level);
+      sl_strlcat   (msg,     val,   256);
+      sl_vsnprintf (&msg[sl_strlen(msg)], 255, tmp, ap);
+      sl_snprintf  (tmp, 255, _(" \t - File %c%s%c at line %d"), 
+		    0x22, file, 0x22, line);
+      sl_strlcat   (msg,     tmp,   512);
+    }
+  va_end (ap);
+  if (flag != 0 || sl_is_suid())
+    retval = sh_log_console (msg);
+  else
+    {
+      if (trace_log[0] == '\0')
+	{
+	  /* sh_log_console (msg); */
+	  fprintf(stderr, "%s\n", msg);
+	}
+      else
+	{
+	  if (!trace_fp)
+	    trace_fp = sl_tracefile_open(trace_log, "a");
+	  if (trace_fp)
 	    {
-	      fprintf (stderr, _("%sTRACE : file=<%s>, line=<%d>, "), 
-		       indent_it, file, line);
-	      va_start(ap, fmt);
-	      vfprintf  (stderr, fmt, ap);
-	      va_end(ap);
+	      fprintf(trace_fp, "%s\n", msg);
 	    }
 	  else
 	    {
-	      dfile = fopen(_("/dev/console"), "w");
-	      fprintf (dfile, 
-		       _("%sTRACE : file=<%s>, line=<%d>, "),
-		       indent_it, file, line);
-	      va_start(ap, fmt);
-	      vfprintf  (dfile, fmt, ap);
-	      va_end(ap);
-	      fclose (dfile);
+	      perror(_("dlog: fopen"));
+	      _exit(1);
 	    }
 	}
     }
-
-  if (mod_ebene < 0)
-    ebene += mod_ebene;
-
-  return;
+  if (flag == 1)
+    sh_log_console (_("\n----------------------------------------------\n"));
+  return retval;
 }
 
-int set_stein (char * c)
+extern char aud_err_message[64];
+char * sl_get_errmsg()
 {
-  stein_set = atoi(c);
-  if (stein_set < -1)
-    {
-      stein_set = -99;
-      return (-1);
-    }
-  return 0;
+  return &aud_err_message[0];
 }
-int set_ebene (char * c)
+
+
+#if defined(SL_DEBUG)
+#define SL_MAX_MYSTACK 128
+
+static char sl_mystack[SL_MAX_MYSTACK][32];
+static int  sl_mystack_count = 0; 
+
+void sl_stack_push(char * c, char * file, int line )
 {
-  mindest = atoi(c);
-  if (mindest < 0)
+  if (slib_do_trace)
+    sl_trace_in(c, file, line);
+  if (c && sl_mystack_count < SL_MAX_MYSTACK)
     {
-      mindest = -99;
-      return (-1);
-    }
-  return 0;
-}
-/* WITH_TRACE */
-#endif
-
-/* ---------------------------------------------------------------- 
- *
- *    Free()/Malloc() with block
- *
- * ---------------------------------------------------------------- */
-
-int myblock_flag = 0;
-
-void my_free(void *ptr)
-{
-  if (myblock_flag == 0)
-    {
-      myblock_flag = 1;
-      free(ptr);
-      myblock_flag = 0;
-    }
-  else
-    {
-#ifdef SIGILL
-      raise(SIGILL);       /* SIGILL is POSIX.1  */
-#endif
-      _exit(EXIT_FAILURE);
+      strncpy(sl_mystack[sl_mystack_count], c, 31);
+      sl_mystack[sl_mystack_count][31] = '\0';
+      ++sl_mystack_count;
+      /*
+      fprintf(stderr, "#%03d %s\n", sl_mystack_count, 
+	      sl_mystack[sl_mystack_count-1]);
+      */
     }
   return;
 }
 
-void * my_malloc(size_t size)
+void sl_stack_pop(char * c, char * file, int line)
 {
-  void * ptr = NULL; /* some compilers don't understand _exit() */
-
-  if (myblock_flag == 0)
+  if (slib_do_trace)
+    sl_trace_out(c, file, line);
+  if (sl_mystack_count > 0)
     {
-      myblock_flag = 1;
-      ptr = malloc(size);
-      myblock_flag = 0;
+      /*
+      fprintf(stderr, " <- #%03d %s\n", sl_mystack_count,
+	      sl_mystack[sl_mystack_count-1]);
+      */
+      --sl_mystack_count;
     }
-  else
-    {
-#ifdef SIGILL
-      raise(SIGILL);       /* SIGILL is POSIX.1  */
-#endif
-      _exit(EXIT_FAILURE);
-    }
-  return ptr;
+  return;
 }
-  
+
+void sl_stack_print()
+{
+  int  i;
+  /* FILE * dfile; */
+
+  if (sl_mystack_count > 0)
+    {
+      sh_log_console(_("\nBacktrace:\n"));
+      /* dlog(3, FIL__, __LINE__, _("\nBacktrace:\n")); */
+      for (i = 0; i < sl_mystack_count; ++i)
+	sh_log_console(sl_mystack[i]);
+      /* dlog(3, FIL__, __LINE__, _("#%03d %s\n"), i, sl_mystack[i]); */
+    } 
+  return;
+}
+
+#endif
+
 
 /*
  *  The global errno.
@@ -227,6 +333,22 @@ long int sl_errno;
  *
  * ---------------------------------------------------------------- */
   
+/*
+ * A memset that does not get optimized away
+ */
+void *sl_memset(void *s, int c, size_t n)
+{
+  size_t i;
+  volatile char *p = s;
+
+  if (s == NULL || n <= 0)
+    return s;
+  
+  for (i = 0; i < n; ++i)
+    p[i] = (char) c;
+  return s;
+}
+
 #if !defined(HOST_IS_I86SOLARIS)
 #if !defined (_GNU_SOURCE)
 extern int vsnprintf ( char *str, size_t n,
@@ -246,11 +368,12 @@ extern int vsnprintf ( char *str, size_t n,
 
 #if !defined(HAVE_VSNPRINTF) || defined(HAVE_BROKEN_VSNPRINTF)
 static
-int sl_printf_count (const char * fmt, va_list  vl)
+size_t sl_printf_count (const char * fmt, va_list  vl)
 {
-  int  length       = 1;
+  size_t  length       = 1;
   int  fini         = 0;
   int  islong       = 0;
+  int  islonglong   = 0;
   int  islongdouble = 0;
   char * string_arg;
 
@@ -291,9 +414,16 @@ int sl_printf_count (const char * fmt, va_list  vl)
 	  break;
 
 	case 'l':   /* 'long' modifier */
-	  islong = 1;
+	  if (islong == 0)
+	    islong = 1;
+	  else
+	    {
+	      islonglong = 1;
+	      islong = 0;
+	    }
 	  ++fmt;
 	  break;
+
 	case 'L':  /* 'long double' modifier */ 
 #ifdef HAVE_LONG_DOUBLE	  
 	  islongdouble = 1;
@@ -309,10 +439,18 @@ int sl_printf_count (const char * fmt, va_list  vl)
 	case 'u':
 	case 'x':
 	case 'X':
-	  if (islong == 1)
+	  if (islonglong == 1)
+#ifdef HAVE_LONG_LONG
+	    (void) va_arg (vl, long long);
+#else
+	    (void) va_arg (vl, long);
+#endif
+	  else if (islong == 1)
 	    (void) va_arg (vl, long);
 	  else
 	    (void) va_arg (vl, int);
+	  islong = 0;
+	  islonglong = 0;
 	  length = length + 24;
 	  ++fmt;
 	  fini = 1;
@@ -334,6 +472,7 @@ int sl_printf_count (const char * fmt, va_list  vl)
 #ifdef HAVE_LONG_DOUBLE	  
 	  if (islongdouble == 1) {
 	    (void) va_arg (vl, long double);
+	    islongdouble = 0;
 	    length = length + 20;
 	    }
 	  else
@@ -395,41 +534,42 @@ int sl_printf_count (const char * fmt, va_list  vl)
 /*
  * An implementation of vsnprintf. va_start/va_end are in the caller
  * function.
- * Returns SL_ENONE on success.
- * ENULL:  src || format == NULL
- * ERANGE: n out of range
- * ETRUNC: truncated
+ * Returns C99 (#bytes that would heve been written) on success.
  */
 int sl_vsnprintf(char *str, size_t n,
 		 const char *format, va_list vl )
 {
-  int           total;
+  int len = 0;
 #if !defined(HAVE_VSNPRINTF) || defined(HAVE_BROKEN_VSNPRINTF)
+  size_t         total;
   va_list       vl2;
 #endif
 
   SL_ENTER(_("sl_vsnprintf"));
   if (str == NULL || format == NULL)
-    SL_IRETURN(SL_ENULL, _("sl_vsnprintf"));
-
-  total = 0;
+    SL_IRETURN(0, _("sl_vsnprintf"));
 
 #if defined(HAVE_VSNPRINTF) && !defined(HAVE_BROKEN_VSNPRINTF)
-  vsnprintf (str, n, format, vl);
+  len = vsnprintf (str, n, format, vl);
+  str[n-1] = '\0';
 #else
   VA_COPY (vl2, vl);                   /* save the argument list           */
   total = sl_printf_count (format, vl);
+  len   = (int) total;
   if (total < n) 
-    vsprintf (str, format, vl2);       /* program has checked that it fits */
+    {
+      vsprintf (str, format, vl2);       /* program has checked that it fits */
+      str[n-1] = '\0';
+    }
   else 
     {
       sl_strlcpy (str, format, n);
       va_end(vl2);
-      SL_IRETURN(SL_ETRUNC, _("sl_vsnprintf"));
+      SL_IRETURN(len, _("sl_vsnprintf"));
     }
   va_end(vl2);
 #endif
-  SL_IRETURN(SL_ENONE, _("sl_vsnprintf"));
+  SL_IRETURN(len, _("sl_vsnprintf"));
 }
 
 /*
@@ -442,9 +582,9 @@ int sl_vsnprintf(char *str, size_t n,
 int sl_snprintf(char *str, size_t n,
 		const char *format, ... )
 {
-  int           total = 0;
   va_list       vl;
 #if !defined(HAVE_VSNPRINTF) || defined(HAVE_BROKEN_VSNPRINTF)
+  size_t          total = 0;
   va_list       vl2;
 #endif
 
@@ -455,16 +595,20 @@ int sl_snprintf(char *str, size_t n,
   va_start (vl, format);
 #if defined(HAVE_VSNPRINTF) && !defined(HAVE_BROKEN_VSNPRINTF)
   vsnprintf (str, n, format, vl);
-  total = sl_strlen(str);
+  str[n-1] = '\0';
 #else
   VA_COPY (vl2, vl);                   /* save the argument list           */
   total = sl_printf_count (format, vl);
   if (total < n) 
-    vsprintf (str, format, vl2);       /* program has checked that it fits */
+    {
+      vsprintf (str, format, vl2);     /* program has checked that it fits */
+      str[n-1] = '\0';
+    }
   else 
     {
       sl_strlcpy (str, format, n);
       va_end(vl2);
+      va_end(vl);
       SL_IRETURN(SL_ETRUNC, _("sl_snprintf"));
     }
   va_end(vl2);
@@ -487,49 +631,42 @@ int sl_strlcat(char * dst, const char *src, size_t siz)
 {
   register size_t dst_end;
   register size_t dst_free;
-  register size_t src_len;
 
-  SL_ENTER(_("sl_strlcat"));
+  register char       * p;
+  register const char * q;
 
   if (dst == NULL)
-    SL_IRETURN (SL_ENULL, _("sl_strlcat"));
+    return SL_ENONE;
   if (src == NULL || src == "") 
-    SL_IRETURN (SL_ENONE, _("sl_strlcat"));
-
-  /* How long is src ?
-   */
-  src_len  = strlen(src);
+    return SL_ENONE;
 
   if (siz > 0) {
 
     /* How much free space do we have ?
      */
     dst_end  = strlen(dst);
-    dst_free = siz - dst_end;
+    dst_free = siz - dst_end - 1;
 
-    /* Copy at most dst_free chars.
-     * The documentation on strncat() is inconclusive as to
-     * whether the NULL counts or not, thus use strncpy().
-     * dst[dst_end] is the NULL terminator of dst.
-     */
-    (void) strncpy(&dst[dst_end], 
-		   src, 
-		   ( (src_len < dst_free) ? src_len : dst_free));
+    p = &dst[dst_end];
+    q = src;
+
+    while (dst_free > 0 && *q != '\0')
+      {
+	*p++ = *q++;
+	--dst_free;
+      }
 
     /* NULL terminate dst.
      */
-    if (src_len < dst_free) 
-      dst[dst_end+src_len] = '\0';
-    dst[siz-1] = '\0';
+    *p = '\0';
 
-    if (src_len >= dst_free) 
-      SL_IRETURN (SL_ETRUNC, _("sl_strlcat"));
+    if (*q != '\0') 
+      return SL_ETRUNC;
   }
 
-  SL_IRETURN (SL_ENONE, _("sl_strlcat"));
+  return SL_ENONE;
 }
 
-
 /*
  * An alternative implementation of the OpenBSD strlcpy() function.
  *
@@ -543,15 +680,15 @@ int sl_strlcat(char * dst, const char *src, size_t siz)
  */
 int sl_strlcpy(char * dst, const char * src, size_t siz)
 {
-  SL_ENTER(_("sl_strlcpy"));
+  /* SL_ENTER(_("sl_strlcpy")); */
 
   if (dst == NULL)
-    SL_IRETURN (SL_ENULL, _("sl_strlcpy"));
+    return SL_ENULL;
   if (src == NULL)
     { 
       if (siz > 0) 
 	dst[0] = '\0';
-      SL_IRETURN (SL_ENONE, _("sl_strlcpy"));
+      return SL_ENONE;
     }
 
 
@@ -563,11 +700,8 @@ int sl_strlcpy(char * dst, const char * src, size_t siz)
     /* NULL terminate
      */
     dst[siz-1] = '\0';
-
-    if (strlen(src) >= siz)
-      SL_IRETURN (SL_ETRUNC, _("sl_strlcpy"));
   }
-  SL_IRETURN (SL_ENONE, _("sl_strlcpy"));
+  return SL_ENONE;
 }
 
 /*
@@ -575,7 +709,6 @@ int sl_strlcpy(char * dst, const char * src, size_t siz)
  */
 char * sl_strncpy(char *dst, const char *src, size_t size)
 {
-  size_t len;
 
 #ifdef SL_FAIL_ON_ERROR
   SL_REQUIRE(dst != NULL, _("dst != NULL"));
@@ -603,7 +736,7 @@ char * sl_strncpy(char *dst, const char *src, size_t size)
   else
     strncpy(dst, src, size);
 
-  if ((len = sl_strlen(src)) >= size)
+  if (sl_strlen(src) >= size)
     {
       errno = ENOSPC;
       dst[size-1] = '\0';
@@ -646,7 +779,7 @@ char * sl_strncat(char *dst, const char *src, size_t n)
 }
 
 
-int sl_strcmp(char * a, char * b)
+int sl_strcmp(const char * a, const char * b)
 {
 #ifdef SL_FAIL_ON_ERROR
   SL_REQUIRE (a != NULL, _("a != NULL"));
@@ -663,7 +796,7 @@ int sl_strcmp(char * a, char * b)
     return (-7); /* arbitrary */
 }
 
-int sl_strncmp(char * a, char * b, size_t n)
+int sl_strncmp(const char * a, const char * b, size_t n)
 {
 #ifdef SL_FAIL_ON_ERROR
   SL_REQUIRE (a != NULL, _("a != NULL"));
@@ -681,42 +814,28 @@ int sl_strncmp(char * a, char * b, size_t n)
     return (-7); /* arbitrary */
 }
 
-
-size_t sl_strlen(const char *s)
-{
-#ifdef SL_FAIL_ON_ERROR
-  SL_REQUIRE (s != NULL, _("s != NULL"));
-#endif
-
-  if (s == NULL)
-    return (0);
-  else
-    return (strlen(s));
-}
-
 /* string searching
  */
 
-char * sl_strstr (char * haystack, char * needle) 
+char * sl_strstr (const char * haystack, const char * needle) 
 {
 #ifndef HAVE_STRSTR
   int             i;
-#endif
   size_t          needle_len;
   size_t          haystack_len;
+#endif
   
   if (haystack == NULL || needle == NULL)
     return NULL;
-    
-  needle_len   = strlen(needle);
-  haystack_len = strlen(haystack);
-
-  if (needle_len == 0 || haystack_len == 0)
+  if (*needle == '\0' || *haystack == '\0')
     return NULL;
 
 #if defined(HAVE_STRSTR)
   return (strstr(haystack, needle));
 #else
+  needle_len   = strlen(needle);
+  haystack_len = strlen(haystack);
+
   for (i = 0; i <= (haystack_len-needle_len); ++i)
     if (0 == sl_strncmp(&haystack[i], needle, needle_len))
       return (needle);
@@ -725,8 +844,6 @@ char * sl_strstr (char * haystack, char * needle)
 }
 
 
-
-
 /* ---------------------------------------------------------------- 
  *
  *    Privilege handling routines
@@ -757,13 +874,20 @@ extern       int setresgid (gid_t trgid, gid_t tegid, gid_t tsgid);
  */
 int sl_is_suid()
 {
-  SL_ENTER(_("sl_is_suid"));
-
-  SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));
-  if (euid == ruid && egid == rgid)
-    SL_IRETURN (0, _("sl_is_suid"));     /* FALSE */
+  if (uids_are_stored == SL_FALSE)
+    {
+      if (getuid() == geteuid() && getgid() == getegid())
+	return (0);     /* FALSE */
+      else
+	return (1);     /* TRUE  */
+    }
   else
-    SL_IRETURN (1, _("sl_is_suid"));     /* TRUE  */
+    {
+      if (euid == ruid && egid == rgid)
+	return (0);     /* FALSE */
+      else
+	return (1);     /* TRUE  */
+    }
 }
 
 /*
@@ -773,11 +897,23 @@ int sl_is_suid()
 int sl_get_euid(uid_t * ret)
 {
   SL_ENTER(_("sl_get_euid"));
-  SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));
-  *ret = euid;
+  /* SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));*/
+  if (uids_are_stored == SL_TRUE)
+    *ret = euid;
+  else
+    *ret = geteuid();
   SL_IRETURN (SL_ENONE, _("sl_get_euid"));
 }
 
+uid_t sl_ret_euid()
+{
+  /* SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));*/
+  if (uids_are_stored == SL_TRUE)
+    return (euid);
+  else
+    return (geteuid());
+}
+
 /*
  * This function returns the saved egid.
  * It calls abort() if the uid's are not saved already.
@@ -785,8 +921,11 @@ int sl_get_euid(uid_t * ret)
 int sl_get_egid(gid_t * ret)
 {
   SL_ENTER(_("sl_get_egid"));
-  SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));
-  *ret = egid;
+  /* SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));*/
+  if (uids_are_stored == SL_TRUE)
+    *ret = egid;
+  else
+    *ret = getegid();
   SL_IRETURN (SL_ENONE, _("sl_get_egid"));
 }
 
@@ -797,8 +936,11 @@ int sl_get_egid(gid_t * ret)
 int sl_get_ruid(uid_t * ret)
 {
   SL_ENTER(_("sl_get_ruid"));
-  SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));
-  *ret = ruid;
+  /* SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));*/
+  if (uids_are_stored == SL_TRUE)
+    *ret = ruid;
+  else
+    *ret = getuid();
   SL_IRETURN (SL_ENONE, _("sl_get_ruid"));
 }
 
@@ -809,8 +951,11 @@ int sl_get_ruid(uid_t * ret)
 int sl_get_rgid(gid_t * ret)
 {
   SL_ENTER(_("sl_get_rgid"));
-  SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));
-  *ret = rgid;
+  /* SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));*/
+  if (uids_are_stored == SL_TRUE)
+    *ret = rgid;
+  else
+    *ret = getgid();
   SL_IRETURN (SL_ENONE, _("sl_get_rgid"));
 }
 
@@ -821,8 +966,11 @@ int sl_get_rgid(gid_t * ret)
 int sl_get_ruid_orig(uid_t * ret)
 {
   SL_ENTER(_("sl_get_ruid_orig"));
-  SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));
-  *ret = ruid_orig;
+  /* SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));*/
+  if (uids_are_stored == SL_TRUE)
+    *ret = ruid_orig;
+  else
+    *ret = getuid();
   SL_IRETURN (SL_ENONE, _("sl_get_ruid_orig"));
 }
 
@@ -833,15 +981,18 @@ int sl_get_ruid_orig(uid_t * ret)
 int sl_get_rgid_orig(gid_t * ret)
 {
   SL_ENTER(_("sl_get_rgid_orig"));
-  SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));
-  *ret = rgid_orig;
+  /* SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));*/
+  if (uids_are_stored == SL_TRUE)
+    *ret = rgid_orig;
+  else
+    *ret = getgid();
   SL_IRETURN (SL_ENONE, _("sl_get_rgid_orig"));
 }
 
 static int suid_warn_flag = 1;
 static void suid_warn(int a)
 {
-  fprintf(stderr, "ERROR:  open set/unset suid !!! %d\n", a);
+  fprintf(stderr, _("ERROR:  open set/unset suid !!! %d\n"), a);
   return;
 }
 
@@ -855,6 +1006,12 @@ int sl_set_suid ()
   int retval;
 
   SL_ENTER(_("sl_set_suid"));
+
+  if (uids_are_stored == SL_FALSE)
+    {
+      SL_IRETURN(SL_ENONE, _("sl_set_suid"));
+    }
+
   SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));  
 
   if (ruid == euid && rgid == egid) 
@@ -865,9 +1022,9 @@ int sl_set_suid ()
   SL_REQUIRE(suid_is_set     == SL_FALSE, _("suid_is_set == SL_FALSE"));  
 
 #if defined(HAVE_SETRESUID)
-  retval = setresuid (-1, euid, -1);
+  retval = setresuid (sh_uid_neg, euid, sh_uid_neg);
   if (retval == 0) 
-    retval = setresgid (-1, egid, -1);
+    retval = setresgid (sh_gid_neg, egid, sh_gid_neg);
 
 #elif defined(HAVE_SETEUID)
   retval = seteuid (egid);
@@ -904,6 +1061,12 @@ int sl_unset_suid ()
   register int retval;
 
   SL_ENTER(_("sl_unset_suid"));
+
+  if (uids_are_stored == SL_FALSE)
+    {
+      SL_IRETURN(SL_ENONE, _("sl_unset_suid"));
+    }
+
   SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));
 
   if (ruid == euid && rgid == egid)
@@ -914,9 +1077,9 @@ int sl_unset_suid ()
   SL_REQUIRE(suid_is_set     == SL_TRUE, _("suid_is_set == SL_TRUE"));  
 
 #if defined(HAVE_SETRESUID)
-  retval = setresgid (-1, rgid, -1);
+  retval = setresgid (sh_gid_neg, rgid, sh_gid_neg);
   if (retval == 0) 
-    retval = setresuid (-1, ruid, -1);
+    retval = setresuid (sh_uid_neg, ruid, sh_uid_neg);
 
 #elif defined(HAVE_SETEUID)
   retval = setegid (rgid);
@@ -968,12 +1131,14 @@ int sl_save_uids()
  * This function drops SUID privileges irrevocably.
  * It set the effective uid to the original real uid.
  */
+extern int  sh_unix_initgroups2 (uid_t in_pid, gid_t in_gid);
 int sl_drop_privileges()
 {
   SL_ENTER(_("sl_drop_privileges"));
   SL_REQUIRE(uids_are_stored == SL_TRUE, _("uids_are_stored == SL_TRUE"));
 
   SL_REQUIRE(setgid(rgid_orig) == 0, _("setgid(rgid_orig) == 0"));
+  SL_REQUIRE(sh_unix_initgroups2(ruid_orig, rgid_orig) == 0, _("sh_unix_initgroups2(ruid_orig,rgid_orig) == 0"));
   SL_REQUIRE(setuid(ruid_orig) == 0, _("setuid(ruid_orig) == 0"));
 
   /* make sure that setuid(0) fails
@@ -1009,34 +1174,58 @@ int sl_policy_get_root()
       rgid = egid;
     }
   suid_is_set = SL_TRUE;
-  SL_IRETURN(SL_ENONE, _("sl_policy_get_root"));
-}
-
-/* 
- * Define a policy: Get real (irrevocably).
- * This function drops SUID privileges irrevocably.
- * Do nothing if not SUID.
- */
-int sl_policy_get_real()
-{
-  SL_ENTER(_("sl_policy_get_real"));
-  SL_REQUIRE(uids_are_stored == SL_FALSE, _("uids_are_stored == SL_FALSE"));
-  SL_REQUIRE (sl_save_uids() == SL_ENONE, _("sl_save_uids() == SL_ENONE"));
-
-  if (euid != ruid || egid != rgid)
+  if (euid == 0)
     {
-      SL_REQUIRE (sl_drop_privileges() == SL_ENONE,
-		  _("sl_drop_privileges() == SL_ENONE"));
+      SL_REQUIRE(sh_unix_initgroups2(euid, egid) == 0, _("sh_unix_initgroups2(euid,egid) == 0"));
     }
-  suid_is_set = SL_TRUE;
-  SL_IRETURN(SL_ENONE, _("sl_policy_get_real"));
+  SL_IRETURN(SL_ENONE, _("sl_policy_get_root"));
 }
 
 #include <pwd.h>
 
+/* 
+ * Define a policy: Get real (irrevocably).
+ * This function drops SUID privileges irrevocably.
+ * Do nothing if not SUID (? not true - drops if root).
+ */
+extern struct passwd * sh_getpwnam(const char *name);
+
+int sl_policy_get_real(char * user)
+{
+  struct passwd * tempres;
+
+  SL_ENTER(_("sl_policy_get_real"));
+  SL_REQUIRE(uids_are_stored == SL_FALSE, _("uids_are_stored == SL_FALSE"));
+  SL_REQUIRE (sl_save_uids() == SL_ENONE, _("sl_save_uids() == SL_ENONE"));
+
+  if (euid == 0 || ruid == 0)
+    {
+      tempres = getpwnam(user);
+      if (!tempres)
+	tempres = sh_getpwnam(user);
+      SL_REQUIRE (NULL != tempres, _("tempres != NULL"));
+  
+      rgid_orig = tempres->pw_gid;
+      ruid_orig = tempres->pw_uid;
+    }
+  else
+    {
+      rgid_orig = rgid;
+      ruid_orig = ruid;
+    }
+
+  SL_REQUIRE (sl_drop_privileges() == SL_ENONE,
+	      _("sl_drop_privileges() == SL_ENONE"));
+
+  suid_is_set = SL_TRUE;
+  SL_IRETURN(SL_ENONE, _("sl_policy_get_real"));
+}
+
+
 /* 
  * Define a policy: Get user.
  * Drops privileges.
+ * Do nothing if not SUID.
  */
 int sl_policy_get_user(char * user)
 {
@@ -1051,6 +1240,8 @@ int sl_policy_get_user(char * user)
   if (euid != ruid || egid != rgid)
     {
       tempres = getpwnam(user);
+      if (!tempres)
+	tempres = sh_getpwnam(user);
       SL_REQUIRE (NULL != tempres, _("tempres != NULL"));
 
       rgid = tempres->pw_gid;
@@ -1070,8 +1261,12 @@ int sl_policy_get_user(char * user)
  *
  * ---------------------------------------------------------------- */
 
-#define TOFFSET 0x1234 
-#define MAXFD   FOPEN_MAX
+#define TOFFSET 0x1234
+
+/* this would prevent opening files if the first 16 fds are open :( */ 
+/* #define MAXFD   FOPEN_MAX                                        */
+
+#define MAXFD   1024
 
 typedef struct openfiles {
   SL_TICKET ticket;     /* The unique  ID.      */ 
@@ -1079,7 +1274,7 @@ typedef struct openfiles {
   char * path;          /* The file path.       */
 } SL_OFILE; 
 
-static SL_OFILE *ofiles[MAXFD]; 
+static SL_OFILE * ofiles[MAXFD]; 
 
 
 static unsigned int nonce_counter = TOFFSET;
@@ -1125,25 +1320,23 @@ int sl_read_ticket (SL_TICKET fno)
   register unsigned myindex; 
   register SL_OFILE *of; 
 
-  SL_ENTER(_("sl_read_ticket"));
-
   myindex = ((fno >> 16) & 0xffff) - TOFFSET;
   if (myindex >= MAXFD)
-    SL_IRETURN(SL_ETICKET, _("sl_read_ticket"));
+    return (SL_ETICKET);
 
   if (ofiles[myindex] == NULL)
-    SL_IRETURN(SL_ETICKET, _("sl_read_ticket"));
+    return (SL_ETICKET);
 
   if (ofiles[myindex]->ticket != fno)
-    SL_IRETURN(SL_ETICKET, _("sl_read_ticket"));
+    return (SL_ETICKET);
 
   if ((of = ofiles[myindex])->fd < 0 || of->fd >= MAXFD )
-    SL_IRETURN(SL_EINTERNAL, _("sl_read_ticket"));
+    return (SL_EINTERNAL);
 
   if (((of->ticket) & 0xffff) == 0)
-    SL_IRETURN(SL_EINTERNAL, _("sl_read_ticket")); 
+    return (SL_EINTERNAL); 
 
-  SL_RETURN(myindex, _("sl_read_ticket")); 
+  return (myindex); 
 }
 
 #define SL_OPEN_MIN          113
@@ -1191,8 +1384,7 @@ int sl_open_file (char *filename, int mode, int priv)
     SL_IRETURN(SL_ENULL, _("sl_open_file"));
   if (mode < SL_OPEN_MIN || mode > SL_OPEN_MAX)
     SL_IRETURN(SL_EINTERNAL, _("sl_open_file"));
-  
-  
+    
   /* "This system call always succeeds and the previous value of
    * the mask is returned." 
    */
@@ -1203,8 +1395,8 @@ int sl_open_file (char *filename, int mode, int priv)
       fd = aud_open (FIL__, __LINE__, priv, filename, 
 		     O_RDONLY|O_NOATIME|O_NONBLOCK, 0);
       if (fd >= 0) {
-	sflags = retry_fcntl(fd, F_GETFL, 0);
-	retry_fcntl(fd, F_SETFL, sflags & ~O_NONBLOCK);
+	sflags = retry_fcntl(FIL__, __LINE__, fd, F_GETFL, 0);
+	retry_fcntl(FIL__, __LINE__, fd, F_SETFL, sflags & ~O_NONBLOCK);
       }
       if (fd < 0)
 	SL_IRETURN(SL_EBADFILE, _("sl_open_file"));
@@ -1213,7 +1405,10 @@ int sl_open_file (char *filename, int mode, int priv)
 
   if (priv == SL_YESPRIV)
     sl_set_suid();
-  lstat_return = retry_lstat(filename, &lbuf);
+  if (mode == SL_OPEN_FOR_READ)
+    lstat_return = retry_stat (FIL__, __LINE__, filename, &lbuf);
+  else
+    lstat_return = retry_lstat(FIL__, __LINE__, filename, &lbuf);
   if (priv == SL_YESPRIV)
     sl_unset_suid();
 
@@ -1243,8 +1438,8 @@ int sl_open_file (char *filename, int mode, int priv)
       fd = aud_open (FIL__, __LINE__, priv, filename, 
 		     O_RDONLY|O_NOATIME|O_NONBLOCK, 0);
       if (fd >= 0) {
-	sflags = retry_fcntl(fd, F_GETFL, 0);
-	retry_fcntl(fd, F_SETFL, sflags & ~O_NONBLOCK);
+	sflags = retry_fcntl(FIL__, __LINE__, fd, F_GETFL, 0);
+	retry_fcntl(FIL__, __LINE__, fd, F_SETFL, sflags & ~O_NONBLOCK);
       }
       break;
     case SL_OPEN_FOR_WRITE:
@@ -1300,7 +1495,7 @@ int sl_open_file (char *filename, int mode, int priv)
 
   if (priv == SL_YESPRIV)
     sl_set_suid();
-  stat_return = retry_fstat(fd, &buf);
+  stat_return = retry_fstat(FIL__, __LINE__, fd, &buf);
   if (priv == SL_YESPRIV)
     sl_unset_suid();
 
@@ -1320,21 +1515,27 @@ int sl_open_file (char *filename, int mode, int priv)
 
   /* Make entry.
    */
-  if (fd >= MAXFD || ofiles[fd] != NULL)
+  if (fd >= MAXFD)
+     {
+	close(fd);
+	SL_IRETURN(SL_TOOMANY, _("sl_open_file"));
+     }
+
+   if (ofiles[fd] != NULL)
     {
       close(fd);
       SL_IRETURN(SL_EINTERNAL, _("sl_open_file"));
     }
 
-
-  if ( (ofiles[fd] = (SL_OFILE *) my_malloc(sizeof(SL_OFILE))) == NULL)
+  if ( (ofiles[fd] = (SL_OFILE *) malloc(sizeof(SL_OFILE))) == NULL)
     {
       close(fd);
       SL_IRETURN(SL_EMEM, _("sl_open_file"));
     }
-  if ( (ofiles[fd]->path = (char *) my_malloc( strlen(filename)+1) ) == NULL)
+
+   if ( (ofiles[fd]->path = (char *) malloc( strlen(filename)+1) ) == NULL)
     {
-      my_free(ofiles[fd]);
+      free(ofiles[fd]);
       ofiles[fd] = NULL;
       close(fd);
       SL_IRETURN(SL_EMEM, _("sl_open_file"));
@@ -1346,8 +1547,8 @@ int sl_open_file (char *filename, int mode, int priv)
 
   if (SL_ISERROR(ticket))
     {
-      (void) my_free (ofiles[fd]->path);
-      (void) my_free (ofiles[fd]);
+      (void) free (ofiles[fd]->path);
+      (void) free (ofiles[fd]);
       close(fd);
       SL_IRETURN(ticket, _("sl_open_file"));
     }
@@ -1391,7 +1592,7 @@ SL_TICKET sl_open_read (char * fname, int priv)
     {
       TPT(( 0, FIL__, __LINE__, 
 	    _("msg=<Error in check_fname_priv.> status=<%ld>\n"), 
-	    fname, status));
+	    status));
       SL_IRETURN(status, _("sl_open_read"));
     }
 
@@ -1464,14 +1665,12 @@ int get_the_fd (SL_TICKET ticket)
 {
   int fd;
 
-  SL_ENTER(_("get_the_fd"));
-
   if (SL_ISERROR(fd = sl_read_ticket(ticket)))
-    SL_IRETURN(fd, _("get_the_fd"));
+    return (fd);
 
   if (ofiles[fd] == NULL || fd != ofiles[fd]->fd || fd < 0)
-    SL_IRETURN(SL_EINTERNAL, _("get_the_fd"));
-  SL_IRETURN(fd, _("get_the_fd"));
+    return (SL_EINTERNAL);
+  return (fd);
 }
 
 int sl_close (SL_TICKET ticket) 
@@ -1494,8 +1693,8 @@ int sl_close (SL_TICKET ticket)
 
   if (ofiles[fd] != NULL)
     {
-      (void) my_free(ofiles[fd]->path);
-      (void) my_free(ofiles[fd]);
+      (void) free(ofiles[fd]->path);
+      (void) free(ofiles[fd]);
       ofiles[fd] = NULL;
     }
 
@@ -1509,8 +1708,8 @@ int sl_dropall(int fd)
       if (ofiles[fd] != NULL)
 	{
 	  if (ofiles[fd]->path != NULL)
-	    (void) my_free(ofiles[fd]->path);
-	  (void) my_free(ofiles[fd]);
+	    (void) free(ofiles[fd]->path);
+	  (void) free(ofiles[fd]);
 	  ofiles[fd] = NULL;
 	}
       ++fd;
@@ -1580,6 +1779,7 @@ int sl_forward (SL_TICKET ticket)
   SL_IRETURN(SL_ENONE, _("sl_forward")); 
 }
 
+
 int sl_sync (SL_TICKET ticket) 
 {
   register int fd;
@@ -1604,15 +1804,22 @@ int sl_read (SL_TICKET ticket, void * buf_in, size_t count)
 
   char * buf;
 
-  SL_ENTER(_("sl_read"));
-
   if (count < 1)
-    SL_IRETURN(SL_ERANGE, _("sl_read"));
+    {
+      TPT(( 0, FIL__, __LINE__, _("msg=<range error>")));
+      return(SL_ERANGE);
+    }
   if (buf_in == NULL)
-    SL_IRETURN(SL_ENULL, _("sl_read"));
+    {
+      TPT(( 0, FIL__, __LINE__, _("msg=<null buffer>")));
+      return (SL_ENULL);
+    }
 
   if (SL_ISERROR(fd = get_the_fd(ticket)))
-    SL_IRETURN(fd, _("sl_read"));
+    {
+      TPT(( 0, FIL__, __LINE__, _("msg=<ticket error> errno=<%d>"), fd));
+      return (fd);
+    }
 
   buf = (char *) buf_in;
 
@@ -1630,9 +1837,11 @@ int sl_read (SL_TICKET ticket, void * buf_in, size_t count)
 
  
   if (byteread == (-1))
-    SL_IRETURN(SL_EREAD, _("sl_read"));
-
-   SL_IRETURN(bytes, _("sl_read"));
+    {
+      TPT(( 0, FIL__, __LINE__, _("msg=<read error> errno=<%d>\n"), errno));
+      return (SL_EREAD);
+    }
+  return (bytes);
 }
 
 
@@ -1697,11 +1906,12 @@ int sl_write_line (SL_TICKET ticket, void * msg, long nbytes)
  *
  * ---------------------------------------------------------------- */
 
-extern long rootonly[];
-extern int  EUIDSLOT;
+extern uid_t rootonly[];
+extern int   EUIDSLOT;
+extern int   ORIG_EUIDSLOT;
 
-extern char tf_path[MAXFILENAME];	/* Error path for trust function. */
-extern long tf_euid;	                /* Space for EUID of process.     */
+extern char  tf_path[MAXFILENAME];	/* Error path for trust function. */
+extern uid_t tf_euid;	                /* Space for EUID of process.     */
 
 
 char * sl_error_string(int errorcode)
@@ -1744,7 +1954,7 @@ char * sl_error_string(int errorcode)
     case SL_EBADFILE:
       return _("File access error.");
     case SL_EBADNAME:
-      return _("Invalid filename (prob. too long).");
+      return _("Invalid filename (prob. too long or null).");
 
     case SL_ETRUNC:
       return _("Truncation occured.");
@@ -1762,29 +1972,55 @@ char * sl_trust_errfile()
   return &tf_path[0];
 }
 
+extern uid_t tf_baduid;
+uid_t   sl_trust_baduid()
+{
+  return tf_baduid;
+}
 
-int  sl_trust_add_user (long pwid)
+extern gid_t tf_badgid;
+gid_t   sl_trust_badgid()
+{
+  return tf_badgid;
+}
+
+
+static int trust_count = 0;
+
+int  sl_trust_purge_user ()
+{
+  int i;
+
+  EUIDSLOT = ORIG_EUIDSLOT;
+  trust_count = 0;
+
+  for (i = EUIDSLOT; i < (EUIDSLOT + 15); ++i) 
+    rootonly[i] = sh_uid_neg;
+  return 0;
+}
+
+int  sl_trust_add_user (uid_t pwid)
 {
-  static int count = 0;
   SL_ENTER(_("sl_trust_add_user"));
 
-  if (count == 6)
+  if (trust_count == 15)
     SL_IRETURN(SL_ERANGE, _("sl_trust_add_user"));
   
   rootonly[EUIDSLOT] = pwid;
   ++EUIDSLOT;
-  ++count;
+  ++trust_count;
+
   SL_IRETURN(SL_ENONE, _("sl_trust_add_user"));
 }
 
-int sl_trustfile_euid(char * filename, long teuid)
+int sl_trustfile_euid(char * filename, uid_t teuid)
 {
   long status;
   SL_ENTER(_("sl_trustfile_euid"));
 
   tf_path[0] = '\0';
-  if (filename == NULL)
-    SL_IRETURN(SL_EBADFILE, _("sl_trustfile_euid"));
+  if (filename == NULL || filename[0] == '\0')
+    SL_IRETURN(SL_EBADNAME, _("sl_trustfile_euid"));
 
   tf_euid = teuid;
   status = sl_trustfile(filename, NULL, NULL);
