@@ -1,6 +1,13 @@
+/* debug problems        */
+/* #define TRUST_DEBUG   */
+
 /* switch off full check */
 /* #define TEST_ONLY     */
 
+/* standalone            */
+/* #define TRUST_MAIN    */
+/* $(CC) -DTRUST_MAIN -DSL_ALWAYS_TRUSTED=...  */
+
 /* LINTLIBRARY */
 /*
  * This is the file with all the library routines in it
@@ -73,9 +80,12 @@
  * is always copied between buffers of length MAXFILENAME.
  */
 
+#ifndef  TRUST_MAIN
 #include "config_xor.h"
 #include "sh_calls.h"
-
+#else
+#define UID_CAST long
+#endif
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -86,21 +96,40 @@
 #include <string.h>
 #include <unistd.h>
 
-/* -- Defined in config.h. --
-
-#ifndef _(string)
-#define _(string) string
-#endif
-
-#ifndef N_(string)
-#define N_(string) string
-#endif
-*/
 
+#ifndef TRUST_MAIN
 #include "slib.h"
+extern struct passwd * sh_getpwnam(const char *name);
+extern struct passwd * sh_getpwuid (uid_t uid);
+extern struct group  * sh_getgrgid(const gid_t gid);
+#else
+
+#define TRUST_DEBUG
+#define SL_FALSE 0
+#define SL_TRUE  1
+#define SL_ENTER(string)
+#define SL_IRETURN(a, b)  return a
+#define retry_lstat(a,b,c,d) lstat(c,d)
+#define _(string)  string 
+#define N_(string) string
+#define MAXFILENAME     4096
+static int sl_errno = 0;
+#define SL_ENONE         0
+#define SL_ENULL     -1024     /* Invalid use of NULL pointer.         */
+#define SL_ERANGE    -1025     /* Argument out of range.               */
+#define SL_ETRUNC    -1026     /* Result truncated.                    */
+#define SL_EINTERNAL -1028     /* Internal error.                      */
+#define SL_EBADFILE  -1030     /* File access error. Check errno.      */
+#define SL_EBADNAME  -1040     /* Invalid name.                        */
+#define SL_ESTAT     -1041     /* stat of file failed. Check errno.    */
+#define SL_EBADUID   -1050	/* Owner not trustworthy.              */
+#define SL_EBADGID   -1051	/* Group writeable and not trustworthy.*/
+#define SL_EBADOTH   -1052	/* World writeable.                    */
+
+#endif
 
 
-#if defined(__linux__)
+#if defined(__linux__) || defined(__FreeBSD__)
 #define STICKY
 #endif
 
@@ -116,12 +145,20 @@
  */
 
 
+#ifndef TRUST_MAIN
+
 #if defined(HAVE_GETCWD) && !defined(HAVE_BROKEN_GETCWD)
 #define CURDIR(buf,nbuf)	getcwd((buf), (nbuf))
 #elif defined(HAVE_GETWD)
 #define CURDIR(buf,nbuf)	getwd((buf))
 #endif
 
+#else
+
+#define CURDIR(buf,nbuf)	getcwd((buf), (nbuf))
+
+#endif
+
 
 
 /*
@@ -157,13 +194,23 @@
 #ifndef SL_ALWAYS_TRUSTED
 #define SL_ALWAYS_TRUSTED  0
 #endif
-static long test_rootonly[] = { SL_ALWAYS_TRUSTED };
+static uid_t test_rootonly[] = { SL_ALWAYS_TRUSTED };
 
-long rootonly[] = { SL_ALWAYS_TRUSTED, -1, -1, -1, -1, -1, -1, -1, -1, -1 };
-long tf_euid = -1;
+#define tf_uid_neg ((uid_t)-1)
+
+uid_t rootonly[] = { SL_ALWAYS_TRUSTED, 
+		    tf_uid_neg, tf_uid_neg, tf_uid_neg, tf_uid_neg, 
+		    tf_uid_neg, tf_uid_neg, tf_uid_neg, tf_uid_neg, 
+		    tf_uid_neg, tf_uid_neg, tf_uid_neg, tf_uid_neg, 
+		    tf_uid_neg, tf_uid_neg, tf_uid_neg, tf_uid_neg };
+
+uid_t tf_euid = tf_uid_neg;
 int EUIDSLOT = sizeof(test_rootonly)/sizeof(long);
+int ORIG_EUIDSLOT = sizeof(test_rootonly)/sizeof(long);
 
-char tf_path[MAXFILENAME];		/* error path for trust function */
+char  tf_path[MAXFILENAME];		/* error path for trust function */
+uid_t tf_baduid;
+gid_t tf_badgid;
 
 static 
 int dirz(char *path)
@@ -171,7 +218,8 @@ int dirz(char *path)
   register char *p = path;/* points to rest of path to clean up */
   register char *q;	/* temp pointer for skipping over stuff */
 
-  char swp[MAXFILENAME];
+  static   char swp[MAXFILENAME];
+
   SL_ENTER(_("dirz"));
   /*
    * standard error checking
@@ -261,8 +309,10 @@ int dirz(char *path)
 static 
 int getfname(char *fname, char *rbuf, int rsz)
 {
+#ifndef TRUST_MAIN
   register int status;
-  
+#endif
+
   SL_ENTER(_("getfname"));
   /*
    * do the initial checking
@@ -280,7 +330,12 @@ int getfname(char *fname, char *rbuf, int rsz)
   else
     {
       if (CURDIR(rbuf, rsz)  == NULL)
-	SL_IRETURN(SL_EBADNAME, _("getfname"));
+	{
+#ifdef TRUST_DEBUG
+	  fprintf(stderr, "trustfile: getcwd failed\n");
+#endif 
+	  SL_IRETURN(SL_EBADNAME, _("getfname"));
+	}
     }
   
   /*
@@ -288,28 +343,52 @@ int getfname(char *fname, char *rbuf, int rsz)
    */
   if (fname != NULL && *fname != '\0')
     {
+#ifndef TRUST_MAIN
       status = sl_strlcat(rbuf, "/", rsz);
       if (status == SL_ENONE)
 	status = sl_strlcat(rbuf, fname, rsz);
       if (status != SL_ENONE)
 	SL_IRETURN(status, _("getfname"));
+#else
+      strncat(rbuf, "/",   rsz-strlen(rbuf));
+      strncat(rbuf, fname, rsz-strlen(rbuf));
+#endif
     }
   SL_IRETURN(dirz(rbuf), _("getfname"));
 }
 
 static 
-int isin(long n, long *list)
+int isin(uid_t n, uid_t *list)
 {
   SL_ENTER(_("isin"));
   if (list == NULL)
     SL_IRETURN(SL_FALSE, _("isin"));
 
-  while(*list != -1 && *list != n)
-    list++;
+  while(*list != tf_uid_neg && *list != n)
+    {
+#ifdef TRUST_DEBUG
+      fprintf (stderr, 
+	       "trustfile: owner_uid=%ld, trusted uid=%ld, no match\n", 
+	       (UID_CAST) n, (UID_CAST) *list);
+#endif 
+      list++;
+    }
 
-  if (*list == -1)
-    SL_IRETURN(SL_FALSE, _("isin"));
+  if (*list == tf_uid_neg)
+    {
+#ifdef TRUST_DEBUG
+      fprintf (stderr, 
+	       "trustfile: owner_uid=%ld, no match with any trusted user --> ERROR\n", 
+	       (UID_CAST) n);
+#endif 
+      SL_IRETURN(SL_FALSE, _("isin"));
+    }
 
+#ifdef TRUST_DEBUG
+  fprintf (stderr, 
+	   "trustfile: owner_uid=%ld, trusted_uid=%ld, match found --> OK\n", 
+	   (UID_CAST)n, (UID_CAST)*list);
+#endif 
   SL_IRETURN(SL_TRUE, _("isin"));
 }
 
@@ -317,18 +396,24 @@ int isin(long n, long *list)
  *  RETURN TRUE if ANYONE in ulist is group member
  */
 static 
-int isingrp(long grp, long *ulist)
+int isingrp(gid_t grp, uid_t *ulist)
 {
   register struct passwd *w;	/* info about group member */
-  register long *u;		/* points to current ulist member */
+  register uid_t *u;		/* points to current ulist member */
   register char **p;		/* points to current group member */
   register struct group *g;	/* pointer to group information */
   
   SL_ENTER(_("isingrp"));
 
   if ((g = getgrgid(grp)) == NULL)
-    SL_IRETURN(SL_FALSE, _("isingrp") );
-
+    {
+#ifdef TRUST_MAIN 
+      SL_IRETURN(SL_FALSE, _("isingrp") );
+#else
+      if ((g = sh_getgrgid(grp)) == NULL)
+	SL_IRETURN(SL_FALSE, _("isingrp") );
+#endif
+    }
   /*
   if(g->gr_mem == NULL || g->gr_mem[0] == NULL )
     SL_IRETURN(SL_FALSE, _("isingrp") );
@@ -337,17 +422,38 @@ int isingrp(long grp, long *ulist)
   /* this will return at the first match
    */
   for(p = g->gr_mem; *p != NULL; p++)
-    for(u = ulist; *u != -1; u++)
-      /* map user name to UID and compare */
-      if ((w = getpwnam(*p)) != NULL && *u == (long)(w->pw_uid) )
-	SL_IRETURN(SL_TRUE, _("isingrp"));
-
+    {
+      for(u = ulist; *u != tf_uid_neg; u++)
+	{
+	  /* map user name to UID and compare */
+#ifdef TRUST_MAIN
+	  if ((w = getpwnam(*p)) != NULL && *u == (uid_t)(w->pw_uid) )
+	    SL_IRETURN(SL_TRUE, _("isingrp"));
+#else
+	  if ((w = getpwnam(*p)) != NULL && *u == (uid_t)(w->pw_uid) )
+	    {
+	      if ((w = sh_getpwnam(*p)) != NULL && *u == (uid_t)(w->pw_uid) )
+		SL_IRETURN(SL_TRUE, _("isingrp"));
+	    }
+#endif
+	}
+    }
   /* added by R. Wichmann Fri Mar 30 08:16:14 CEST 2001: 
    * a user can have a GID but no entry in /etc/group
    */
-  for(u = ulist; *u != -1; u++)
-    if ((w = getpwuid(*u)) != NULL && grp == (long)(w->pw_gid) )
-      SL_IRETURN(SL_TRUE, _("isingrp"));
+  for(u = ulist; *u != tf_uid_neg; u++)
+    {
+#ifdef TRUST_MAIN
+      if ((w = getpwuid(*u)) != NULL && grp == (gid_t)(w->pw_gid) )
+	SL_IRETURN(SL_TRUE, _("isingrp"));
+#else
+      if ((w = getpwuid(*u)) != NULL && grp == (gid_t)(w->pw_gid) )
+	{
+	  if ((w = sh_getpwuid(*u)) != NULL && grp == (gid_t)(w->pw_gid) )
+	    SL_IRETURN(SL_TRUE, _("isingrp"));
+	}
+#endif
+    }
 
   SL_IRETURN(SL_FALSE, _("isingrp"));
 }
@@ -356,18 +462,36 @@ int isingrp(long grp, long *ulist)
  *  RETURN TRUE only if ALL group members are trusted
  */
 static 
-int onlytrustedingrp(long grp, long *ulist)
+int onlytrustedingrp(gid_t grp, uid_t *ulist)
 {
   register struct passwd *w;	/* info about group member */
-  register long *u;		/* points to current ulist member */
+  register uid_t *u;		/* points to current ulist member */
   register char **p;		/* points to current group member */
   register struct group *g;	/* pointer to group information */
   register int flag = -1;       /* group member found */
 
   SL_ENTER(_("onlytrustedingrp"));
 
+#ifdef TRUST_DEBUG
+  fprintf(stderr, "trustfile: group writeable, group_gid: %ld\n", 
+	  (UID_CAST)grp); 
+#endif
+
   if ((g = getgrgid(grp)) == NULL)
-    SL_IRETURN(SL_FALSE, _("onlytrustedingrp") );
+    {
+#ifndef TRUST_MAIN
+      g = sh_getgrgid(grp);
+#endif
+      if (g == NULL)
+	{
+#ifdef TRUST_DEBUG
+	  fprintf(stderr, 
+		  "trustfile: group_gid: %ld, no such group --> ERROR\n", 
+		  (UID_CAST)grp); 
+#endif
+	  SL_IRETURN(SL_FALSE, _("onlytrustedingrp") );
+	}
+    }
 
   /* empty group -> no problem
    
@@ -380,16 +504,61 @@ int onlytrustedingrp(long grp, long *ulist)
   for(p = g->gr_mem; *p != NULL; p++)
     {
       flag = -1;
-      /* fprintf(stderr, "GROUP MEMBER: %s\n", *p); */
-      for(u = ulist; *u != -1; u++)
+#ifdef TRUST_DEBUG
+      fprintf(stderr, "trustfile: group_member: %s\n", *p); 
+#endif
+      /* map user name to UID and compare 
+       */
+      w = getpwnam(*p);
+#ifndef TRUST_MAIN
+      if (!w)
+	w = sh_getpwnam(*p);
+#endif
+      if (w == NULL)    /* not a valid user, ignore    */
 	{
-	  /* map user name to UID and compare */
-	  if ((w = getpwnam(*p)) != NULL && *u == (long)(w->pw_uid) )
-	    flag = 0;
+	  flag = 0; 
 	}
-      /* not found */
+      else              /* check list of trusted users */
+	{
+#ifdef TRUST_DEBUG
+	  fprintf (stderr, 
+		   "trustfile: uid=%ld, checking whether it is trusted\n",
+		   (UID_CAST)(w->pw_uid));
+#endif 
+	  for(u = ulist; *u != tf_uid_neg; u++)
+	    {
+	      if (*u == (w->pw_uid) )
+		{
+#ifdef TRUST_DEBUG
+		  fprintf (stderr, 
+			   "trustfile: uid=%ld, trusted_uid=%ld, match found --> OK\n", 
+			   (UID_CAST)(w->pw_uid), (UID_CAST)*u);
+#endif 
+		  flag = 0;
+		  break;
+		}
+	      else
+		{
+#ifdef TRUST_DEBUG
+		  fprintf (stderr, 
+			   "trustfile: uid=%ld, trusted_uid=%ld, no match\n", 
+			   (UID_CAST)(w->pw_uid), (UID_CAST)*u);
+#endif 
+		  ;
+		}
+	    }
+	}
+      /* not found 
+       */
       if (flag == -1)
+	{
+#ifdef TRUST_DEBUG
+	  fprintf (stderr, 
+		   "trustfile: user=%s (gid %ld), not a trusted user --> ERROR\n", *p, (UID_CAST)grp);
+#endif 
+	  tf_baduid = w->pw_uid;
 	  SL_IRETURN(SL_FALSE, _("onlytrustedingrp"));
+	}
     }
 
 #ifndef TEST_ONLY	
@@ -398,20 +567,46 @@ int onlytrustedingrp(long grp, long *ulist)
    */
   while (NULL != (w = getpwent())) 
     {
-      if (grp == (long)(w->pw_gid))
+      if (grp == (gid_t)(w->pw_gid))
 	{
-	  /* fprintf(stderr, "USER: %s\n", w->pw_name); */
+#ifdef TRUST_DEBUG
+	  fprintf(stderr, "trustfile: checking group member %s, uid %ld\n", 
+		  w->pw_name, (UID_CAST)w->pw_uid); 
+#endif
 	  /* is it a trusted user ?
 	   */
 	  flag = -1;
-	  for(u = ulist; *u != -1; u++)
+	  for(u = ulist; *u != tf_uid_neg; u++)
 	    {
-	      if (*u == (long)(w->pw_uid))
-		flag = 0;
+	      if (*u == (uid_t)(w->pw_uid))
+		{
+#ifdef TRUST_DEBUG
+		  fprintf (stderr, 
+			   "trustfile: uid=%ld, trusted_uid=%ld, match found --> OK\n", 
+			   (UID_CAST)(w->pw_uid), *u);
+#endif 
+		  flag = 0;
+		  break;
+		}
+	      else
+		{
+#ifdef TRUST_DEBUG
+		  fprintf (stderr, 
+			   "trustfile: uid=%ld, trusted_uid=%ld, no match\n", 
+			   (UID_CAST)(w->pw_uid), (UID_CAST)*u);
+#endif 
+		  ;
+		}
 	    }
 	  /* not found */
 	  if (flag == -1)
-	    SL_IRETURN(SL_FALSE, _("onlytrustedingrp"));
+	    {
+#ifdef TRUST_DEBUG
+	      fprintf(stderr,"trustfile: group member %s not found in trusted users --> ERROR\n", w->pw_name); 
+#endif
+	      tf_baduid = w->pw_uid;
+	      SL_IRETURN(SL_FALSE, _("onlytrustedingrp"));
+	    }
 	}
     }
 
@@ -424,12 +619,17 @@ int onlytrustedingrp(long grp, long *ulist)
   /* #ifdef HAVE_GETPWENT */
 #endif
 
+#ifdef TRUST_DEBUG
+  fprintf(stderr,
+	  "trustfile: group %ld:  all members are trusted users --> OK\n", 
+	  (UID_CAST)grp);
+#endif
   /* all found
    */
   SL_IRETURN(SL_TRUE, _("onlytrustedingrp"));
 }
 
-int sl_trustfile(char *fname, long *okusers, long *badusers)
+int sl_trustfile(char *fname, uid_t *okusers, uid_t *badusers)
 {
   char fexp[MAXFILENAME];	/* file name fully expanded        */
   register char *p = fexp;      /* used to hold name to be checked */
@@ -444,8 +644,14 @@ int sl_trustfile(char *fname, long *okusers, long *badusers)
    * next expand to the full file name
    * getfname sets sl_errno as appropriate
    */
+#ifdef TRUST_MAIN
+  sl_errno = getfname(fname, fexp, MAXFILENAME);
+  if (sl_errno != 0)
+    return sl_errno;
+#else
   if (SL_ISERROR(getfname(fname, fexp, MAXFILENAME)))
       SL_IRETURN(sl_errno, _("sl_trustfile"));
+#endif
 
   if (okusers == NULL && badusers == NULL)
     {
@@ -485,9 +691,15 @@ int sl_trustfile(char *fname, long *okusers, long *badusers)
       /*
        * now get the information
        */
-      if (retry_lstat(fexp, &stbuf) < 0)
+      if (retry_lstat(FIL__, __LINE__, fexp, &stbuf) < 0)
 	{
 	  (void) strcpy(tf_path, fexp);                  /* known to fit  */
+#ifdef TRUST_MAIN
+	  fprintf(stderr, "---------------------------------------------\n");
+	      fprintf(stderr, "trustfile: ESTAT: stat(%s) failed, maybe the file does not exist\n",
+		      fexp);
+	  fprintf(stderr, "---------------------------------------------\n");
+#endif
 	  SL_IRETURN(SL_ESTAT, _("sl_trustfile"));
 	}
 
@@ -532,7 +744,15 @@ int sl_trustfile(char *fname, long *okusers, long *badusers)
 	  if (lsym >= 0) 
 	    csym[lsym] = '\0';
 	  else
-	    SL_IRETURN(SL_EBADNAME, _("sl_trustfile"));
+	    {
+#ifdef TRUST_MAIN
+	      fprintf(stderr, "---------------------------------------------\n");
+	      fprintf(stderr, "trustfile: EBADNAME: readlink(%s) failed\n",
+		      fexp);
+	      fprintf(stderr, "---------------------------------------------\n");
+#endif
+	      SL_IRETURN(SL_EBADNAME, _("sl_trustfile"));
+	    }
 
 	  /*
 	   * relative or absolute referent?
@@ -568,6 +788,13 @@ int sl_trustfile(char *fname, long *okusers, long *badusers)
 		  /* yes -- error 
 		   */
 		  (void) strcpy(tf_path, fexp);          /* known to fit  */
+#ifdef TRUST_MAIN
+		  fprintf(stderr, "---------------------------------------------\n");
+		  fprintf(stderr, 
+			  "trustfile: ETRUNC: normalized path too long (%s)\n",
+			  fexp);
+		  fprintf(stderr, "---------------------------------------------\n");
+#endif
 		  SL_IRETURN(SL_ETRUNC, _("sl_trustfile"));
 		}
 	      *b = '\0';
@@ -607,15 +834,30 @@ int sl_trustfile(char *fname, long *okusers, long *badusers)
 #endif
 
 			
+#ifdef TRUST_DEBUG
+      fprintf(stderr, "\ntrustfile: checking path=%s\n", fexp); 
+#endif 
       /*
        * if the owner is not trusted then -- as the owner can
        * change protection modes -- he/she can write to the
        * file regardless of permissions, so bomb
        */
-      if (((okusers != NULL && SL_FALSE == isin((long)stbuf.st_uid,okusers))||
-	   (badusers != NULL && SL_TRUE == isin((long)stbuf.st_uid,badusers))))
+      if (((okusers != NULL && SL_FALSE == isin((uid_t)stbuf.st_uid,okusers))||
+	   (badusers != NULL && SL_TRUE == isin((uid_t)stbuf.st_uid,badusers))))
 	{
+#ifdef TRUST_DEBUG
+	  fprintf(stderr, "---------------------------------------------\n");
+	  fprintf(stderr, "trustfile: EBADUID %s (owner not trusted)\n", 
+		  fexp); 
+	  fprintf(stderr, "The owner of this file/directory is not in samhains\n"); 
+	  fprintf(stderr, "list of trusted users.\n");
+	  fprintf(stderr, "Please run ./configure again with the option\n");
+	  fprintf(stderr, " ./configure [more options] --with-trusted=0,...,UID\n"); 
+	  fprintf(stderr, "where UID is the UID of the (yet) untrusted user.\n"); 
+	  fprintf(stderr, "---------------------------------------------\n");
+#endif 
 	  (void) strcpy(tf_path, fexp);                  /* known to fit  */
+	  tf_baduid = (uid_t) stbuf.st_uid;
 	  SL_IRETURN(SL_EBADUID, _("sl_trustfile"));
 	}
 
@@ -631,15 +873,28 @@ int sl_trustfile(char *fname, long *okusers, long *badusers)
        * (untrusted) users in the group
        */
       if (((stbuf.st_mode & S_IWGRP) == S_IWGRP) &&
-	  ((okusers != NULL && !onlytrustedingrp((long)stbuf.st_gid,okusers))||
-	   (badusers != NULL && isingrp((long)stbuf.st_gid, badusers)))
+	  ((okusers != NULL && !onlytrustedingrp((gid_t)stbuf.st_gid,okusers))||
+	   (badusers != NULL && isingrp((gid_t)stbuf.st_gid, badusers)))
 #ifdef STICKY
 	  && ((stbuf.st_mode&S_IFDIR) != S_IFDIR ||
 	      (stbuf.st_mode&S_ISVTX) != S_ISVTX)
 #endif
 	  )
 	{
+#ifdef TRUST_DEBUG
+	  fprintf(stderr, "---------------------------------------------\n");
+	  fprintf(stderr, 
+		  "trustfile: EBADGID %ld %s (group member not trusted)\n", 
+		  (UID_CAST)stbuf.st_gid, fexp);
+	  fprintf(stderr, "This file/directory is group writeable, and one of the group members\n");
+	  fprintf(stderr, "is not in samhains list of trusted users.\n"); 
+	  fprintf(stderr, "Please run ./configure again with the option\n");
+	  fprintf(stderr, " ./configure [more options] --with-trusted=0,...,UID\n"); 
+	  fprintf(stderr, "where UID is the UID of the (yet) untrusted user.\n"); 
+	  fprintf(stderr, "---------------------------------------------\n");
+#endif 
 	  (void) strcpy(tf_path, fexp);                  /* known to fit  */
+	  tf_badgid = (gid_t) stbuf.st_gid;
 	  SL_IRETURN(SL_EBADGID, _("sl_trustfile"));
 	}
       /*
@@ -653,6 +908,13 @@ int sl_trustfile(char *fname, long *okusers, long *badusers)
 #endif
 	  )
 	{
+#ifdef TRUST_DEBUG
+	  fprintf(stderr, "---------------------------------------------\n");
+	  fprintf(stderr, "trustfile: EBADOTH (world writeable): %s\n", 
+		  fexp);
+	  fprintf(stderr, "This file/directory is world writeable.\n");
+	  fprintf(stderr, "---------------------------------------------\n");
+#endif 
 	  (void) strcpy(tf_path, fexp);                  /* known to fit  */
 	  SL_IRETURN(SL_EBADOTH, _("sl_trustfile"));
 	}
@@ -680,3 +942,44 @@ int sl_trustfile(char *fname, long *okusers, long *badusers)
   SL_IRETURN(SL_ENONE, _("sl_trustfile"));
 }
 
+#ifdef TRUST_MAIN
+int main (int argc, char * argv[])
+{
+  int status;
+#if defined(SH_WITH_SERVER)
+  struct passwd * pass;
+#endif
+
+  if (argc < 2) {
+    fprintf(stderr, "%s: Usage: %s <fullpath>\n", argv[0], argv[0]);
+    return 1;
+  }
+
+  tf_path[0] = '\0';
+#if defined(SH_WITH_SERVER)
+  pass = getpwnam(SH_IDENT);
+  if (pass != NULL)
+    tf_euid = pass->pw_uid;
+  else
+    {
+      fprintf(stderr, "trustfile: ERROR: getpwnam(%s) failed\n",
+	      SH_IDENT);
+      return 1;
+    }
+#else
+  tf_euid = geteuid();
+#endif
+
+  status = sl_trustfile(argv[1], NULL, NULL);
+  if (status != SL_ENONE)
+    {
+      fprintf(stderr, "trustfile: ERROR: not a trusted path: %s\n",
+	      argv[1]);
+      return 1;
+    }
+  return 0;
+}
+#endif
+
+
+
